;; Reading guile-gtk .defs files.

;; Copyright 2006, 2007, 2008 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (guile-lint defs)
  #:use-module (guile-lint common)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13))


;;-----------------------------------------------------------------------------

(define (within-range? lo hi x)
  (and (>= x lo)
       (<= x hi)))

;; trim PART from the right of STR
(define-public (string-trim-right-str str part)
  (if (string-suffix? part str)
      (string-drop-right str (string-length part))
      str))

(define (name-c->scheme name)
  (if (symbol? name)
      (set! name (symbol->string name)))
  (set! name (string-trim-right-str name "_interp"))
  (set! name (string-downcase name))
  (set! name (string-map (lambda (c)
			   (if (eqv? c #\_) #\- c))
			 name))
  (string->symbol name))

(define (name-mixedcase->func name)
  (if (symbol? name)
      (set! name (symbol->string name)))
  (set! name
	(regexp-substitute/global
	 #f "([a-z])([A-Z]+)" name
	 'pre
	 (lambda (m)
	   (string-append (match:substring m 1)
			  "-"
			  (match:substring m 2)))
	 'post))
  (string-downcase name))

(if (not (defined? 'hash-for-each))  ;; new in guile 1.8
    (define (hash-for-each proc table)
      (hash-fold (lambda (key value prev)
		   (proc key value))
		 #f table)))


;;-----------------------------------------------------------------------------
;; types builtin to build-guile-gtk-1.2

(define defs-type-table (make-hash-table 97))
(define defs-return-type-table (make-hash-table 97))

;; record a type which might be found in a .defs file
;; NAME is a symbol, like GdkWindow
;; TYPE and RETURN-TYPE are lint type objects
;;
(define* (defs-type-new name type #:optional (return-type type))
  (hashq-set! defs-type-table name type)
  (hashq-set! defs-return-type-table name type))

;; return a lint type object for NAME (a symbol) in a .defs file
(define-public (defs-type-find name)
  (or (hashq-ref defs-type-table name)
      (and (list>=? name 2)
	   (eq? 'list (first name))
	   (type-list-of (defs-type-find (second name))))
      (error "Unknown .defs file type:" name)))

;; return a lint type object for NAME (a symbol) in a .defs file, used as a
;; return type from a define-func
(define (defs-return-type-find name)
  (or (hashq-ref defs-return-type-table name)
      (error "Unknown .defs file type:" name)))

(define defs-type-or-f-table (make-hash-table 97))
(define-public (defs-type-or-f type)
  (or (hashq-ref defs-type-or-f-table type)
      (let ((type-or-f (type-or type type-f)))
	(hashq-set! defs-type-or-f-table type type-or-f)
	type-or-f)))


(defs-type-new 'none   	 type-unspecified)
(define-public type-gtk-atom type-symbol)
(defs-type-new 'atom   	 type-gtk-atom)
(define-public type-gtk-int type-exact-integer)
(defs-type-new 'int    	 type-gtk-int)
(defs-type-new 'int16  	 type-exact-integer)
(defs-type-new 'uint   	 type-exact-integer-nonneg)
(defs-type-new 'uint16 	 type-exact-integer-nonneg)
(defs-type-new 'uint32 	 type-exact-integer-nonneg)
(defs-type-new 'long   	 type-exact-integer)
(defs-type-new 'ulong 	 type-exact-integer-nonneg)
(defs-type-new 'float    type-real)
(defs-type-new 'double 	 type-real)
(defs-type-new 'bool   	 type-any              type-boolean)
(defs-type-new 'SCM    	 type-any)
(defs-type-new 'char     type-character)
(defs-type-new 'string   type-string)
(defs-type-new 'cstring  type-string)
(defs-type-new 'static_string  type-string)
(defs-type-new 'list   	 type-list-or-vector   type-list)
(defs-type-new 'slist  	 type-list-or-vector   type-list)
(defs-type-new 'cvec  	 type-list-or-vector   type-vector)
(defs-type-new 'cvecr  	 type-list-or-vector   type-vector)
(defs-type-new 'fvec  	 type-list-or-vector   type-vector)
(defs-type-new 'raw-data-r  (type-or type-f type-string type-vector))

;; should try to figure the actual size of a `gint' somehow
(define gtk-int-min (- (ash 1 31)))
(define gtk-int-max (1- (ash 1 31)))
(define (gtk-int? obj)
  (and (exact-integer? obj)
       (within-range? gtk-int-min gtk-int-max obj)))

(define (gdk-point? obj)
  (and (pair? obj)
       (gtk-int? (car obj))
       (gtk-int? (cdr obj))))
(define-public type-gdk-point (type-pair-of type-gtk-int))
(defs-type-new 'point type-gdk-point)

(define (gdk-rectangle? obj)
  (and (pair? obj)
       (gdk-point? (car obj))
       (gdk-point? (cdr obj))))
(define-public type-gtk-rect (type-pair-of type-gdk-point))
(defs-type-new 'rect type-gtk-rect)

(defs-type-new 'rect-null-ok
  (make-type "GtkRectangle or #f"
	     #:test (lambda (obj)
		      (or (gdk-rectangle? obj)
			  (eq? #f obj)))))

(defs-type-new 'full-callback   type-procedure)
(defs-type-new 'file-descriptor type-fd-or-port)
(define type-gtk-GtkType (make-type "GtkType"))
(defs-type-new 'type (type-or type-symbol type-gtk-GtkType) type-gtk-GtkType)


;;-----------------------------------------------------------------------------

(define (spec->type-name type)
  (cond ((and (list? type)
	      (eq? 'ret (first type)))
	 (second type))
	((list? type)
	 (first type))
	(else
	 type)))

(define (type-supertype! type supertype)
  (type-supertype-list! type (cons type (type-supertype-list supertype))))

(define-public (defs-file-process module filename)

  (define unhandled-list '())

  ;; object types have some forward references, so supertype settings are
  ;; deferred until the whole .defs file has been read
  ;;
  ;; object-supertype-alist is a list of (NAME . PARENT), both symbols

  (define object-supertype-alist '())

  (define (object-supertypes-now? elem)
    (not (assq (cdr elem) object-supertype-alist)))

  (define (object-supertypes-apply)
    (let more ()
      (if (not (null? object-supertype-alist))
	  (let ((elem (or (find object-supertypes-now? object-supertype-alist)
			  (error "oops no supertypes ready:"
				 object-supertype-alist))))
	    (set! object-supertype-alist (delq elem object-supertype-alist))
	    (type-supertype! (defs-type-find (car elem))
			     (defs-type-find (cdr elem)))
	    (more))))

    ;; redo the null-ok supertypes
    (hash-for-each
     (lambda (base-type type)
       (type-supertype-list! type
			     (cons type
				   (delete-duplicates
				    (append-map type-supertype-list
						(type-or-list type))
				    eq?))))
     defs-type-or-f-table))

  (define (make-flags-type name symbols)
    (make-type name
	       #:supertype (type-list-of type-symbol)
	       #:test (lambda (obj)
			(and (list? obj)
			     (lset<= eq? obj symbols)))))

  (define (make-enum-type name symbols)
    (make-type name
	       #:supertype type-symbol
	       #:test (lambda (obj)
			(and (symbol? obj)
			     (memq obj symbols)))))

  (define (lookup-func name-str)
    (let ((var (module-variable module (string->symbol name-str))))
      (or var
	  (error "gtk field accessor not found:" name-str module))
      (variable-ref var)))

  ;; FIELD-LIST is elements (FIELD-TYPE FIELD-NAME)
  (define (process-fields name type field-list)
    (set! name (name-mixedcase->func name))
    (for-each
     (lambda (elem)
       (receive-list (field-type field-name . options)
	   elem
	 (set! field-name (symbol->string (name-c->scheme field-name)))
	 (set! field-type (defs-type-find field-type))

	 ;; getter types
	 (let ((proc (lookup-func (string-append name "-" field-name))))
	   (lint-types! proc 'pure field-type type))

	 ;; setter types, if setter defined
	 (if (assoc-ref options 'setter)
	     (let ((proc (lookup-func (string-append name "-set-"
						     field-name))))
	       (lint-types! proc type-unspecified type field-type)))))
     field-list))

  (define (process-object obj)
    (case (first obj)
      ((define-type-alias)
       (receive-list (name existing-type)
	   (cdr obj)
	 (defs-type-new name (defs-type-find existing-type))))

      ((define-enum)
       (receive-list (name . values)
	   (cdr obj)
	 ;; strip option list like ((gettypefuncname "NULL")) in guile-gtk 2.0
	 (if (list? (first (first values)))
	     (set! values (cdr values)))
	 (defs-type-new name (make-enum-type name (map first values)))))

      ((define-flags)
       (receive-list (name . values)
	   (cdr obj)
	 (defs-type-new name (make-flags-type name (map first values)))))

      ((define-boxed define-struct)
       (receive-list (name . values)
	   (cdr obj)
	 (let ((type (make-type (symbol->string name))))
	   ;; conversion func means a string is also accepted, and interned
	   (if (assoc 'conversion values)
	       (set! type (type-or type-string type)))
	   (defs-type-new name type)

	   ;; field accessors
	   (and-let* ((field-list (assq-ref values 'fields)))
	     (process-fields name type field-list)))))

      ((define-boxed-union)
       (receive-list (name types)
	   (cdr obj)
	 (defs-type-new name (apply type-or (map defs-type-find types)))))

      ((define-func)
       (receive-list (name return args . flags)
	   (cdr obj)
	 (set! name (cond ((assq 'scm-name flags)
			   => (lambda (elem)
				(string->symbol (second elem))))
			  (else
			   (name-c->scheme name))))
	 (let ((retattrs '()))
	   (if (list? return)
	       (begin
		 (set! retattrs (cdr return))
		 (set! return (first return))))
	   ;; ENHANCE-ME: if multi-value return types are the same, then use
	   ;; `type-list-of'
	   (if (assq 'values retattrs)
	       (set! return type-list)
	       (set! return (defs-return-type-find (spec->type-name return)))))

	 (let ((var (module-variable module name)))
	   (if var
	       (let ((proc (variable-ref var)))
		 (apply lint-types-proc!
			name proc return
			(map (lambda (arg)
			       (let ((type (defs-type-find (spec->type-name
							    (first arg)))))
				 (if (member '(null-ok) (cddr arg))
				     (set! type (defs-type-or-f type)))
				 type))
			     args))
		 (if (lint-arity-any-optional? proc)
		     (apply lint-defaults!
			    proc
			    (map (lambda (arg)
				   (and-let* ((c-def (assq-ref '= (cddr arg))))
				     (c-default->scheme c-def)))
				 args))))
	       (lint-warn obj "warning, defs function not found: ~a"
			  name)))))

      ((define-object)
       (receive-list (name parent . attributes)
	   (cdr obj)
	 (let ((type (make-type (symbol->string name))))
	   (defs-type-new name type)
	   (if (not (null? parent))
	       (set! object-supertype-alist (acons name (first parent)
						   object-supertype-alist)))

	   ;; field accessors
	   (and-let* ((field-list (assq-ref attributes 'fields)))
	     (process-fields name type field-list)))))

      (else
       (set! unhandled-list
	     (lset-adjoin eq? unhandled-list (first obj))))))

  ;; allow for defs file to be perhaps missing from a packaged version
  (let ((port (false-if-exception (open-input-file filename))))
    (if port
	(begin
	  (do ((obj (read port) (read port)))
	      ((eof-object? obj))
	    (process-object obj))
	  (close-port port))
	(lint-warn #f "warning, defs file not found: ~a" filename)))

  ;; (format #t ";; unhandled: ~a\n" unhandled-list)
  (object-supertypes-apply))

(define (lint-arity-any-optional? proc)
  (and-let* ((arity (lint-arity proc)))
    (positive? (second arity))))

;; C-DEF is a value from a default (= "SOMETHING"), eg. "NULL" or "TRUE"
;; return the corresponding scheme value, or #f if unrecognised
(define (c-default->scheme c-def)
  ;; can have symbol NULL instead of string
  (if (symbol? c-def)
      (set! c-def (symbol->string c-def)))
  (cond ((number? c-def) ;; can have literal number instead of string number
	 => noop)
	((and (string? c-def)
	      (string->number c-def)) ;; integer like "-1"
	 => noop)
	((assoc c-def '(("NULL" . #f)
			("TRUE" . #t)
			("FALSE" . #f)
			("SCM_BOOL_F" . #f)
			;; could do something more general for options like
			;; this, but just help gtk_table_attach for now
			("GTK_EXPAND|GTK_FILL" . (expand fill)))
		=> cdr)
	 (else
	  (format (current-error-port) "guile-lint: warning, unrecognised guile-gtk defs file default: ~s\n" c-def)
	  #f))))
