;; Checking (ice-9 format).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (guile-lint ice-9 format)
  #:use-module (guile-lint common)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 format)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 receive)
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-13))


;;-----------------------------------------------------------------------------
;; helpers

(define-public (number>=2? obj)
  (and (number? obj) (>= obj 2)))

(define (number-positive? obj)
  (and (number? obj) (positive? obj)))

(define (number-non-negative? obj)
  (and (number? obj) (not (negative? obj))))


;;-----------------------------------------------------------------------------

(define lint-ice-9-format-param-type-alist
  `((,number?              . "number")
    (,char?                . "character")
    (,number-positive?     . "positive")
    (,number-non-negative? . "non-negative")
    (,number>=2?           . "number>=2")))

(define*-public (lint-ice-9-format-string where str arg-count env
					  #:key or-more)

  (define (char-name c)
    (if (char=? #\newline c) "\newline" c))

  (let ((arg-pos           0)  ;; 0 upwards, or #f for unknown
	(str-pos           0)  ;; within str
	(arg-used          (make-vector arg-count #f))
	(str-length        (string-length str))
	(done-not-enough   #f)
	(done-before-first #f)
	(done-lost-arg-pos #f)

	;; list of:
	;; ('conditional MODIFIER CLAUSECOUNT)
	;; ('iteration   STARTPOS)
	;; ('case)
	(nest-list         '()))

    ;; return character from str and advance str-pos, or #f at end of string
    (define (next-char)
      (and (< str-pos str-length)
	   (begin
	     (set! str-pos (1+ str-pos))
	     (string-ref str (1- str-pos)))))

    ;; if nest-list contains some element of TYPE return it, or #f if none
    (define (nest-list-contains? type)
      (find (lambda (elem)
	      (eq? type (first elem)))
	    nest-list))

    ;; return true if nest-list contains only TYPE elements
    (define (nest-list-only? type)
      (every (lambda (elem)
	       (eq? type (first elem)))
	     nest-list))

    ;; if nest-list first element is TYPE return it, or #f otherwise
    (define (nest-list-first? type)
      (and (not (null? nest-list))
	   (eq? type (caar nest-list))))

    ;; find and delete TYPE element, and return it
    (define (nest-list-remove type)
      (let ((elem (nest-list-contains? type)))
	(set! nest-list (delq! elem nest-list))
	elem))

    ;; find and handle the next tilde conversion
    (let next-conversion ()
      (and-let* ((tilde-pos (string-index str #\~ str-pos)))
	(let ((modifier   #f)  ;; #f 'at 'colon or 'colon-at
	      (param-list '())
	      (seen-param #f)) ;; when comma doesn't mean an empty param

	  (define (add-param p)
	    (set! seen-param #t)
	    (set! param-list (append! param-list (list p))))

	  (define (modifiers-after-params)
	    (if modifier
		(lint-error where "`format' modifier should be after parameter(s): ~s" (tilde-str))))

	  (define (no-modifiers)
	    (if modifier
		(lint-error where "`format' no modifier allowed: ~s"
			    (tilde-str))))

	  ;; use current arg, move to the next
	  (define (take-arg)
	    ;; Lost pos if take in a conditional, because don't know which
	    ;; clause is used.
	    ;; ENHANCE-ME: Count what arguments are taken by the various
	    ;; clauses, and if there's a default clause then that would
	    ;; establish a minimum required.
	    (if (nest-list-contains? 'conditional)
		(lost-arg-pos))
	    (if (and arg-pos
		     (nest-list-only? 'case))
		(begin
		  (cond ((>= arg-pos arg-count)
			 (if (and (not or-more)
				  (not done-not-enough))
			     (begin
			       (set! done-not-enough #t)
			       (lint-error where "`format' not enough arguments, at ~s" (tilde-str)))))
			((< arg-pos 0)
			 (if (not done-before-first)
			     (begin
			       (set! done-before-first #t)
			       (lint-error where "`format' stepped back before first argument: ~s" (tilde-str)))))
			(else
			 (vector-set! arg-used arg-pos #t)))
		  (move-arg 1))))

	  ;; move current arg, +ve or -ve
	  (define (move-arg n)
	    (if (and arg-pos
		     (not (nest-list-contains? 'iteration)))
		(set! arg-pos (+ n arg-pos))))

	  ;; current tilde conversion string
	  (define (tilde-str)
	    (substring str tilde-pos str-pos))

	  ;; use all remaining args, don't move
	  (define (use-rest)
	    (do ((i arg-pos (1+ i)))
		((>= i arg-count))
	      (vector-set! arg-used i #t)))

	  (define (lost-arg-pos)
	    (set! arg-pos #f)
	    (vector-fill! arg-used #t)
	    (if (not done-lost-arg-pos)
		(begin
		  (set! done-lost-arg-pos #t)
		  (if option-verbose
		      (lint-error where "`format' argument position lost at: ~s"
				  (tilde-str))))))


	  (set! str-pos (1+ tilde-pos))

	  ;; look for more within a tilde conversion, ie. parameter,
	  ;; modifier, or the conversion code
	  (let more-conversion ()
	    (let ((c (next-char)))

	      (define (parameter-separator)
		(if seen-param
		    (lint-error where "`format' missing \",\" between parameters: ~s" (tilde-str))))

	      (define (no-parameters)
		(or (null? param-list)
		    (lint-error where "`format' no parameters allowed for `~a': ~s" (char-name c) (tilde-str))))

	      (define (param-types . rest)
		(let ((max-params (quotient (length rest) 2)))
		  (if (> (length param-list) max-params)
		      (lint-error where "`format' maximum ~a parameters for `~a': ~s" max-params (char-name c) (tilde-str))))
		(receive (desc-list type-list)
		    (partition string? rest)
		  (for-each
		   (lambda (n type? desc value)
		     (cond ((or (eq? 'variable value)  ;; unknown
				(not value)            ;; #f default
				(type? value)))        ;; good type
			   (else
			    (lint-error where "`format' bad parameter ~a to `~a', expect ~a ~s: ~s"
					n (char-name c)
					(assq-ref lint-ice-9-format-param-type-alist
						  type?) desc
						  (tilde-str)))))
		   '(1 2 3 4 5 6 7 8 9 10) type-list desc-list param-list)))

	      (define (param-get n)
		(if (list>=? param-list n)
		    (list-ref param-list (1- n))
		    #f))

	      (define (valid-modifiers lst)
		(or (memq modifier lst)
		    (lint-error where "`format' invalid modifier for `~a': ~s"
				(char-name c) (tilde-str))))

	      (case (and c (char-downcase c))
		((#f)    ;; end of string
		 (lint-error where "`format' end of string within `~~' sequence: ~s" (tilde-str)))

		((#\:)   ;; modifier :
		 (if (memq modifier '(colon colon-at))
		     (lint-error where "`format' duplicate `:' modifier: ~s"
				 (tilde-str)))
		 (set! modifier (if (eq? modifier 'at) 'colon-at 'colon))
		 (more-conversion))

		((#\@)   ;; modifier @
		 (if (memq modifier '(at colon-at))
		     (lint-error where "`format' duplicate `@' modifier: ~s"
				 (tilde-str)))
		 (set! modifier (if (eq? modifier 'colon) 'colon-at 'at))
		 (more-conversion))

		((#\#)   ;; parameter is number of remaining args
		 (modifiers-after-params)
		 (parameter-separator)
		 (add-param (if (and arg-pos
				     (not (nest-list-contains? 'iteration)))
				(- arg-count arg-pos)
				'variable))
		 (more-conversion))

		((#\v)   ;; parameter from next arg
		 (modifiers-after-params)
		 (parameter-separator)
		 (add-param 'variable)
		 (take-arg)
		 (more-conversion))

		((#\')   ;; parameter from next format char
		 (modifiers-after-params)
		 (parameter-separator)
		 ;; more-conversion will report if we're at end of string
		 (add-param (next-char))
		 (set! seen-param #t)
		 (more-conversion))

		((#\0 #\1 #\2 #\3 #\4 #\5   ;; numeric parameter
		  #\6 #\7 #\8 #\9 #\+ #\-)
		 (modifiers-after-params)
		 (parameter-separator)
		 (let* ((num-end (or (string-skip str char-numeric? str-pos)
				     str-length))
			(num-str (substring str (1- str-pos) num-end))
			(num     (string->number num-str)))
		   ;; just "+" or "-" isn't fatal in format, but it's
		   ;; unlikely to be intentional
		   (if (not num)
		       (lint-error where "`format' incomplete numeric parameter: ~s" num-str))
		   (add-param num)
		   (set! str-pos num-end))
		 (more-conversion))

		((#\,)   ;; parameter separator
		 (modifiers-after-params)
		 (if (not seen-param)  ;; nothing since last comma (or start)
		     (add-param #f))   ;; gives empty param
		 (set! seen-param #f)
		 (more-conversion))

		((#\()   ;; case conversion begin
		 ;; default=downcase, @=cap first word, :=cap each word,
		 ;; :@=upcase
		 (if (nest-list-contains? 'case)
		     (lint-error where "`format' case conversions don't nest (currently): ~s" (tilde-str)))
		 (no-parameters)
		 (set! nest-list (cons '(case) nest-list))
		 (next-conversion))

		((#\))   ;; case conversion end
		 (no-parameters)
		 (cond ((nest-list-first? 'case)
			)  ;; good
		       ((nest-list-contains? 'case)
			(lint-error where "`format' case conversion end improper nesting: ~s" (tilde-str)))
		       (else
			(lint-error where "`format' case conversion end without begin: ~s" (tilde-str))))
		 (nest-list-remove 'case)
		 (next-conversion))

		((#\|    ;; formfeed
		  #\%    ;; newline
		  #\&    ;; newline if not already
		  #\_    ;; space
		  #\/    ;; tab char
		  #\~)   ;; tilde
		 (param-types number-non-negative? "repetition count")
		 (no-modifiers)
		 (next-conversion))

		((#\!)   ;; flush
		 (no-parameters)
		 (no-modifiers)
		 (next-conversion))

		((#\newline)  ;; continuation
		 (no-parameters)
		 (valid-modifiers '(#f      ;; no print
				    at      ;; print newline
				    colon)) ;; print all whitespace
		 ;; whitespace skipped by next-conversion like any non-tilde
		 (next-conversion))

		((#\q)   ;; inquiry messages
		 (no-parameters)
		 (valid-modifiers '(#f       ;; copyright
				    colon))  ;; version
		 (next-conversion))

		((#\t)   ;; tab stop
		 (param-types number-non-negative? "column number"
			      number-non-negative? "column increment"
			      char?                "pad char")
		 (valid-modifiers '(#f   ;; absolute
				    at)) ;; relative
		 (next-conversion))

		((#\y)   ;; pretty print
		 (no-parameters)
		 (no-modifiers)
		 (take-arg)
		 (next-conversion))

		((#\a    ;; display
		  #\s)   ;; write
		 ;; @ = pad left, : = read-proof style on internal objs
		 (param-types number-non-negative? "min width"
			      number-non-negative? "column increment"
			      number-non-negative? "minimum pad"
			      char?                "pad char")
		 (if (and (param-get 4)
			  (not (param-get 1)))
		     (lint-error where "`format' pad char parameter not used without minwidth: ~s" (tilde-str)))
		 (take-arg)
		 (next-conversion))

		((#\f    ;; fixed point
		  #\i)   ;; complex
		 (param-types number-non-negative? "width"
			      number-non-negative? "decimals"
			      number?              "scale"
			      char?                "overflow char"
			      char?                "pad char")
		 (valid-modifiers '(#f
				    at))  ;; show "+"
		 (let ((width        (param-get 1))
		       ;; (decimals     (param-get 2))
		       ;; (scale        (param-get 3))
		       (overflowchar (param-get 4))
		       (padchar      (param-get 5)))
		   (if (and padchar
			    (not width))
		       (lint-error where "`format' pad char parameter not used without width: ~s" (tilde-str)))
		   (if (and overflowchar
			    (not width))
		       (lint-error where "`format' overflow char parameter not used without width: ~s" (tilde-str))))
		 (take-arg)
		 (next-conversion))

		((#\e    ;; scientific
		  #\g)   ;; general
		 (param-types number-non-negative? "width"
			      number-non-negative? "mant digits"
			      number-non-negative? "exponent digits"
			      number?              "int digits"
			      char?                "overflow char"
			      char?                "pad char"
			      char?                "exponent char")
		 (valid-modifiers '(#f
				    at))  ;; show "+"
		 (take-arg)
		 (next-conversion))

		((#\$)    ;; fixed point two decimals
		 ;; @ = show "+", : = sign before padding
		 (param-types number-non-negative? "decimals"
			      number-non-negative? "integer digits"
			      number-non-negative? "width"
			      char?                "pad char")
		 (take-arg)
		 (next-conversion))

		((#\c)    ;; char
		 ;; FIXME: Should numbers be allowed as the parameter, to
		 ;; get ascii values?  Is this really meant to take a
		 ;; parameter at all?
		 (param-types char? "char to print")
		 (valid-modifiers '(#f      ;; direct output
				    at      ;; `write' style
				    colon)) ;; ^X style
		 (if (null? param-list)
		     (take-arg))
		 (next-conversion))

		((#\p)    ;; plural
		 ;; :=reuse prev arg, @="y"/"ies"
		 (if (memq modifier '(colon colon-at))
		     (move-arg -1))
		 (take-arg)
		 (next-conversion))

		((#\b    ;; binary
		  #\d    ;; decimal
		  #\o    ;; octal
		  #\x)   ;; hex
		 ;; :=commas, @="+" on positives
		 (param-types number-non-negative? "min width"
			      char?                "pad char"
			      char?                "comma char"
			      number-positive?     "comma width")
		 (let ((minwidth   (param-get 1))
		       (padchar    (param-get 2))
		       (commachar  (param-get 3))
		       (commawidth (param-get 4)))
		   (if (and (number? minwidth) ;; literal number
			    (< minwidth 2))
		       (lint-error where "`format' minwidth < 2 does nothing: ~s" (tilde-str)))
		   (if (and padchar
			    (not minwidth))
		       (lint-error where "`format' pad char parameter not used without minwidth: ~s" (tilde-str)))
		   (if (and (or commachar commawidth)
			    (not (memq modifier '(colon colon-at))))
		       (lint-error where "`format' comma parameters not used without `:' modifier: ~s" (tilde-str))))
		 (take-arg)
		 (next-conversion))

		((#\r)   ;; number
		 ;; no params is words, :=ordinal, @=roman :@=old roman
		 ;; otherwise radix etc, :=commas, @="+" on positives
		 (param-types number>=2?           "radix"
			      number-non-negative? "min width"
			      char?                "pad char"
			      char?                "comma char"
			      number-positive?     "comma width")
		 (let ((minwidth   (param-get 2))
		       (padchar    (param-get 3))
		       (commachar  (param-get 4))
		       (commawidth (param-get 5)))
		   (if (and (number? minwidth) ;; literal number
			    (< minwidth 2))
		       (lint-error where "`format' minwidth < 2 does nothing: ~s" (tilde-str)))
		   (if (and padchar
			    (not minwidth))
		       (lint-error where "`format' pad char parameter not used without minwidth: ~s" (tilde-str)))
		   (if (and (or commachar commawidth)
			    (not (memq modifier '(colon colon-at))))
		       (lint-error where "`format' comma parameters not used without `:' modifier: ~s" (tilde-str))))
		 (take-arg)
		 (next-conversion))

		((#\?    ;; sub-format STR and LST
		  #\k)   ;; same
		 (no-parameters)
		 (valid-modifiers '(#f    ;; take fmt + arglist
				    at))  ;; take fmt then rest is arglist
		 (take-arg)      ;; fmt
		 (if (not modifier)
		     (take-arg)    ;; arglist
		     (use-rest))   ;; args used, but pos unchanged
		 (next-conversion))

		((#\*)    ;; argument jump
		 (param-types number? "move")
		 (valid-modifiers '(#f       ;; forward
				    at       ;; absolute
				    colon))  ;; backward
		 ;; move by param, default 1
		 (if (and arg-pos
			  (not (nest-list-contains? 'iteration)))
		     (let ((n (if (null? param-list) 1 (first param-list))))
		       (if (number? n)
			   (case modifier
			     ((at)    ;; absolute
			      (move-arg (- n arg-pos)))
			     ((colon) ;; backward
			      (move-arg (- n)))
			     (else    ;; forward
			      (move-arg n)))
			   (lost-arg-pos)))) ;; non-literal move
		 (next-conversion))

		((#\[)    ;; conditional begin
		 (param-types number-non-negative? "clause selector")
		 (valid-modifiers '(#f      ;; num-case
				    at      ;; if-then/consume
				    colon)) ;; if-else-then
		 (if (null? param-list)
		     (take-arg))  ;; selector if no param
		 (if (eq? modifier 'at)
		     (lost-arg-pos))  ;; don't know if consume or not
		 (set! nest-list (cons (list 'conditional modifier 1)
				       nest-list))
		 (next-conversion))

		((#\;)   ;; conditional separator
		 (no-parameters)
		 (valid-modifiers '(#f
				    colon))  ;; default clause
		 (cond ((nest-list-first? 'conditional)
			) ;; good
		       ((nest-list-contains? 'conditional)
			(lint-error where "`format' conditional separator improper nesting: ~s" (tilde-str)))
		       (else
			(lint-error where "`format' conditional separator without begin: ~s" (tilde-str))))
		 (and-let* ((elem (nest-list-contains? 'conditional)))
		   (case (second elem) ;; modifier
		     ((at)
		      (if (= 1 (third elem))
			  (lint-error where "`format' conditional `~~@[' expects just 1 clause")))
		     ((colon)
		      (if (= 2 (third elem))
			  (lint-error where "`format' conditional `~~:[' expects just 2 clauses"))))
		   (set-third! elem (1+ (third elem))))
		 (next-conversion))

		((#\])    ;; conditional end
		 (no-parameters)
		 (no-modifiers)
		 (cond ((nest-list-first? 'conditional)
			) ;; good
		       ((nest-list-contains? 'conditional)
			(lint-error where "`format' conditional end improper nesting: ~s" (tilde-str)))
		       (else
			(lint-error where "`format' conditional end without begin: ~s" (tilde-str))))
		 (and-let* ((elem (nest-list-remove 'conditional)))
		   (if (and (eq? 'colon (second elem)) ;; modifier
			    (< (third elem) 2))
		       (lint-error where "`format' conditional `~~:[' expects 2 clauses")))
		 (next-conversion))

		((#\{)    ;; iteration begin
		 (param-types number-non-negative? "repetitions")
		 ;; actually the args for these are taken after the possible
		 ;; fmt in ~}, but that doesn't matter to our counting
		 (case modifier
		   ((colon)
		    ;; sublists, arg is a list, each elem being a list of
		    ;; arguments to fmt
		    (take-arg))
		   ((at)
		    ;; rest-args, rep fmt until it uses all remaining args
		    (lost-arg-pos))
		   ((colon-at)
		    ;; rest-sublists, remaining args are lists, rep once for
		    ;; each passing as args
		    (lost-arg-pos))
		   (else
		    ;; list, arg is a list, rep until elems used up by fmt
		    (take-arg)))
		 (set! nest-list (cons (list 'iteration str-pos) nest-list))
		 (next-conversion))

		((#\})    ;; iteration end
		 (no-parameters)
		 (valid-modifiers '(#f
				    colon))  ;; at least one iteration (?)
		 (cond ((nest-list-first? 'iteration)
			) ;; good
		       ((nest-list-contains? 'iteration)
			(lint-error where "`format' iteration end improper nesting: ~s" (tilde-str)))
		       (else
			(lint-error where "`format' iteration end without begin: ~s" (tilde-str))))
		 (and-let* ((elem (nest-list-remove 'iteration)))
		   ;; ~{~} with nothing in between means fmt arg
		   (if (= tilde-pos (second elem))
		       (take-arg)))
		 (next-conversion))

		((#\^)    ;; up and out
		 (param-types number? "arg"  ;; single equal 0
			      number? "arg"  ;; two equal
			      number? "arg") ;; three each <=
		 (no-modifiers)
		 (cond ((null? nest-list)
			;; possibly terminate formatting
			(lost-arg-pos))
		       ((eq? 'iteration (caar nest-list))
			;; terminate iteration
			)
		       ((eq? 'conditional (caar nest-list))
			;; terminate conditional clause
			(lost-arg-pos))
		       ((eq? 'case (caar nest-list))
			;; stop converting, but keep formatting
			;; FIXME: Is this supposed to terminate any
			;; containing iteration/conditional too?
			))
		 (next-conversion))

		(else
		 (lint-error where "`format' unrecognised ~~ escape `~a': ~s"
			     c (tilde-str)))))))))

    (let ((nest-open (delete-duplicates! (map first nest-list))))
      (if (memq 'iteration nest-open)
	  (lint-error where "`format' iteration begin `~~{' without end"))
      (if (memq 'conditional nest-open)
	  (lint-error where "`format' conditional begin `~~[' without end"))
      ;; format.scm isn't worried about unclosed case conversion, but let's
      ;; be stricter about that
      (if (memq 'case nest-open)
	  (lint-error where "`format' case conversion begin `~~(' without end")))

    (let* ((arg-used-list (vector->list arg-used))
	   (tail          (drop-while noop arg-used-list)))
      (cond ((and (not done-not-enough) ;; don't print not enough then too many
		  (not (null? tail))    ;; if some #f's
		  (every not tail))     ;; if only #f's after leading #t's
	     ;; this is the usual message
	     (lint-error where "`format' too many arguments, want ~a got ~a~a"
			 (- arg-count (length tail))
			 arg-count (if or-more " or more" "")))
	    ((any not arg-used-list)
	     ;; this message is when ~* has been jumping about
	     (lint-error where "`format' some arguments unused: ~{~d~^ ~}"
			 (remove! (lambda (i)
				    (vector-ref arg-used (1- i)))
				  (iota arg-count 1))))))))


;; return true if `simple-format' suffices to print string STR 
(define (lint-simple-format-suffices? str)
  (regexp-exec lint-simple-format-regexp str))
(define lint-simple-format-regexp
  (make-regexp "^[^~]*(~[sa][^~]*)*$" regexp/icase))

(define (lint-ice-9-format-needed)
  (set-object-property! lint-current-module-env
			'lint-ice-9-format-needed #t))

;; only (0 0 #t) in ice-9 format
;; the code allows the port argument to be omitted (start with a format
;; string), but that's not documented, so don't allow it
(lint-arity! format '(2 0 #t))

(lint-types! format  (type-or type-string type-unspecified)
	     (type-or type-port type-boolean type-number) type-string type-any)

(define (lint-ice-9-format-string-from-format-call where str)
  (or (lint-simple-format-suffices? str)
      (lint-ice-9-format-needed))
  (if (not (string-any #\~ str))
      (lint-warn where "hint, `format' without `~~' can use `display': ~s" str)))

(define (lint-ice-9-format where now expr env)
  (if (not (null? (string-gettext-list (second expr) env)))
      (lint-error where "`format' probably missing port argument: ~s" expr))

  (let ((fmt-list (string-gettext-list (third expr) env)))
    (for-each
     (lambda (str)
       (lint-ice-9-format-string-from-format-call where str)
       (lint-ice-9-format-string where str (length (cdddr expr)) env))
     fmt-list)
    ;; if the format isn't a literal string (or gettext form) then can't
    ;; tell if ice-9 format is really needed, assume it is
    (if (null? fmt-list)
	(lint-ice-9-format-needed)))

  (lint-list where now (cdr expr) env))

(define (lint-ice-9-format-apply where now expr env)
  ;; `(apply format port fmt X...X lst)' has fixed args of len-5
  (if (length>=? expr 5)
      (let ((fmt-list (string-gettext-list (third expr) env)))
	(for-each
	 (lambda (str)
	   (lint-ice-9-format-string-from-format-call where str)
	   (lint-ice-9-format-string where (fourth expr) (- (length expr) 5)
				     env #:or-more #t))
	 fmt-list)
	;; if the format isn't a literal string (or gettext form) then can't
	;; tell if ice-9 format is really needed, assume it is
	(if (null? fmt-list)
	    (lint-ice-9-format-needed)))))

(lint-handler! format lint-ice-9-format)
(lint-apply-handler! format lint-ice-9-format-apply)

(add-hook! lint-finalize-env-hook
	   (lambda (env)
	     (or (object-property env 'lint-ice-9-format-needed)
		 (and-let* ((usemod (find (lambda (usemod)
					    (equal? '(ice-9 format)
						    (module-name
						     (usemod-module usemod))))
					  (env-usemod-list env))))
		   (lint-warn (usemod-where usemod) "hint, `(ice-9 format)' not needed, `simple-format' is enough for all uses")))))
