;; Checking (oop goops).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (guile-lint oop goops)
  #:use-module (guile-lint common)
  #:use-module (oop goops)
  #:use-module (srfi srfi-1))


;;-----------------------------------------------------------------------------
;; helpers

;; like `for-each' but taking two elements at a time from LST
(define (for-each-two proc lst)
  (if (not (null? lst))
      (begin
	(proc (first lst) (second lst))
	(for-each-two proc (cddr lst)))))


;;-----------------------------------------------------------------------------
;; various types

(define (class? obj)
  (is-a? obj <class>))

(define-public type-goops-class
  (make-type "goops class"
	     #:test class?))

(lint-types! class-of      	 type-goops-class type-any)
(lint-types! instance?      	 type-boolean type-any)
(lint-types! is-a?      	 type-boolean type-any type-goops-class)


;;-----------------------------------------------------------------------------
;; define-class

(lint-arity! define-class '(2 0 #t))

(define (form-define-class-slot where expr)
  (set! where (where-follow where expr))
  (and (or (list>=? expr 1)
	   (lint-error where "`define-class' slot descriptor must be a list: ~s" expr))
       (or (symbol? (first expr))
	   (lint-error where "`define-class' slot name must be a symbol: ~s" expr))
       (let more ((options (cdr expr)))
	 (cond ((null? options)
		#t) ;; good
	       ((not (keyword? (first options)))
		(lint-error (where-follow where (first options)) "`define-class' slot option not a keyword: ~s" (first options)))
	       ((null? (cdr options))
		(lint-error where "`define-class' missing value for slot option: ~s" (first options)))
	       (else
		(more (cddr options)))))))

(define (form-define-class-options where expr)
  #t)

(define (form-define-class where expr)
  (and? (or (symbol? (second expr))
	    (lint-error where "`define-class' class name should be a symbol: ~s" (second expr)))
	(apply and? (map (lambda (elem)
			   (form-define-class-slot where elem))
			 (take-while pair? (cddr expr))))
	(form-define-class-options where (drop-while pair? (cddr expr)))))

(define (lint-define-class where now expr env)
  (if (form-define-class where expr)
      (begin
	(env-bind-one env where (second expr) #f)
	(lint-list where now (third expr) env)
	(for-each (lambda (slot-desc)
		    (let ((where     (where-follow where slot-desc))
			  (name      (first slot-desc))
			  (virtual   #f)
			  (slot-ref  #f)
			  (slot-set! #f)
			  (seen      '()))
		      (for-each-two
		       (lambda (key arg)
			 (if (memq key seen)
			     (lint-error where "`define-class' slot `~a' duplicate slot option: ~s" name key)
			     (set! seen (cons key seen)))
			 (case key
			   ((#:allocation)
			    (if (not (memq arg '(#:instance #:class #:each-subclass #:virtual)))
				(lint-error (where-follow where arg) "`define-class' slot `~a' unknown #:allocation choice: ~s" name arg))
			    (if (eq? arg #:virtual)
				(set! virtual #t)))
			   ((#:accessor)
			    (env-bind-one env where arg
					  (set-procedure-name!
					   arg (make-procedure-with-setter
						(lambda (obj) #f)
						(lambda (obj arg) #f)))))
			   ((#:getter)
			    (env-bind-one env where arg
					  (set-procedure-name!
					   arg (lambda (obj) #f))))
			   ((#:init-form)
			    ;; evaluated later
			    (lint-object where #f arg env))
			   ((#:init-keyword)
			    (if (not (keyword? arg))
				(lint-error (where-follow where arg) "`define-class' slot `~a' #:init-keyword should be a keyword: ~s" name arg)))
			   ((#:init-thunk)
			    ;; evaluated later
			    (lint-expr where #f (list arg) env))
			   ((#:init-value)
			    ;; evaluated now
			    (lint-object where now arg env))
			   ((#:setter)
			    (env-bind-one env where arg
					  (set-procedure-name!
					   arg (lambda (obj arg) #f))))
			   ((#:slot-ref)
			    (set! slot-ref #t)
			    (env-reference-add env where arg '(obj)))
			   ((#:slot-set!)
			    (set! slot-set! #t)
			    (env-reference-add env where arg '(obj val)))
			   (else
			    (lint-error where "`define-class' slot `~a' unknown option: ~s" name key))))
		       (cdr slot-desc))
		      (if (and virtual (not slot-ref))
			  (lint-error where "`define-class' virtual slot `~a' needs slot-ref" name))
		      (if (and virtual (not slot-set!))
			  (lint-error where "`define-class' virtual slot `~a' needs slot-set!" name))
		      (if (and slot-ref (not virtual))
			  (lint-error where "`define-class' slot `~a' #:slot-ref unused" name))
		      (if (and slot-set! (not virtual))
			  (lint-error where "`define-class' slot `~a' #:slot-set! unused" name))))
		  (take-while pair? (cdddr expr))))))

(lint-handler! define-class lint-define-class)


;;-----------------------------------------------------------------------------
;; define-method

(lint-arity! define-method '(1 0 #t))

(define (form-define-method where expr)
  (let ((formals (second expr)))
    (and? (and (or (not (null? formals))
		   (lint-error where "`define-method' need method name: ~s" formals))
	       (or (symbol? (first formals))
		   (lint-error where "`define-method' method name must be a symbol: ~s" (first formals))))
	  (let ((result #t))
	    (let more ((formals (cdr formals)))
	      (cond ((null? formals))   ;; end of list
		    ((pair? formals)
		     (let ((arg (first formals)))
		       (or (symbol? arg)
			   (and (list? arg)
				(= 2 (length arg))
				(symbol? (first arg))
				(symbol? (second arg)))
			   (begin
			     (lint-error (where-follow where arg) "`define-method' formal parameter must be symbol or (symbol type): ~s" arg)
			     (set! result #f))))
		     (more (cdr formals)))
		    ((symbol? formals)) ;; . rest arg
		    (else
		     (lint-error (where-follow where formals) "`define-method' bad \"rest\" parameter: ~s" formals)
		     (set! result #f)))
	      result)))))

;; FIXME: env-bind is wrong
(define (lint-define-method where now expr env)
  (if (form-define-method where expr)
      (let ((params (second expr)))
	;; an existing imported binding will be extended rather than
	;; creating a new binding
	(if (not (env-binding env (first params)))
	    (env-bind env where (first params) #t))
	(lint-subenv env
		     (lambda (subenv)
		       (env-bind subenv where 'next-method #t #t)
		       (for-each-improper
			(lambda (formal)
			  (if (symbol? formal)
			      (env-bind subenv where formal #t #t)
			      (env-bind subenv where (first formal) #t #t)))
			params)
		       (lint-list where #f (cddr expr) subenv))))))

(lint-handler! define-method lint-define-method)


;;-----------------------------------------------------------------------------
;; goops-error

(lint-handler! goops-error (make-lint-format 'simple-format 0))


;;-----------------------------------------------------------------------------
;; make

(set-procedure-property! make 'arglist '((class) () (key) #t #f))


;;-----------------------------------------------------------------------------
;; process-define-generic
;;
;; This arises from the macro expansion of `define-generic' in a top-level
;; environment.  It creates a binding if not already an import.

(if (defined? 'process-define-generic) ;; new in 1.8
    (begin
      (lint-types! process-define-generic type-unspecified type-symbol)

      (define (lint-process-define-generic where now expr env)
	(let ((sym (lint-value (second expr))))
	  (if (symbol? sym)
	      (if (not (env-binding env sym))
		  (env-bind env where sym (make <generic> #:name sym))))))

      (lint-handler! process-define-generic lint-process-define-generic)))

