;; Guile-Lint html manual href link munging.

;; Copyright 2005, 2006, 2007 Kevin Ryde
;;
;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: guile -s mung-html.scm filename.html ...
;;
;; This program modifies the given "filename.html" etc input files, changing
;; links like href="../foo.html#Node", as generated by makeinfo, into links
;; to the GNU web site, or wherever.
;;
;; The links recognised are just for the guile-lint manual, there's no
;; attempt at being completely general.
;;

(use-modules (ice-9 regex)
	     (ice-9 rw)
	     (srfi srfi-13))


;; return string read from PORT up to eof
;; (copied from misc.scm)
(define (read-string-to-eof port)
  (let ((buf (make-string 4096)))
    (do ((lst   '()
		(cons (substring buf 0 nread) lst))
	 (nread (read-string!/partial buf port)
		(read-string!/partial buf port)))
	((not nread)                       ;; stop at #f
	 (string-concatenate-reverse lst)) ;; join parts
      ;; allow for 0 meaning EWOULDBLOCK
      ;; if 0 is in fact from a signal then this select does no harm
      (if (= 0 nread)
	  (select (list port) '() '())))))

;; return STR with all occurrances of OLD changed to NEW
(define (string-substitute/global str old new)
  (let more ((upto 0)
	     (lst  '()))
    (let ((match (string-contains str old upto)))
      (if match
	  (more (+ match (string-length old))
		(cons* new (substring str upto match) lst))
	  (string-concatenate-reverse (cons (substring str upto) lst))))))

;; return STR with all matches of REGEXP changed according to SUBST-BITS
;; SUBST-BITS is `regexp-substitute' parameters
;;
;; writing biggish blocks to a string port is very slow in 1.6.7, avoid that
;;
(define (string-regexp-substitute/global str regexp . subst-bits)
  (let ((re (make-regexp regexp)))
    (let more ((upto 0)
	       (lst  '()))
      (let ((m (regexp-exec re str upto)))
	(if m
	    (begin
	      (more (match:end m)
		    (cons* (apply regexp-substitute #f m subst-bits)
			   (substring str upto (match:start m))
			   lst)))
	    (string-concatenate-reverse (cons (substring str upto) lst)))))))

(for-each
 (lambda (filename)
   (let ((body (call-with-input-file filename read-string-to-eof)))

     ;; remove doubled names in "Top" links
     ;; this must be before munging those #Top bits
     (set! body
	   (string-regexp-substitute/global
	    body "#Top\">([^(]+) \\(\\1\\)</a>"
	    "#Top\">" 1 "</a>"))

     ;; autoconf
     (set! body
	   (string-substitute/global
     	    body "href=\"autoconf.html#Top"
     	    "href=\"http://www.gnu.org/software/autoconf/manual/index.html"))

     ;; automake
     (set! body
	   (string-substitute/global
     	    body "href=\"automake.html#Top"
     	    "href=\"http://sourceware.org/automake/automake.html"))

     ;; bash
     (set! body
	   (string-substitute/global
	    body "href=\"bashref.html"
	    "href=\"http://www.gnu.org/software/bash/manual/index.html"))

     ;; emacs
     (set! body
	   (string-substitute/global
	    body "href=\"emacs.html#Top"
	    "href=\"http://www.gnu.org/software/emacs/manual/index.html"))
     (set! body
     	   (string-regexp-substitute/global
     	    body "href=\"emacs\\.html#([^\"]+)"
	    "href=\"http://www.gnu.org/software/emacs/manual/html_node/"
	    1 ".html"))

     ;; emacs ses - no online copy of ses.texi yet
     (set! body
	   (string-regexp-substitute/global
     	    body "href=\"ses\\.html#([^\"]+)"
	    "href=\"http://home.comcast.net/~jyavner/ses/"))

     ;; elisp
     ;; FIXME: The html_node form has gone away.  The others are 3Mb or
     ;; whatever, far too big to want to link to ...
     (set! body
     	   (string-substitute/global
     	    body "href=\"elisp.html#Buffer-Contents"
	    "href=\"http://www.gnu.org/software/emacs/elisp-manual/elisp.html"))

     ;; gettext
     (set! body
	   (string-substitute/global
     	    body "href=\"gettext.html#Top"
     	    "href=\"http://www.gnu.org/software/gettext/manual/gettext.html"))
     (set! body
	   (string-substitute/global
     	    body "href=\"gettext.html#Language-Codes"
     	    "href=\"http://www.gnu.org/software/gettext/manual/html_node/gettext_221.html"))
     (set! body
	   (string-substitute/global
     	    body "href=\"gettext.html#Country-Codes"
     	    "href=\"http://www.gnu.org/software/gettext/manual/html_node/gettext_222.html"))

     ;; guile ref
     (set! body
     	   (string-substitute/global
	    body "href=\"guile.html#Top"
     	    "href=\"http://www.gnu.org/software/guile/docs/docs-1.8/guile-ref/index.html"))
     (set! body
     	   (string-regexp-substitute/global
     	    body "href=\"guile\\.html#([^\"]+)"
	    "href=\"http://www.gnu.org/software/guile/docs/docs-1.8/guile-ref/"
	    1 ".html"))

     ;; guile tut
     (set! body
     	   (string-substitute/global
	    body "href=\"guile-tut.html#Top"
     	    "href=\"http://www.gnu.org/software/guile/docs/guile-tut/index.html"))

     ;; guile gtk
     (set! body
     	   (string-substitute/global
	    body "href=\"guile-gtk.html#Top"
     	    "href=\"http://www.gnu.org/software/guile-gtk/docs/guile-gtk/index.html"))
     (set! body
     	   (string-regexp-substitute/global
     	    body "href=\"guile-gtk\\.html#([^\"]+)"
	    "href=\"http://www.gnu.org/software/guile-gtk/docs/guile-gtk/"
	    1 ".html"))

     ;; gzip
     (set! body
	   (string-substitute/global
     	    body "href=\"gzip.html#Top"
     	    "href=\"http://www.gnu.org/software/gzip/manual/gzip.html"))

     ;; glibc
     (set! body
     	   (string-regexp-substitute/global
	    body "href=\"libc\\.html#([^\"]+)"
	    "href=\"http://www.gnu.org/software/libc/manual/html_node/"
	    1 ".html"))

     ;; info-stnd
     (set! body
     	   (string-substitute/global
	    body "href=\"info-stnd.html#Top"
     	    "href=\"http://www.gnu.org/software/texinfo/manual/info-stnd/index.html"))

     ;; slib
     (set! body
     	   (string-substitute/global
	    body "href=\"slib.html#Top"
     	    "href=\"http://swissnet.ai.mit.edu/~jaffer/slib_toc.html"))
     (set! body
     	   (string-substitute/global
	    body "href=\"slib.html#Require"
     	    "href=\"http://swissnet.ai.mit.edu/~jaffer/slib_1.html#SEC3"))


     ;; texinfo
     (set! body
     	   (string-substitute/global
	    body "href=\"texinfo.html#Top"
     	    "href=\"http://www.gnu.org/software/texinfo/manual/texinfo/index.html"))
     (set! body
     	   (string-regexp-substitute/global
     	    body "href=\"texinfo\\.html#([^\"]+)"
	    "href=\"http://www.gnu.org/software/texinfo/manual/texinfo/html_node/"
	    1 ".html"))


     (let ((m-list (list-matches "href=\"[a-z]+\\.html(#[^\"]+)?\"" body)))
       (or (null? m-list)
	   (begin
	     (display "Unmunged links:\n")
	     (for-each (lambda (m)
			 (format #t "    ~a\n" (match:substring m)))
		       m-list)
	     (exit 1))))

     (call-with-output-file filename
       (lambda (port)
	 (display body port)))))

 (cdr (program-arguments)))

(exit 0)
