;; Checking (ice-9 optargs).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (guile-lint ice-9 optargs)
  #:use-module (guile-lint common)
  #:use-module (ice-9 optargs)
  #:use-module (srfi srfi-1))


;;-----------------------------------------------------------------------------
;; define*

(lint-arity! define* '(2 0 #t))

(define (form-define*-arglist where expr)
  (and (or (not (null? expr))
	   (lint-error where "`define*' formals cannot be empty: ~s" expr))
       (or (pair?   (first expr))
	   (symbol? (first expr))
	   (lint-error where "`define*' arglist must start with symbol or currying list: ~s" expr))
       (and? (if (pair? (first expr))
		 (form-define*-arglist where (car expr))
		 #t)
	     (form-lambda*-formals where 'define* expr (cdr expr)))))

(define (form-define* where expr)
  (and (or (symbol? (second expr))
	   (pair?   (second expr))
	   (lint-error where "`define*' takes symbol or list: ~s"
		       (second expr)))
       (or (not (symbol? (second expr)))
	   (= 3 (length expr))
	   (lint-error where "`define*' of symbol takes two arguments"))
       (or (not (pair? (second expr)))
	   (form-define*-arglist where (second expr)))))

(define (lint-define* where now expr env)
  (or (env-toplevel env)
      (lint-error where "`define*' bad placement, can only be at start of body"))
  (lint-define*-allowed where now expr env))

(define* (lint-define*-allowed where now expr env #:optional (used #f))
  (if (form-define* where expr)
      (if (symbol? (second expr))
	  ;; (define* foo ...)
	  (begin
	    (lint-error where "`define*' of symbol can use plain `define'")
	    (lint-body-list where now (cddr expr) env)
	    (env-bind-define env where (second expr)
				 (lint-value where (third expr) env) used))

	  ;; (define* (foo ...) ...)
	  (let* ((formals       (second expr))
		 (formals-where (where-follow where formals)))
	    (let ((flat (currying-formals-flatten formals)))
	      ;; duplicates between currying levels don't make sense,
	      ;; though they're not an error
	      (lint-formals-duplicates formals-where 'define* flat)

	      (if (not (any keyword? (improper->proper flat)))
		  (lint-error where "`define*' can be plain `define'"))

	      ;; new definition
	      (let ((name (car-to-object formals))
		    (val (lint-value where
				     (list lambda* (cdr (car-to-last-list
							 formals)) #f) env)))
		(set-procedure-property! val 'name name)
		(env-bind-define env where name val used))

	      (let loop ((env env)
			 (formals flat))
		(lint-subenv
		 env
		 (lambda (subenv)
		   (if (null? formals)
		       (lint-body where #f (lint-strip-docstring (cddr expr))
				  subenv)
		       (begin
			 (if (symbol? formals) ;; . rest
			     (set! formals (list formals)))
			 (cond ((symbol? (first formals))
				(env-bind-one subenv formals-where
					      (first formals) #f #t))
			       ((pair? (first formals))
				(let ((where (where-follow formals-where
							   (first formals)))
				      (sym   (first (first formals)))
				      (val   (second (first formals))))
				  (lint-object where #f val subenv)
				  ;; don't want to put in "(lint-value where
				  ;; val subenv)" as the value, it's only a
				  ;; default, not its definite setting
				  (env-bind-one subenv where sym #f #t))))
			 (loop subenv (cdr formals))))))))))))

(lint-handler! define* lint-define*)

(set! lint-body-define-alist (acons define* lint-define*-allowed
				    lint-body-define-alist))


;;-----------------------------------------------------------------------------
;; define*-public

(lint-arity! define*-public '(2 0 #t))

(define (lint-define*-public where now expr env)
  (lint-require-toplevel where now expr env)
  (lint-define*-allowed where now expr env #t))

(lint-handler! define*-public lint-define*-public)


;;-----------------------------------------------------------------------------
;; lambda*

(lint-arity! lambda* '(2 0 #t))

(define (form-lambda*-formals where name source-expr formals)

  (and (or (symbol? formals)
	   (pair? formals)
	   (null? formals)
	   (lint-error where "`~a' formals must be list or symbol: ~s"
		       name formals))
       
       (let ((formals formals))
	 (define (ext-var-decl? elem)
	   (or (symbol? elem)
	       (and (list? elem)
		    (= 2 (length elem))
		    (symbol? (car elem)))))
	 
	 ;; leading symbols
	 (let symbols ()
	   (if (pair? formals)
	       (if (symbol? (car formals))
		   (begin
		     (set! formals (cdr formals))
		     (symbols)))))
	 
	 ;; #:optional symbols or (symbol expr)
	 (if (and (pair? formals)
		  (eq? #:optional (car formals)))
	     (begin
	       (set! formals (cdr formals))
	       (let optionals ()
		 (if (pair? formals)
		     (if (ext-var-decl? (car formals))
			 (begin
			   (set! formals (cdr formals))
			   (optionals)))))))

	 ;; #:key symbols or (symbol expr), and optional #:allow-other-keys
	 (if (and (pair? formals)
		  (eq? #:key (car formals)))
	     (begin
	       (set! formals (cdr formals))
	       (let keys ()
		 (if (pair? formals)
		     (cond ((ext-var-decl? (first formals))
			    (set! formals (cdr formals))
			    (keys))
			   ((eq? #:allow-other-keys (first formals))
			    (set! formals (cdr formals))))))))
	 
	 ;; #:rest symbol or . symbol
	 (cond ((and (pair? formals)
		     (eq? #:rest (car formals))
		     (pair? (cdr formals))
		     (symbol? (cadr formals)))
		(set! formals (cddr formals)))
	       ((symbol? formals)
		(set! formals '())))
	 
	 (or (null? formals)
	     (lint-error where "`~a' invalid formals: ~s" name formals)))))

(define (form-lambda* where expr)
  (and (or (length>=? expr 3)
	   (lint-error where "`lambda*' must have bindings and at least one expression: ~s" expr))
       (form-lambda*-formals where 'lambda* (second expr) (second expr))))

(define (lint-lambda* where now expr env)
  (if (form-lambda* where expr)
      (let* ((formals       (second expr))
	     (formals-where (where-follow where formals)))
	(lint-formals-duplicates formals-where 'lambda* formals)
	(if (not (any keyword? (improper->proper formals)))
	    (lint-error where "`lambda*' can be plain `lambda'"))
	(let loop ((env env)
		   (formals formals))
	  (lint-subenv
	   env
	   (lambda (subenv)
	     (if (null? formals)
		 (lint-body where #f (cddr expr) subenv)
		 (begin
		   (if (symbol? formals) ;; . rest
		       (set! formals (list formals)))
		   (cond ((symbol? (first formals))
			  (env-bind-one subenv formals-where
					(first formals) #f #t))
			 ((pair? (first formals))
			  (let ((where (where-follow formals-where
						     (first formals)))
				(sym   (first (first formals)))
				(val   (second (first formals))))
			    ;; default evaled when procedure called
			    (lint-object where #f val subenv)
			    (env-bind-one subenv where sym
					  (lint-value where val subenv)
					  #t))))
		   (loop subenv (cdr formals))))))))))

(lint-handler! lambda* lint-lambda*)


;;-----------------------------------------------------------------------------
;; let-keywords

;; force non-empty body
(lint-arity! let-keywords '(4 0 #t))

(define (lint-let-keywords where now expr env)
  (lint-object where now (third expr) env) ;; allow-other-keywords
  ;; mung to let-optional args
  (set! expr (cons* (car expr) (cadr expr) (cdddr expr)))
  (if (form-let-optional where 'let-keywords expr)
      (lint-let-optional-action where now expr env)))

(lint-handler! let-keywords lint-let-keywords)


;;-----------------------------------------------------------------------------
;; let-keywords*

;; force non-empty body
(lint-arity! let-keywords* '(4 0 #t))

(define (lint-let-keywords* where now expr env)
  (lint-object where now (third expr) env) ;; allow-other-keywords
  ;; mung to let-optional* args
  (set! expr (cons* (car expr) (cadr expr) (cdddr expr)))
  (if (form-let-optional where 'let-keywords* expr)
      (lint-let-optional*-action where now expr env)))

(lint-handler! let-keywords* lint-let-keywords*)


;;-----------------------------------------------------------------------------
;; let-optional

;; force non-empty body
(lint-arity! let-optional '(3 0 #t))

;; used for let-keywords, let-keywords*, let-optional* too
(define (form-let-optional where name expr)
  (let* ((binding-list (third expr))
	 (where        (where-follow where binding-list)))
    (and (or (list? binding-list)
	     (lint-error where "`~a' bindings must be a (proper) list: ~s"
			 name binding-list))
	 (apply and?
		(iterate-where
		 map-in-order where binding-list
		 (lambda (where binding)
		   (or (or (symbol? binding)
			   (and (list? binding)
				(= 2 (length binding))
				(symbol? (first binding))))
		       (lint-error where "`~a' binding must be symbols and value: ~s" name binding))))))))

(define (lint-let-optional where now expr env)
  (if (form-let-optional where 'let-optional expr)
      (lint-let-optional-action where now expr env)))

;; used for let-keywords too
(define (lint-let-optional-action where now expr env)
  (let ((binding-list (third expr)))
    ;; rest arg
    (lint-object where now (second expr) env)
    (lint-bindings-duplicates where 'let-optional binding-list)
    ;; default values
    (for-each-bindings where (filter list? binding-list)
		       (lambda (where sym val)
			 (lint-object where now val env)))
    (lint-subenv env
		 (lambda (subenv)
		   ;; bindings
		   (for-each (lambda (binding)
			       (let ((sym   binding)
				     (val   #f)
				     (where (where-follow where binding)))
				 (if (list? binding)
				     (begin
				       (set! sym (first binding))
				       (set! val (second binding))))
				 (env-bind-one subenv where sym val #t)))
			     binding-list)
		   ;; body
		   (lint-body where now (cdddr expr) subenv)))))

(lint-handler! let-optional lint-let-optional)


;;-----------------------------------------------------------------------------
;; let-optional*

;; force non-empty body
(lint-arity! let-optional* '(3 0 #t))

(define (lint-let-optional* where now expr env)
  (if (form-let-optional where 'let-optional* expr)
      (lint-let-optional*-action where now expr env)))

;; used for let-keywords* too
(define (lint-let-optional*-action where now expr env)
  (lint-object where now (second expr) env)
  (let loop ((binding-list (third expr))
	     (env          env))
    (lint-subenv env
		 (lambda (subenv)
		   (if (null? binding-list)
		       (lint-body where now (cdddr expr) subenv)
		       (let* ((binding (first binding-list))
			      (where   (where-follow where binding)))
			 (if (pair? binding)  ;; (SYM DEFAULT)
			     (begin
			       (lint-object where now (second binding)
					    env)
			       (set! binding (first binding))))
			 (env-bind-one subenv where binding #f #t)
			 (loop (cdr binding-list) subenv)))))))

(lint-handler! let-optional* lint-let-optional*)



;;-----------------------------------------------------------------------------
;; optargs call checking

(define-public (lint-call-optarity where proc-where proc-name optarity arglist)
  (receive-list (non-optional optional keys+defaults aok? rest)
      optarity
    ;; turn (name <default>) keys into just symbols
    (define keys (map (lambda (key)
			(if (pair? key) (first key) key))
		      keys+defaults))

    (let ((proc-error #f)    ;; to print proc location at end
	  (seen-keys  '()))  ;; list of keywords

      (define (check-non-optional)
	(cond ((null? non-optional)
	       #t)  ;; no more non-optionals, keep checking rest
	      ((symbol? arglist)
	       #f)  ;; `apply' args can be anything, stop
	      ((null? arglist)
	       (lint-error where "`~a' not enough arguments" proc-name)
	       (set! proc-error #t)
	       #f)  ;; stop
	      (else ;; (pair? arglist)
	       (set! non-optional (cdr non-optional))
	       (set! arglist (cdr arglist))
	       (check-non-optional))))

      (define (check-optional)
	(cond ((null? optional)
	       #t)  ;; no more optionals, keep checking rest
	      ((symbol? arglist)
	       #f)  ;; `apply' args can be anything, stop
	      ((null? arglist)
	       #f)  ;; stop
	      (else ;; (pair? arglist)
	       (set! optional (cdr optional))
	       (set! arglist (cdr arglist))
	       (check-optional))))

      (define (check-keys)
	(cond ((symbol? arglist)
	       #f)  ;; `apply' args can be anything, stop
	      ((null? arglist)
	       #t)
	      ((and (pair? arglist)
		    (keyword? (first arglist)))
	       ;; examine the keyword
	       (let ((key (first arglist)))
		 (or aok?
		     (memq (keyword->symbol key) keys)
		     (begin
		       (lint-error where "`~a' invalid keyword argument: ~a"
				   proc-name key)
		       (set! proc-error #t)))
		 (if (memq key seen-keys)
		     (lint-error where "`~a' duplicate keyword argument: ~a"
				 proc-name key)
		     (set! seen-keys (cons key seen-keys)))
		 (set! arglist (cdr arglist))

		 ;; examine its arg
		 (cond ((symbol? arglist)
			#f)  ;; `apply' args can be anything
		       ((null? arglist)
			(lint-error where "`~a' missing value for ~a"
				    proc-name key)
			#f)  ;; stop
		       (else ;; (pair? arglist)
			(set! arglist (cdr arglist))
			(check-keys)))))

	      (else ;; (pair? arglist)
	       #t))) ;; no more keywords, check rest

      (define (check-mandatory-keys)
	(for-each (lambda (elem)
		    (if (pair? elem)
			(let ((key     (symbol->keyword (first elem)))
			      (default (second elem)))
			  (if (and (pair? default)
				   (or (eq? error (first default))
				       (eq? 'error (first default))))
			      (or (memq key seen-keys)
				  (begin
				    (lint-error where "`~a' keyword ~a is mandatory: ~s" proc-name key default)
				    (set! proc-error #t)))))))
		  keys+defaults))

      (define (check-rest)
	(cond ((pair? arglist)
	       (or rest
		   (begin
		     (lint-error where "`~a' additional arguments without `rest': ~s" proc-name arglist)
		     (set! proc-error #t))))))

      (and (check-non-optional)
	   (check-optional)
	   (check-keys)
	   (check-mandatory-keys)
	   (check-rest))

      (if (and proc-error proc-where)
	  (lint-error proc-where "  `~a' definition is here" proc-name)))))

(define optargs-module (resolve-module '(ice-9 optargs)))

;; return a procedure taking the given optargs style arguments
;;
;; This eval seems to bomb with formals ending in ". r" style, so change to
;; #:rest style.  ice-9 syncase might be the culprit.
;;
(define-public (lambda*-with-args formal-list)
  (if (dotted-list? formal-list)
      (begin
	(set! formal-list (improper->proper formal-list))
	(set! formal-list (append (drop-right formal-list 1)
				  (list #:rest (last formal-list))))))
  (eval `(lambda* ,formal-list #f) optargs-module))
