;; Checking (ice-9 regex).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (guile-lint ice-9 regex)
  #:use-module (guile-lint common)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 regex)
  #:use-module (srfi srfi-1))


;;-----------------------------------------------------------------------------
;; helpers

;; check regexp PATTERN+FLAGS, return error message string, or #f if good
(define (regexp-pattern-check pattern . flags)
  (catch 'regular-expression-syntax
    (lambda ()
      (apply make-regexp pattern flags)
      #f)
    (lambda (key func msg . rest)
      msg)))


;;-----------------------------------------------------------------------------
;; types

(define type-match
  (make-type "regexp match"
	     #:test regexp-match?))

(define type-regexp
  (make-type "compiled regexp"
	     #:test regexp?))

(define type-port-or-f
  (type-or type-port type-f))

(define type-string-or-regexp
  (type-or type-string type-regexp))

(define type-symbol-pre-post
  (make-type "'pre or 'post"
	     #:test (lambda (obj)
		      (and (symbol? obj)
			   (memq obj '(pre post))))
	     #:supertype type-symbol))

(lint-types! match:count     type-exact-integer-nonneg type-match)
(lint-types! match:prefix    type-string type-match)
(lint-types! match:string    type-string type-match)
(lint-types! match:suffix    type-string type-match)

(lint-types! regexp?        type-boolean type-any)
(lint-types! regexp-exec  type-match
	     type-regexp type-string type-exact-integer-nonneg)
(lint-types! regexp-match?  type-boolean type-any)
(lint-types! regexp-quote   type-string type-string)
(lint-types! regexp-substitute  (type-or type-unspecified type-string)
	     type-port-or-f type-match
	     (type-or type-string type-exact-integer-nonneg
		      type-symbol-pre-post))


;;-----------------------------------------------------------------------------
;; fold-matches

;; only (4 0 #t) in regex.scm
(lint-arity! fold-matches '(4 1 #f))

(lint-types! fold-matches  type-any
	     type-string-or-regexp type-string type-any type-procedure
	     type-exact-integer-nonneg)

(define (lint-fold-matches where now expr env)
  (lint-procexpr-arglist where now (fifth expr) env '(match prev))
  (lint-list where now expr env))

(lint-handler! fold-matches lint-fold-matches)


;;-----------------------------------------------------------------------------
;; list-matches

;; only (2 0 #t) in regex.scm
(lint-arity! list-matches '(2 1 #f))

(lint-types! list-matches  type-list
	     type-string-or-regexp type-string type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; make-regexp

(lint-types! make-regexp  type-regexp type-string type-exact-integer-nonneg)

(define lint-make-regexp-flags-alist
  `((regexp/icase    . ,regexp/icase)
    (regexp/newline  . ,regexp/newline)
    (regexp/basic    . ,regexp/basic)
    (regexp/extended . ,regexp/extended)))

(define (lint-make-regexp where now expr env)
  ;; if the pattern and flags are constant then validate now
  ;; FIXME: check symbols regexp/icase etc are really the core defines
  ;; within env, not shadowed
  (and-let*
      ((pattern (second expr))
       (        (string? pattern))
       (flags   (map (lambda (flag)
		       (assq-ref lint-make-regexp-flags-alist flag))
		     (cddr expr)))
       (errmsg  (if (apply and? flags)
		    ;; flags are constants
		    (apply regexp-pattern-check pattern flags)
		    ;; flags not constant, give an error if pattern bad as
		    ;; both basic and extended
		    (and (regexp-pattern-check pattern regexp/basic)
			 (regexp-pattern-check pattern regexp/extended)))))
    (lint-error where "`make-regexp' pattern ~s: ~a" pattern errmsg))
  (lint-list where now (cdr expr) env))

(lint-handler! make-regexp lint-make-regexp)


;;-----------------------------------------------------------------------------
;; match:end

;; only (1 0 #t) in regex.scm
(lint-arity! match:end '(1 1 #f))

(lint-types! match:end  type-exact-integer-nonneg
	     type-match type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; match:start

;; only (1 0 #t) in regex.scm
(lint-arity! match:start '(1 1 #f))

(lint-types! match:start  type-exact-integer-nonneg
	     type-match type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; match:substring

;; only (1 0 #t) in regex.scm
(lint-arity! match:substring '(1 1 #f))

(lint-types! match:substring  type-string type-match type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; regexp-substitute/global

;; only (3 0 #t) in regex.scm, but demand at least one action arg
(lint-arity! regexp-substitute/global '(4 0 #t))

(lint-types! regexp-substitute/global  (type-or type-unspecified type-string)
	     type-port-or-f
	     type-string-or-regexp
	     type-string
	     (type-or type-string type-exact-integer-nonneg
		      type-symbol-pre-post type-predicate))

(define (lint-regexp-substitute/global where now expr env)
  ;; if the pattern is a constant string then validate it now
  (and-let* ((pattern (third expr))
	     (        (string? pattern))
	     (errmsg  (regexp-pattern-check pattern))) ;; no flags
    (lint-error where "`regexp-substitute/global' pattern ~s: ~a"
		pattern errmsg))
  (for-each (lambda (elem)
	      ;; elements which are not procedures will be ignored
	      (lint-procexpr-arglist where now elem env '(match)))
	    (cdddr expr))
  (lint-list where now (cdr expr) env))

(lint-handler! regexp-substitute/global lint-regexp-substitute/global)


;;-----------------------------------------------------------------------------
;; string-match

;; only (2 0 #t) in regex.scm
(lint-arity! string-match '(2 1 #f))

(lint-types! string-match  type-match
	     type-string type-string type-exact-integer-nonneg)

(define (lint-string-match where now expr env)
  ;; if the pattern is a constant string then validate it now
  (and-let* ((pattern (second expr))
	     (        (string? pattern))
	     (errmsg  (regexp-pattern-check pattern))) ;; no flags
    (lint-error where "`string-match' pattern ~s: ~a" pattern errmsg))
  (lint-list where now (cdr expr) env))

(lint-handler! string-match lint-string-match)
