;; Checking slib, setups shared by --slib and (ice-9 slib).

;; Copyright 2003, 2004, 2005, 2006, 2007, 2008 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;;-----------------------------------------------------------------------------
;; helpers

;; return the toplevel environment at or above ENV
(define (env->toplevel env)
  (if (env-toplevel env)
      env
      (env->toplevel (env-prev env))))


;;-----------------------------------------------------------------------------
;; make-random-state

;; only (0 0 #t) in guile.init
(lint-arity! make-random-state '(0 1 #f))


;;-----------------------------------------------------------------------------
;; output-port-height, output-port-width

;; only (0 0 #t) in guile.init
(lint-arity! output-port-height '(0 1 #f))

;; only (0 0 #t) in guile.init
(lint-arity! output-port-width '(0 1 #f))


;;-----------------------------------------------------------------------------
;; require

;; expr is "(define (NAME ARGLIST) BODY...)"		
;;      or "(define NAME VALUE)"		
(define (lint-slib-define where expr env)
  (if (symbol? (second expr))
      (let ((val (third expr)))
	(env-bind-one env where (second expr) (lint-value where val env)))
      (let* ((formals (second expr))
	     (name    (first formals))
	     (val     (lint-value where (list lambda (cdr formals) #f) env)))
	(set-procedure-property! val 'name name)
	(env-bind-one env where name val #t))))

(define (lint-slib-object where obj env)
  (cond ((and (list>=? obj 3)
	      (memq (first obj) '(define
				   define-operation))) ;; yasos
	 (lint-slib-define where obj env))
	((and (list>=? obj 3)
	      (eq? (first obj) 'defmacro))
	 (lint-slib-define where obj env))
	((and (list>=? obj 2)
	      (eq? (first obj) 'require)
	      (list>=? (second obj) 2)
	      (eq? (first (second obj)) 'quote))
	 (lint-slib-feature where (second (second obj)) env))
	((and (list>=? obj 3)
	      (eq? (first obj) 'require-if)
	      (list>=? (third obj) 2)
	      (eq? (first (third obj)) 'quote))
	 (lint-slib-feature where (second (third obj)) env))))

(define lint-slib-file-env-hash (make-hash-table 67))

(define (lint-slib-file-env filename)
  (or (hash-ref lint-slib-file-env-hash filename)
      (let ((env   (env-new #f))
	    (where #f))
	(catch #t
	  (lambda ()
	    (call-with-input-file filename
	      (lambda (port)
		(do ((obj (read port) (read port)))
		    ((eof-object? obj))
		  (set! where (where-port-object port obj))
		  (lint-slib-object where obj env)))))
	  (lambda (key . args)
	    (lint-error where "error loading slib file: ~s\n~s ~s\n"
			filename key args)))
	(hash-set! lint-slib-file-env-hash filename env)
	env)))

;; We don't have multiple-inheritance for environments, so just copy senv
;; into ENV.  This is a bit wasteful, especially when an slib file itself
;; loads in lots of things.
(define (lint-slib-load filename env)
  (let ((senv (lint-slib-file-env filename)))
    (hash-fold (lambda (name bind prior)
		 (env-bind-one env (binding-where bind) name (binding-val bind)
			       #t))
	       #f
	       (env-binding-hash senv))))

(define (lint-slib-feature where feature env)
  (cond ((string? feature)
	 (lint-slib-load (string-append feature ".scm") env))
	((symbol? feature)
	 (if (not (provided? feature))
	     (begin
	       (set! feature (catalog:get feature))
	       (if (and (list? feature)
			(= 2 (length feature))
			(eq? 'source (first feature)))
		   (set! feature (second feature)))
	       (lint-slib-feature where feature env))))
	((list>=? feature 2)
	 (case (first feature)
	   ((aggregate)
	    (for-each (lambda (feature)
			(lint-slib-feature where feature env))
		      (cddr feature)))
	   ((color-names)
	    ;; FIXME: need to do anything with this?
	    #f)
	   ((defmacro macro)
	    (lint-slib-feature where (second feature) env))
	   ((spectral-tristimulus-values)
	    ;; load-ciexyz of some file, no need to do anything, it just
	    ;; adds to variable values
	    #f)
	   (else
	    (lint-error where "`require' unrecognised feature: ~s" feature))))
	(else
	 (lint-error where "`require' unrecognised feature: ~s" feature))))

(define (form-slib-require where expr env)
  (let ((feature (second expr)))
    (or (and (list? feature)
	     (= 2 (length feature))
	     ;; FIXME: check `quote' is really the quote macro in env
	     (eq? 'quote (first feature))
	     (symbol? (second feature)))
	(lint-error (where-follow where feature)
		    "lint `require' only supports a quoted feature symbol: ~s"
		    feature))))

(define (lint-slib-require where now expr env)
  (if (form-slib-require where expr env)
      (lint-slib-feature where (second (second expr)) (env->toplevel env))))

(lint-handler! require lint-slib-require)


;;-----------------------------------------------------------------------------
;; require-if

(define (lint-slib-require-if where now expr env)
  (lint-slib-require where now (cdr expr) env))

(lint-handler! require-if lint-slib-require-if)
