;; Checking (srfi srfi-1).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (guile-lint srfi srfi-1)
  #:use-module (ice-9 and-let-star)
  #:use-module (guile-lint common)
  #:use-module (srfi srfi-1))


(define (module-unexport-maybe! m names)
  (let ((public-i (module-public-interface m)))
    (for-each (lambda (name)
		(let ((var (module-local-variable public-i name)))
		  (if var
		      (module-remove! public-i name)
		      ;; (error "module-unexport!: no such variable:" name)
		      )))
	      names)))

;; In guile 1.8, the full set of srfi-1 specified procedures is exported,
;; even those which are in r5rs.  Change the module so r5rs bits are not
;; exported, so that merely using those funcs doesn't make it look like the
;; module is used.
;;
;; Likewise remove a few guile core bindings which are re-exported, but only
;; those which have been in guile forever.  In particular don't drop
;; `filter' which is only in the core in guile 1.8 and higher.
;;
(module-unexport-maybe!
 (resolve-module '(srfi srfi-1))
 '(cons list make-list pair? null?
	car cdr caar cadr cdar cddr
	caaar caadr cadar caddr cdaar cdadr cddar cdddr
	caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr
	cdaaar cdaadr cdadar cdaddr cddaar cddadr cdddar cddddr
	list-ref length append append! reverse reverse!
	memq memv assq assv set-car! set-cdr!))



;;-----------------------------------------------------------------------------
;; srfi-1 types

(if (not (eq? alist-cons acons))  ;; different defn in 1.6
    (lint-types! alist-cons  'pure type-alist type-any type-any type-alist))
(lint-types! append-reverse    'pure type-list type-list)
(lint-types! append-reverse!   type-list type-list)

(lint-types! circular-list  'pure type-circular-list type-any)
(lint-types! circular-list? type-boolean type-any)
(lint-types! concatenate    'pure type-list type-list)
(lint-types! concatenate!   type-list type-list)
(lint-types! count          type-exact-integer-nonneg type-procedure type-list)

(lint-types! dotted-list? 'pure type-boolean type-any)

(if (not (eq? drop list-tail)) ;; alias in 1.8
    (lint-types! drop     'pure type-list type-list type-exact-integer-nonneg))
(lint-types! drop-right   'pure type-list type-list type-exact-integer-nonneg)
(lint-types! drop-right!       	type-list type-list type-exact-integer-nonneg)

;; aliases for car, cadr, caddr, cadddr in the core
;; (lint-types! first        	'pure type-any type-pair)
;; (lint-types! second        	'pure type-any type-pair)
;; (lint-types! third        	'pure type-any type-pair)
;; (lint-types! fourth        	'pure type-any type-pair)
(lint-types! fifth        	'pure type-any type-pair)
(lint-types! sixth        	'pure type-any type-pair)
(lint-types! seventh        	'pure type-any type-pair)
(lint-types! eighth        	'pure type-any type-pair)
(lint-types! ninth        	'pure type-any type-pair)
(lint-types! tenth        	'pure type-any type-pair)

(lint-types! last        	'pure type-any type-list)
(lint-types! length+       	'pure type-exact-integer-nonneg type-list)
(lint-types! list=        	'pure type-boolean type-procedure type-list)
(lint-types! list-tabulate type-list
	     type-exact-integer-nonneg type-procedure)

(lint-types! not-pair?  'pure type-boolean type-any)
(lint-types! null-list? 'pure type-boolean type-list)

(lint-types! proper-list? 'pure type-boolean type-any)

(if (not (eq? take list-head)) ;; alias in 1.8
    (lint-types! take     'pure type-list type-list type-exact-integer-nonneg))
(lint-types! take!        	type-list type-list type-exact-integer-nonneg)
(lint-types! take-right   'pure type-list type-list type-exact-integer-nonneg)

(lint-types! xcons        'pure type-pair type-any)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) alist-delete
;; (srfi srfi-1) alist-delete!

;; only (2 0 #t) in guile 1.6 srfi-1.scm
(lint-arity! alist-delete  '(2 1 #f))
(lint-arity! alist-delete! '(2 1 #f))

(lint-types! alist-delete 'pure type-alist type-any type-alist type-equal-proc)
(lint-types! alist-delete!      type-alist type-any type-alist type-equal-proc)

(define (lint-srfi1-member where now expr env)
  (if (list>=? expr 4)
      (lint-procexpr-arglist where now (fourth expr) env '(x elem)))
  (lint-list where now expr env))

(lint-handler! alist-delete  lint-srfi1-member)
(lint-handler! alist-delete! lint-srfi1-member)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) any

(lint-types! any  type-any type-predicate type-list)
(lint-handler! any lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) append-map
;; (srfi srfi-1) append-map!

(lint-types! append-map   type-list type-procedure type-list)
(lint-types! append-map!  type-list type-procedure type-list)
(lint-handler! append-map lint-for-each)
(lint-handler! append-map! lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) assoc

;; only (2 0 #t) in guile 1.6 srfi-1.scm
(lint-arity! assoc '(2 1 #f))
(lint-types! assoc 'pure type-alist type-any type-alist type-equal-proc)
(lint-handler! assoc lint-srfi1-member)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) break
;; (srfi srfi-1) break!

(lint-handler! break lint-for-each)
(lint-handler! break! lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) delete
;; (srfi srfi-1) delete!

;; only (2 0 #t) in guile 1.6 srfi-1.scm
(lint-arity! delete  '(2 1 #f))
(lint-arity! delete! '(2 1 #f))

(lint-types! delete  'pure type-list type-any type-list type-equal-proc)
(lint-types! delete!       type-list type-any type-list type-equal-proc)

(lint-handler! delete  lint-srfi1-member)
(lint-handler! delete! lint-srfi1-member)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) delete-duplicates
;; (srfi srfi-1) delete-duplicates!

;; only (1 0 #t) in guile 1.6 srfi-1.scm
(lint-arity! delete-duplicates  '(1 1 #f))
(lint-arity! delete-duplicates! '(1 1 #f))

(lint-types! delete-duplicates  'pure type-list type-list type-equal-proc)
(lint-types! delete-duplicates!       type-list type-list type-equal-proc)

(define (lint-srfi1-delete-duplicates where now expr env)
  (if (list>=? expr 3)
      (lint-procexpr-arglist where now (third expr) env '(elem1 elem2)))
  (lint-list where now expr env))

(lint-handler! delete-duplicates  lint-srfi1-delete-duplicates)
(lint-handler! delete-duplicates! lint-srfi1-delete-duplicates)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) every

(lint-types! every  type-any type-predicate type-list)
(lint-handler! every lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) filter-map

(lint-types! filter-map  type-list type-procedure type-list)
(lint-handler! filter-map lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) find
;; (srfi srfi-1) find-tail

(lint-types! find        type-any type-predicate type-list)
(lint-types! find-tail   type-any type-predicate type-list)
(lint-handler! find      lint-for-each)
(lint-handler! find-tail lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) fold
;; (srfi srfi-1) fold-right

(lint-types! fold       type-any type-procedure type-any type-list)
(lint-types! fold-right type-any type-procedure type-any type-list)
(lint-handler! fold       lint-fold)
(lint-handler! fold-right lint-fold)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) for-each

(lint-handler! for-each lint-for-each)
(lint-types! for-each  type-unspecified type-procedure type-list)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) iota

(lint-arity! iota '(1 2 #f))
(lint-types! iota 'pure type-list type-exact-integer-nonneg type-number)
(lint-defaults! iota #f 0 1)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) list-index

(lint-handler! list-index lint-for-each)
(lint-types! list-index type-exact-integer-nonneg type-procedure type-list)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) lset=
;; (srfi srfi-1) lset<=
;; (srfi srfi-1) lset-adjoin
;; (srfi srfi-1) lset-union
;; (srfi srfi-1) lset-union!
;; (srfi srfi-1) lset-intersection
;; (srfi srfi-1) lset-intersection!
;; (srfi srfi-1) lset-difference
;; (srfi srfi-1) lset-difference!
;; (srfi srfi-1) lset-xor
;; (srfi srfi-1) lset-xor!
;; (srfi srfi-1) lset-diff+intersection
;; (srfi srfi-1) lset-diff+intersection!

(lint-types! lset<=       	'pure type-boolean type-equal-proc type-list)
(lint-types! lset=        	'pure type-boolean type-equal-proc type-list)
(lint-types! lset-adjoin  	'pure type-list type-equal-proc type-list type-any)
(lint-types! lset-union   	'pure type-list type-equal-proc type-list)
(lint-types! lset-union!              type-list type-equal-proc type-list)
(lint-types! lset-difference    'pure type-list type-equal-proc type-list)
(lint-types! lset-difference!         type-list type-equal-proc type-list)
(lint-types! lset-intersection  'pure type-list type-equal-proc type-list)
(lint-types! lset-intersection!       type-list type-equal-proc type-list)
(lint-types! lset-xor           'pure type-list type-equal-proc type-list)
(lint-types! lset-xor!                type-list type-equal-proc type-list)

(define (lint-lset= where now expr env)
  (if (<= (length expr) 3)
      (lint-warn where "warning, `~a' usually should have 3 or more args (2 args is often a missing equality proc)"
		  (first expr)))
  (lint-procexpr-arglist where now (second expr) env '(x y))
  (lint-list where now expr env))

(lint-handler! lset=                   lint-lset=)
(lint-handler! lset<=                  lint-lset=)
(lint-handler! lset-adjoin             lint-lset=)
(lint-handler! lset-union              lint-lset=)
(lint-handler! lset-union!             lint-lset=)
(lint-handler! lset-intersection       lint-lset=)
(lint-handler! lset-intersection!      lint-lset=)
(lint-handler! lset-difference         lint-lset=)
(lint-handler! lset-difference!        lint-lset=)
(lint-handler! lset-xor                lint-lset=)
(lint-handler! lset-xor!               lint-lset=)
(lint-handler! lset-diff+intersection  lint-lset=)
(lint-handler! lset-diff+intersection! lint-lset=)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) map
;; (srfi srfi-1) map-in-order

(lint-handler! map lint-for-each)
(if (not (eq? map-in-order map)) ;; same as map in 1.6
    (lint-handler! map-in-order lint-for-each))


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) map!

(if (not (eq? map map!)) ;; just an alias for map in pre-1.8
    (begin
      (lint-types! map!  type-list type-procedure type-list)
      (lint-handler! map! lint-for-each)))


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) member

;; only (2 0 #t) in guile 1.6 srfi-1.scm
(lint-arity! member '(2 1 #f))
(lint-types! member 'pure type-list type-any type-list type-equal-proc)
(lint-handler! member lint-srfi1-member)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) pair-fold
;; (srfi srfi-1) pair-fold-right

(lint-types! pair-fold       type-any type-procedure type-any type-list)
(lint-types! pair-fold-right type-any type-procedure type-any type-list)
(lint-handler! pair-fold       lint-fold)
(lint-handler! pair-fold-right lint-fold)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) pair-for-each

(lint-types! pair-for-each  type-unspecified type-procedure type-list)
(lint-handler! pair-for-each lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) partition
;; (srfi srfi-1) partition!

(lint-types! partition    type-values type-predicate type-list)
(lint-types! partition!   type-values type-predicate type-list)
(lint-handler! partition  lint-for-each)
(lint-handler! partition! lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) reduce
;; (srfi srfi-1) reduce-right

(lint-types! reduce       type-any type-procedure type-any type-list)
(lint-types! reduce-right type-any type-procedure type-any type-list)
(lint-handler! reduce lint-fold)
(lint-handler! reduce-right lint-fold)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) remove
;; (srfi srfi-1) remove!

(lint-types! remove    type-list type-predicate type-list)
(lint-types! remove!   type-list type-predicate type-list)
(lint-handler! remove  lint-for-each)
(lint-handler! remove! lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) span
;; (srfi srfi-1) span!

(lint-types! span    type-values type-predicate type-list)
(lint-types! span!   type-values type-predicate type-list)
(lint-handler! span  lint-for-each)
(lint-handler! span! lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) take-while
;; (srfi srfi-1) take-while!

(lint-types! take-while    type-list type-predicate type-list)
(lint-types! take-while!   type-list type-predicate type-list)
(lint-handler! take-while  lint-for-each)
(lint-handler! take-while! lint-for-each)


;;-----------------------------------------------------------------------------
;; (srfi srfi-1) unfold
;; (srfi srfi-1) unfold-right

;; only (4 0 #t) in srfi-1.scm
(lint-arity! unfold '(4 1 #f))
(lint-arity! unfold-right '(4 1 #f))

;; (unfold PRED THIS NEXT SEED [TAILGEN])
(define (lint-unfold where now expr env)
  (lint-procexpr-arglist where now (second expr) env '(seed))
  (lint-procexpr-arglist where now (third expr) env '(seed))
  (lint-procexpr-arglist where now (fourth expr) env '(seed))
  (if (list>=? expr 6)
      (lint-procexpr-arglist where now (sixth expr) env '(seed)))
  (lint-list where now expr env))

(lint-handler! unfold lint-unfold)
(lint-handler! unfold-right lint-unfold)
