;; Checking (ice-9 syncase).

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

(define-module (guile-lint ice-9 syncase)
  #:use-module (guile-lint common)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 syncase)
  #:use-module (srfi srfi-1))


(let ((original-define-syntax define-syntax))
  (set! common-sc-macro (lambda ()
			  original-define-syntax)))

;; All the syncase macros share an sc-macro handler, create aliases so we
;; can attach different lint-handler! routines to each.
;;
(if (defined? '@)
    (begin
      (set! (@ (ice-9 syncase) define-syntax)
	    (procedure->memoizing-macro (macro-transformer define-syntax)))
      (set! (@ (ice-9 syncase) let-syntax)
	    (procedure->memoizing-macro (macro-transformer let-syntax)))
      (set! (@ (ice-9 syncase) letrec-syntax)
	    (procedure->memoizing-macro (macro-transformer letrec-syntax))))
    (begin
      (set! define-syntax
	    (procedure->memoizing-macro (macro-transformer define-syntax)))
      (set! let-syntax
	    (procedure->memoizing-macro (macro-transformer let-syntax)))
      (set! letrec-syntax
	    (procedure->memoizing-macro (macro-transformer letrec-syntax)))))


;;-----------------------------------------------------------------------------
;; helpers

;; return the value associated with KEY in ALIST, or DEFAULT if none
(define (assq-ref-default alist key default)
  (let ((pair (assq key alist)))
    (if pair
	(cdr pair)
	default)))


;;-----------------------------------------------------------------------------
;; define-syntax

;; return the second last element of LST
(define (second-last lst)
  (first (take-right lst 2)))

(define (transform-tree alist tree)
  (if (and (list>=? tree 2)
	   (eq? '... (last tree)))
      (append! (transform-tree alist (drop-right tree 2))
	       (assq-ref alist (second-last tree)))
      (map (lambda (elem)
	     (if (pair? elem)
		 (transform-tree alist elem)
		 (assq-ref-default alist elem elem)))
	   tree)))

(define (syntax-rules-make-macro rules)
  (let ((macro (procedure->memoizing-macro
		(lambda (expr env)
		  (syntax-rules-expand rules expr)))))
    (lint-arity! macro '(1 0 #t))
    macro))

(define (syntax-rules-expand rules expr)
  (let ((literals (second rules)))
    (any (lambda (rule)
	   (and-let* ((bindings (syntax-rules-match-pattern
				 literals
				 (cdr (first rule))
				 (cdr expr))))
	     (transform-tree bindings (second rule))))
	 (cddr rules))))

;; return an alist of bindings, or #f
(define (syntax-rules-match-pattern literals pattern expr)
  (let more ((bindings '())
	     (pattern pattern)
	     (expr    expr))
    (cond ((null? pattern)
	   (if (null? expr)
	       bindings
	       #f))

	  ((symbol? pattern)
	   (alist-cons pattern expr bindings))

	  ((and (pair? pattern)
		(symbol? (first pattern))
		(memq (first pattern) literals)
		(pair? expr)
		(eq? (first pattern) (first expr)))
	   (more bindings (cdr pattern) (cdr expr)))

	  ((and (pair? pattern)
		(symbol? (first pattern))
		(pair? (cdr pattern))
		(eq? '... (second pattern)))
	   (alist-cons (first pattern) expr bindings))

	  ((and (pair? pattern)
		(symbol? (first pattern)))
	   (if (pair? expr)
	       (more (alist-cons (first pattern) (first expr) bindings)
		     (cdr pattern)
		     (cdr expr))))
	     
	  ((and (pair? pattern)
		(pair? (first pattern))
		(pair? expr)
		(pair? (first expr)))
	   (more (syntax-rules-match-pattern literals
					      (first pattern) (first expr))
		 (cdr pattern)
		 (cdr expr)))

	  (else
	   (error "oops, unrecognised syntax-rules pattern element: "
		  pattern)))))

(define (pseudo-syntax-rules-func literals rule1 . rest)
  #f)

(define (form-syntax-rules-literals where literals)
  (and (or (and (list? literals)
		(every symbol? literals))
	   (lint-error where "`syntax-rules' literals must be a (proper) list of symbols: ~s" literals))
       (or (not (memq '... literals))
	   (lint-error where "`syntax-rules' literals cannot contain `...': ~s" literals))))

(define (form-syntax-rules where sym expr)
  (and (or (list? expr)
	   (lint-error where "Syntax rules must be a (proper) list: ~s" expr))
       (or (and (pair? expr)
		(eq? 'syntax-rules (first expr)))
	   (lint-error where "Syntax rules must start with `syntax-rules': ~s" expr))
       (form-arity where #f pseudo-syntax-rules-func expr)
       (apply and?
	      (form-syntax-rules-literals (where-follow where (second expr))
					  (second expr))
	      (map (lambda (rule)
		     (let ((where (where-follow where rule)))
		       (and (or (and (list? rule)
				     (= 2 (length rule)))
				(lint-error where "`syntax-rules' rule must be a list (PATTERN TEMPLATE): ~s" rule))
			    (let* ((pattern (first rule))
				   (where   (where-follow where pattern)))
			      (or (and (pair? pattern)
				       (or (eq? sym (first pattern))
					   (and (not option-r5rs)
						(eq? '_ (first pattern)))))
				  (lint-error where "`syntax-rules' pattern must be a list starting with syntax symbol: ~s" pattern)))
			    )))
		   (cddr expr)))))


(lint-arity! define-syntax '(2 0 #f))

(define (form-define-syntax where expr)
  (form-syntax-rules where (second expr) (third expr)))

(define (lint-define-syntax where now expr env)
  (if (form-define-syntax where expr)
      (env-bind-one env where (second expr)
		    (syntax-rules-make-macro (third expr)))))

(lint-handler! define-syntax lint-define-syntax)


;;-----------------------------------------------------------------------------
;; let-syntax

(lint-arity! let-syntax '(2 0 #t))

;; also used for letrec-syntax
(define (form-let-syntax where expr)
  (and (form-bindings where (first expr) (second expr))
       (for-each-bindings where (second expr)
			  (lambda (where sym val)
			    (form-syntax-rules where sym val)))))

(define (lint-let-syntax where now expr env)
  (if (form-let-syntax where expr)
      (begin
	(lint-bindings-duplicates where 'let-syntax (second expr))
	(lint-subenv env
		     (lambda (subenv)
		       (lint-body where now (cddr expr) subenv))))))

(lint-handler! let-syntax lint-let-syntax)


;;-----------------------------------------------------------------------------
;; letrec-syntax

(lint-arity! letrec-syntax '(2 0 #t))

(define (lint-letrec-syntax where now expr env)
  (if (form-let-syntax where expr)
      (lint-subenv env
		   (lambda (subenv)
		     (lint-body where now (cddr expr) subenv)))))

(lint-handler! letrec-syntax lint-letrec-syntax)


