;; Run a test through guile-lint.

;; Copyright 2003, 2004, 2005, 2006, 2007 Kevin Ryde

;; This file is part of Guile-Lint.
;;
;; Guile-Lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; Guile-Lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


;; Usage: guile -s run-test.scm testfile.scm
;;
;; A test file has lines annotated with ";; bad" comments.  guile-lint
;; should report errors or warnings for those lines, and only those lines.
;;
;; Extra text is allowed after ";; bad", eg. ";; bad (my reminder)" for when
;; the problem to be detected isn't obvious.
;;
;; The setups here to run guile-lint uninstalled are a bit nasty.  There's
;; also some slightly nasty pattern matching on the "testfile.scm" filename,
;; allowing guile 1.6 or 1.8 specifics to be confined to when running the
;; respective version, or for gtk tests to be suppressed if not available or
;; no X server.


(use-modules (ice-9 and-let-star)
	     (ice-9 popen)
	     (ice-9 rdelim)
	     (ice-9 regex)
	     (srfi srfi-1)
	     (srfi srfi-13))

(define option-verbose #f)
(putenv "GUILE_WARN_DEPRECATED=no")


;;-----------------------------------------------------------------------------
;; helpers

(if (not (defined? 'hash-for-each))  ;; new in guile 1.8
    (define (hash-for-each proc table)
      (hash-fold (lambda (key value prev)
		   (proc key value))
		 #f table)))

(define (status:description status)
  (or (and-let* ((code (status:exit-val status)))
	(format #f "~a" code))
      (and-let* ((code (or (status:term-sig status)
			   (status:stop-sig status))))
	(format #f "signal ~a"))
      "unknown exit"))

;; compare pairs of numbers xh:xl yh:yl (the "h"s being the high parts,
;; ie. most significant)
(define-public (two>= xh xl yh yl)
  (or (> xh yh)
      (and (= xh yh)
	   (>= xl yl))))

;; as per `resolve-interface' in boot-9.scm
;; an error can be thrown if the module exists but there's bad stuff in it
(define (module-exists? spec)
  (if (pair? (first spec))
      (set! spec (first spec)))
  (and-let* ((module (resolve-module spec)))
    (module-public-interface module)))

(define (path-absolute path)
  (if (string-prefix? "/" path)
      path
      (string-append (getcwd) "/" path)))


;;-----------------------------------------------------------------------------

(define srcdir (or (getenv "srcdir")
		   (error "oops, srcdir not set")))
(set! srcdir (path-absolute srcdir))

(define top_srcdir (string-append srcdir "/.."))

(or (file-exists? "guile-lint")
    (begin
      (mkdir "guile-lint")
      (symlink (string-append top_srcdir "/common.scm") "guile-lint/common.scm")
      (symlink (string-append top_srcdir "/defs.scm") "guile-lint/defs.scm")
      (symlink (string-append top_srcdir "/gdk-common.scm") "guile-lint/gdk-common.scm")
      (symlink (string-append top_srcdir "/slib-option.scm") "guile-lint/slib-option.scm")
      (symlink (string-append top_srcdir "/slib-setups.scm") "guile-lint/slib-setups.scm")
      (symlink top_srcdir "guile-lint/ice-9")
      (symlink top_srcdir "guile-lint/srfi")
      (symlink top_srcdir "guile-lint/oop")
      (symlink (string-append top_srcdir "/gtk") "guile-lint/gtk")
      (symlink (string-append top_srcdir "/gtk-1.2") "guile-lint/gtk-1.2")
      (symlink (string-append top_srcdir "/gtk-2.0") "guile-lint/gtk-2.0")))

(define guile-prog      (or (getenv "GUILE") "guile"))
(define guile-lint-prog "../guile-lint")
(define guile-lint-options "")

(define sourcefile (second (program-arguments)))

;; foo-1.6.scm files are for 1.6 only
(if (string-suffix? "-1.6.scm" sourcefile)
    (or (string=? "1.6" (effective-version))
	(exit 77)))

;; foo-1.8.scm files are for 1.8 and up only
(if (string-suffix? "-1.8.scm" sourcefile)
    (or (two>= (string->number (major-version))
	       (string->number (minor-version))
	       1 8)
	(exit 77)))

;; srfi-NN.scm files check the module exists
(and-let* ((m (string-match "(srfi-[0-9]+)\\.scm$" sourcefile)))
  (or (module-exists? (list 'srfi (string->symbol (match:substring m 1))))
      (begin
	(format #t "(~a not available in guile ~a)\n" 
		(match:substring m 1) (version))
      (exit 77))))

;; slib-option.scm file only if guile.init available
(if (string-suffix? "/slib-option.scm" sourcefile)
    (begin
      (if (getenv "SCHEME_LIBRARY_PATH")
	  (or (file-exists? (string-append (getenv "SCHEME_LIBRARY_PATH")
					   "/guile.init"))
	      (file-exists? (string-append (getenv "SCHEME_LIBRARY_PATH")
					   "/init/guile.init"))
	      (begin
		(format #t "Slib not available - \"SCHEME_LIBRARY_PATH\" file not found: ~a)\n"
			(string-append (getenv "SCHEME_LIBRARY_PATH")
				       "/guile.init"))
		(exit 77)))
	  (or (%search-load-path "slib/guile.init")
	      (%search-load-path "slib/init/guile.init")
	      (begin
		(format #t "Slib not available: slib/guile.init not found in %load-path ~a\n" %load-path)
		(exit 77))))
      (set! guile-lint-options " --slib")))

;; ice9-slib.scm file only if (ice-9 slib) works
(if (string-suffix? "/ice9-slib.scm" sourcefile)
    (begin
      (or (eqv? 0 (status:exit-val
		   (system (string-append guile-prog
					  " -c '(use-modules (ice-9 slib))' >/dev/null 2>&1"))))
	  (begin
	    (display "(ice-9 slib) module doesn't work\n")
	    (exit 77)))))

(define (need-X)
  (or (getenv "DISPLAY")
      (begin
	(display "(X not available, no \"DISPLAY\" environment variable)\n")
	(exit 77))))

;; gdk.scm file only if module available
(if (or (string-suffix? "/gtk.scm" sourcefile)
	(string-suffix? "/gdk.scm" sourcefile))
    (begin
      (need-X)
      (or (module-exists? '(gtk gdk))
	  (exit 77))))

(if (string-suffix? "/gdk-2.0.scm" sourcefile)
    (begin
      (need-X)
      (or (module-exists? '(gtk-2.0 gdk))
	  (exit 77))))
(if (string-suffix? "/gdk-pixbuf-1.2.scm" sourcefile)
    (begin
      (need-X)
      (or (module-exists? '(gtk-1.2 gdk-pixbuf))
	  (exit 77))))
(if (string-suffix? "/gdk-pixbuf-2.0.scm" sourcefile)
    (begin
      (need-X)
      (or (module-exists? '(gtk-2.0 gdk-pixbuf))
	  (exit 77))))

(define want (make-hash-table 997))
(call-with-input-file sourcefile
  (lambda (port)
    (do ((line (read-line port) (read-line port)))
	((eof-object? line))
      (if (string-contains line ";; bad")
	  (hashv-set! want (port-line port) #t)))))

(define got (make-hash-table 997))
(let ((command (string-append guile-prog
			      " -c '(set! %load-path (cons \"" (getcwd) "\"
                                                           %load-path))
                                    (load \"" guile-lint-prog "\")'"
			      guile-lint-options
			      " --exit0 "
			      sourcefile)))
  (if option-verbose
      (format #t "~a\n" command))

  (let ((port    (open-input-pipe command))
	(re      (make-regexp (string-append "^" (regexp-quote sourcefile)
					     ":([0-9]+)"))))
    (do ((line (read-line port) (read-line port)))
	((eof-object? line))
      (cond ((string-contains line "is here")
	     ;; ignore
	     )
	    ((regexp-exec re line)
	     => (lambda (m)
		  (hashv-set! got (string->number (match:substring m 1)) line)))
	    (else
	     (format #t "~a\n" line))))

    (let ((status (close-pipe port)))
      (or (eqv? 0 (status:exit-val status))
	  (begin
	    (format #t "error from guile-lint run: ~a\n"
		    (status:description status))
	    (exit 1))))))

(define bad '())
(hash-for-each
 (lambda (linenum dummy-value)
   (or (hashv-ref got linenum)
       (set! bad (acons linenum
			(format #f "~a:~a: expect bad, but isn't"
				sourcefile linenum)
			bad))))
 want)
(hash-for-each
 (lambda (linenum linestr)
   (or (hashv-ref want linenum)
       (set! bad (acons linenum linestr bad))))
 got)

(define (car< x y)
  (< (car x) (car y)))

(define exit-code 0)

(for-each (lambda (elem)
	    (format #t "~a\n" (cdr elem))
	    (set! exit-code 1))
	  (sort bad car<))

(exit exit-code)
