;; Syntax checking of guile scripts.

;; Copyright 2003, 2004, 2005, 2006, 2007, 2008 Kevin Ryde

;; guile-lint is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; guile-lint is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.


(define-module (guile-lint common)
  #:use-module (ice-9 and-let-star)
  #:use-module (ice-9 format)
  #:use-module (ice-9 optargs)
  #:use-module (ice-9 poe)
  #:autoload   (ice-9 pretty-print) (pretty-print)
  #:autoload   (ice-9 r5rs)         (scheme-report-environment)
  #:autoload   (ice-9 safe-r5rs)    (null-environment)
  #:autoload   (oop goops)  (is-a? <arbiter> <async> <condition-variable>
				   <continuation> <guardian> <mutex> <observer>
				   <random-state>)
  #:use-module (srfi srfi-14)
  #:autoload   (srfi srfi-31)       (rec)
  #:autoload   (guile-lint ice-9 format)  (lint-ice-9-format-string)
  #:autoload  (guile-lint ice-9 optargs) (lint-call-optarity lambda*-with-args)
  #:export-syntax (dv lint-types! receive-list))


;; note the core versions of these funcs, before overriden by srfi-1
(define core:assoc        assoc)
(define core:delete       delete)
(define core:delete!      delete!)
(define core:for-each     for-each)
(define core:map          map)
(define core:map-in-order map-in-order)
(define core:member       member)
(use-modules (srfi srfi-1))

;; note the core string-index before overridden by srfi-13 (for 1.6, where
;; srfi-13 is not in the core)
(define-public core:string-downcase  string-downcase)
(define-public core:string-downcase! string-downcase!)
(define-public core:string-index     string-index)
(define-public core:string-upcase    string-upcase)
(define-public core:string-upcase!   string-upcase!)
(use-modules (srfi srfi-13))



;;-----------------------------------------------------------------------------
;; globals

(define-public option-verbose            #f)
(define-public option-r5rs               #f)
(define-public option-use-srfi-list      '())
(define-public option-no-redefine-macros #f)

;; true if all top level `define's are considered used, as for an slib module
(define-public option-toplevel-used      #f)

;; set to #f if any error printed
(define-public result-all-good #t)


;;-----------------------------------------------------------------------------
;; misc

(define core:provided? provided?) ;; shadowed by slib

(define dump-port (current-error-port))

;; Print the arguments, for diagnostic purposes.
(define-public (d . args)
  (for-each (lambda (elem)
	      (if (pair? elem)
		  (pretty-print-no-newline elem dump-port)
		  (display elem dump-port))
	      (display " " dump-port))
	    args)
  (display "\n" dump-port))

;; Usage: (dv expr ...)
;; Print names and values of expressions, for diagnostic purposes.
;; Literal strings are printed as prefixes to the following argument.
(define-macro (dv . args)
  `(begin
     ,@(map (lambda (elem)
	      (if (string? elem)
		  `(display (string-append ,elem " ") dump-port)
		  `(d ',elem ,elem)))
	    args)))

(define (pretty-print-no-newline obj port)
  (display (string-trim-right (pretty-print-to-string obj)) port))

(define (pretty-print-to-string obj)
  (call-with-output-string
   (lambda (port)
     (pretty-print obj port))))


;; Like `for-each' but allowing an improper list.
;; If LST is a list, PROC is called for each element, including any improper
;; final cdr.  If LST is some other object, PROC is called just once on it.
(define-public (for-each-improper proc lst)
  (cond ((pair? lst)
	 (proc (car lst))
	 (for-each-improper proc (cdr lst)))
	((not (null? lst))
	 (proc lst))))

;; return true if object X is a list of symbols, possibly an improper list,
;; meaning the last cdr can be a symbol too
(define-public (list-of-symbols-improper? x)
  (or (null? x)
      (and (pair? x)
	   (symbol? (car x))
	   (or (symbol? (cdr x))
	       (list-of-symbols-improper? (cdr x))))))

;; Return true if LST length is >= N.
;; Unlike `length', this doesn't traverse the whole of LST, just as much as
;; needed to do the test.
(define-public (length>=? lst n)
  (if (zero? n)
      #t
      (if (null? lst)
	  #f
	  (length>=? (cdr lst) (1- n)))))

;; return true if X is a proper list of at least N elements
(define-public (list>=? x n)
  (and (list? x)
       (length>=? x n)))

;; `and' but as a function so it can be used with `apply'
;; same as in (ice-9 common-list), but don't use that because it provokes
;; redefinition warnings when used with srfi-1
(define-public (and? . args)
  (not (memq #f args)))

;; apply `car' until reaching a non-pair
(define-public (car-to-object x)
  (if (pair? x)
      (car-to-object (car x))
      x))

;; apply `car' until reaching a list starting with a non-pair
(define-public (car-to-last-list lst)
  (if (pair? (car lst))
      (car-to-last-list (car lst))
      lst))

;; number of symbols defined by MODULE
(define (module-count module)
  (let ((n 0))
    (module-for-each (lambda (symbol variable)
		       (set! n (1+ n)))
		     module)
    n))

;; as per `resolve-interface' in boot-9.scm
;; an error can be thrown if the module exists but there's bad stuff in it
(define-public (module-exists? spec)
  (if (pair? (first spec))
      (set! spec (first spec)))
  (and-let* ((module (resolve-module spec)))
    (module-public-interface module)))

;; return a procedure taking the given arguments
(define-public (lambda-with-args formal-list)
  (eval `(lambda ,formal-list #f) (null-environment 5)))

;; compare two symbols, by name
(define (symbol<? x y)
  (string<? (symbol->string x) (symbol->string y)))

;; SPEC is a list of symbols like (foo bar), or like ((foo bar) #:renamer ...)
;; renamer procedures can only draw on the guile core
(define (apply-resolve-interface spec)
  (let ((module (eval (cons resolve-interface
			    (compile-interface-spec spec)) guile-module)))
    (read-enable 'positions) ;; in case module disables this

    (let ((lint-module-spec (cons 'guile-lint (if (list? (first spec))
						  (first spec)
						  spec))))
      (resolve-module lint-module-spec))

    module))

;; procedure name if OBJ is a procedure, or #f if not
(define (procedure-name-maybe obj)
  (cond ((procedure? obj)
	 (or (procedure-name obj)
	     '[anonymous-procedure]))
	((macro? obj)
	 (or (macro-name obj)
	     '[anonymous-macro]))
	(else
	 #f)))

(define  set-first!             set-car!)
;;(define (set-second! lst val)  (set-car! (cdr  lst) val))
(define-public (set-third!  lst val)  (set-car! (cddr lst) val))

(define (not-any pred lst)
  (not (any pred lst)))

(define (integer-non-negative? obj)
  (and (integer? obj) (not (negative? obj))))

;; return a proper list from a possibly improper list or object OBJ
(define-public (improper->proper obj)
  (let more ((obj obj)
	     (ret '()))
    (cond ((pair? obj)
	   (more (cdr obj) (cons (car obj) ret)))
	  ((null? obj)
	   (reverse! ret))
	  (else
	   (reverse! (cons obj ret))))))

(define-macro (receive-list vars expr . body)
  `(apply (lambda ,vars ,@body) ,expr))

;; return true if exception KEY+ARGS is a reader "unexpected )"
(define (error-reader-close-paren? key args)
  (and (memq key '(misc-error   ;; guile 1.6
		   read-error)) ;; guile 1.8
       (list>=? args 2)
       (string-suffix? "unexpected \")\"" (second args))))

;; return true if exception KEY+ARGS is some reader error
(define (error-reader? key args)
  (and (memq key '(misc-error   ;; guile 1.6
		   read-error)) ;; guile 1.8
       (list>=? args 1)
       (equal? (first args) "scm_lreadr")))

;; return the throw key for a reader error; this differs in guile 1.6 and 1.8
(define error-reader-key
  (catch #t (lambda () (read (open-input-string ")")))
	 (lambda (key . args) key)))

;; set the name of PROC to NAME, and return PROC
(define-public (set-procedure-name! name proc)
  (set-procedure-property! proc 'name name)
  proc)

;; `take' but limiting N to the LST length
(define (take-maybe lst n)
  (if (length>=? lst n)
      (take lst n)
      lst))


;;-----------------------------------------------------------------------------
;; arity
;;
;; An "arity" is a list as per procedure-properties 'arity, namely
;; (required-count optional-count more?).
;; REQUIRED-COUNT many arguments are mandatory
;; if MORE? is #f then anything up to OPTIONAL-COUNT many more are allowed,
;; if MORE? is #t then an arbitrary number more are allowed.

;; return true if ARITY will accept N arguments
;; N can be a pair (n . 'more) for n or more from an `apply'
(define (arity-match? arity n)
  (let ((lo        (first arity))
	(hi        (+ (first arity) (second arity)))
	(unlimited (third arity))
	(normore   (pair? n))
	(n         (if (pair? n) (car n) n)))
    (and (or (>= n lo) normore)
	 (or (<= n hi) unlimited))))

;; return true if ARITY will accept N arguments
(define (arity-match-n? arity n)
  (receive-list (required-count optional-count more?)
      arity
    (and (>= n required-count)
	 (or (<= n (+ required-count optional-count))
	     more?))))

;; return true if ARITY will accept N or more arguments
;;
;; ok if ARITY takes unlimited, otherwise need n <= required+optional since
;; an n bigger than that certainly overflows
;;
(define (arity-match-n-or-more? arity n)
  (receive-list (required-count optional-count more?)
      arity
    (or more?
	(<= n (+ required-count optional-count)))))

;; return true if ARITY will accept the given ARGLIST
;; ARGLIST can be an proper list like (1 2 3), or an improper (1 2 . lst)
;; for unknown additional arguments
;;
(define (arity-match-arglist? arity arglist)
  (if (proper-list? arglist)
      (arity-match-n? arity (length arglist))
      (arity-match-n-or-more? arity (1- (length (improper->proper arglist))))))

;; return a string describing ARITY
(define (arity-desc arity)
  (cond ((third arity)
	 (format #f "~s or more" (first arity)))
	((zero? (second arity))
	 (format #f "~s" (first arity)))
	((= 1 (second arity))
	 (format #f "~s or ~s"
		 (first arity) (1+ (first arity))))
	(else
	 (format #f "~s to ~s"
		 (first arity)
		 (+ (first arity) (second arity))))))

;; (define (arity-union x y)
;;   (let ((required (min (first x) (first y)))
;; 	(more?    (or  (third x) (third y)))
;; 	(optional 0))
;;     (if (not more?)
;; 	(set! optional (- (max (+ (first x) (second x))
;; 			       (+ (first y) (second y))) required)))
;;     (list required optional more?)))

;; return an arity for the given lambda style formals
(define (formals->arity formal-list)
  (let ((req 0))
    (while (pair? formal-list)
	   (set! req (1+ req))
	   (set! formal-list (cdr formal-list)))
    (list req 0 (not (null? formal-list)))))


;;-----------------------------------------------------------------------------
;; arity
;;
;; record arity separate from the procedure-property 'arity, to allow for
;; stricter checks than arise from the way they're defined (eg. a "rest" arg
;; whose length ends up checked, etc)

;; PROC vs ARITY, with arity as a list like (1 0 #t)
(define-public lint-arity-hash (make-hash-table 167))

(define-public (lint-arity! obj arity)
  (if (hashq-ref lint-arity-hash obj)
      (error "Already have lint-arity for " obj))
  (hashq-set! lint-arity-hash obj arity))

;; return an arity list like (1 0 #t), or #f if no information (including
;; when OBJ is not a procedure)
(define-public (lint-arity obj)
  (or (hashq-ref lint-arity-hash obj)
      (and (procedure? obj)
	   (procedure-property obj 'arity))))


;;-----------------------------------------------------------------------------
;; special handlers
;;
;; The lint-foo functions are called with a list expression (a proper list),
;; the first element being the procedure as registered with lint-handler!.
;;
;; The form-foo functions check the form of the given expression and return
;; true if it's good.

(define-public lint-handler-property (make-object-property))

(define-public (lint-handler! obj proc)
  (if (lint-handler-property obj)
      (error "Already have lint-handler for " obj))
  (set! (lint-handler-property obj) proc))


;; a procedure to be called (PROC expr) which returns the return value,
;; usually a function taking args
;;
(define lint-return-property (make-object-property))

(define-public (lint-return! obj proc)
  (if (lint-return-property obj)
      (error "Already have lint-return for " obj))
  (set! (lint-return-property obj) proc))

;; (fold PROC INIT LST...) calling (PROC ELEM... PREV)
(define-public (lint-fold where now expr env)
  (lint-procexpr-arglist where now (second expr) env
			 (append (make-list (length (cdddr expr)) 'elem)
				 '(prev)))
  (lint-list where now expr env))


(define-public (lint-apply-handler! proc handler)
  (set-object-property! proc 'lint-apply-handler handler))


;;-----------------------------------------------------------------------------
;; `where' source locations

(define where-type     (make-record-type "where" '(filename line column)))
(define where-type?    (record-predicate where-type))
(define where-new      (record-constructor where-type))
(define where-filename-raw (record-accessor where-type 'filename))
(define where-line-raw     (record-accessor where-type 'line))
(define where-column-raw   (record-accessor where-type 'column))

(define (where-filename where)
  (if where
      (where-filename-raw where)
      "!unknown"))
(define (where-line where)
  (if where
      (where-line-raw where)
      0))
(define (where-column where)
  (if where
      (where-column-raw where)
      0))

(define (where-pair pair)
  (let ((filename (source-property pair 'filename)))
    (if filename
	(where-new filename
		   (1+ (or (source-property pair 'line)
			   -1))
		   (1+ (or (source-property pair 'column)
			   -1)))
	#f)))

;; get a `where' from PORT
(define (where-port port)
  (where-new (port-filename port)
	     (1+ (port-line port))
	     (1+ (port-column port))))

;; Get a `where' from PORT + OBJ.
;; If OBJ is a pair then it ought to have source properties, in which case
;; there should be no need for `where-port', but record that anyway in case
;; a #,() reader extension loses the properties, or whatever.
(define-public (where-port-object port obj)
  (where-follow (where-port port) obj))

;; return a string describing the WHERE location
(define (where-string where)
  (set! where (where-resolve where))
  (format #f "~a:~a:~a:"
	  (where-filename where)
	  (where-line     where)
	  (where-column   where)))

(define (where-resolve-list lst)
  (if (null? lst)
      #f
      (let ((obj (car lst)))
	(cond ((where-type? obj)
	       obj)
	      ((and (pair? obj)
		    (source-property obj 'filename))
	       (where-pair obj))
	      (else
	       (where-resolve-list (cdr lst)))))))

;; return a `where' which is WHERE, but with any deferred source properties
;; examinations resolved (this avoids continuing to reference possibly large
;; input expressions)
(define (where-resolve where)
  (cond ((or (where-type? where)
	     (not where))
	 where)
	((pair? where)
	 (where-resolve-list where))
	(else
	 (error "unrecognised type for where-resolve"))))

;; return a new `where' which is for OBJ if it has location info, or
;; otherwise the given WHERE
(define-public (where-follow where obj)
  (if (pair? where)
      (cons obj where)
      (list obj where)))

(define (where=? x y)
  (and (string=? (where-filename x) (where-filename y))
       (= (where-line x) (where-line y))
       (= (where-column x) (where-column y))))

(define (where<? x y)
  (if (not (string=? (where-filename x) (where-filename y)))
      (string<?      (where-filename x) (where-filename y))
      (if (not (= (where-line x) (where-line y)))
	  (<      (where-line x) (where-line y))
	  (<      (where-column x) (where-column y)))))


;;-----------------------------------------------------------------------------
;; environments

;; TOPLEVEL is the module name like `(guile-user)' or `(foo bar)', or #f for
;; a non-toplevel environment
;;
;; BINDING-HASH is keyed by symbol, each value is a list
;; "(where val used module-elem)"
;;
;; VAL may be a procedure or some other object, but not #f.
;;
;; USED is #t/#f according to whether the binding has been referenced.
;;
;; REFERENCE-LIST is a list of lists (where sym arglist recursive)
;;    WHERE is a where record or list where symbol SYM is used
;;    ARGLIST is a (possibly improper) list of symbols which is the call,
;;    or #f for just the value used
;;    RECURSIVE is true if the reference occurred during the definition of
;;    SYM itself, and thus shouldn't mark that binding as used
;;
;; PREV is a previous environment record, chained to
;;
;; USEMOD-LIST is a list of usemod records for modules brought into scope in
;; the env
;;
(define env-type  (make-record-type
		   "env" '(toplevel binding-hash reference-list prev
				     usemod-list)))
(define-public env-toplevel        (record-accessor env-type 'toplevel))
(define        env-new-raw         (record-constructor env-type))
(define-public env-binding-hash    (record-accessor env-type 'binding-hash))
(define        env-reference-list  (record-accessor env-type 'reference-list))
(define-public env-prev            (record-accessor env-type 'prev))
(define-public env-usemod-list     (record-accessor env-type 'usemod-list))
(define env-toplevel!       (record-modifier env-type 'toplevel))
(define env-reference-list! (record-modifier env-type 'reference-list))
(define env-usemod-list!    (record-modifier env-type 'usemod-list))

(define binding-type  (make-record-type "binding" '(where val used usemod)))
(define binding-new         (record-constructor binding-type))
(define-public binding-where       (record-accessor binding-type 'where))
(define-public binding-val         (record-accessor binding-type 'val))
(define binding-used        (record-accessor binding-type 'used))
(define binding-usemod      (record-accessor binding-type 'usemod))
(define binding-where!      (record-modifier binding-type 'where))
(define binding-val!        (record-modifier binding-type 'val))
(define binding-used!       (record-modifier binding-type 'used))

(define usemod-type  (make-record-type "usemod" '(module where used)))
(define usemod-new         (record-constructor usemod-type))
(define-public usemod-module      (record-accessor usemod-type 'module))
(define-public usemod-where       (record-accessor usemod-type 'where))
(define-public usemod-used        (record-accessor usemod-type 'used))
(define-public usemod-where!      (record-modifier usemod-type 'where))
(define-public usemod-used!       (record-modifier usemod-type 'used))

(define-public (env-dump env)
  (display "env:\n")
  (format #t " toplevel ~a\n" (env-toplevel env))
  (format #t " usemod-list ~a\n" (env-usemod-list env))
  (display " bindings:\n")
  (hash-fold (lambda (sym bind prior)
	       (format #t "  ~a  ~a = ~s  [~a]\n"
		       (where-string (binding-where bind))
		       sym (binding-val bind)
		       (if (binding-used bind) "used" "unused")))
	     #f (env-binding-hash env)))

(define (env-dump-recursive env)
  (env-dump env)
  (cond ((env-prev env) => env-dump-recursive)))


(define-public (env-unbind env sym)
  (hashq-remove! (env-binding-hash env) sym))

(define dummy-usemod (usemod-new #t #t #t))

(define*-public (env-bind-one env where sym val
			      #:optional (used #f) (usemod dummy-usemod))
  ;; "no-redefine-macros" applies to new defines in a top-level environment,
  ;; shadowings in sub-environments are allowed.  This is mostly to help
  ;; when linting boot-9.scm itself.
  (if (not (and option-no-redefine-macros
		(env-toplevel env)
		(macro? (env-binding-val env sym))))
      (begin
	(or (symbol? sym)
	    (error "expect symbol"))
	(if (not val)
	    (set! val #t))
	(hashq-set! (env-binding-hash env) sym
		    (binding-new where val used usemod)))))

;; bind some symbols within ENV
;; OBJ can be a symbol, or a list (proper or improper) of symbols or
;; further lists
(define*-public (env-bind env where obj val #:optional (used #f))
  (if (symbol? obj)
      (env-bind-one env where obj val used)
      (begin
	(set! where (where-follow where obj))
	(for-each-improper (lambda (elem)
			     (env-bind env where elem val used))
			   obj))))

;; bind SYM to VAL in ENV, or if there's already such a binding then just
;; set the value
(define*-public (env-bind-define env where sym val #:optional (used #f))

  (or (and-let* ((binding (hashq-ref (env-binding-hash env) sym #f))
		 (        (eq? dummy-usemod (binding-usemod binding))))
	;; redefinition of binding in this ENV
	;; either an internal env or the toplevel redefining a non-import
	(if (not (env-toplevel env))
	    (lint-error where "duplicate internal define: ~a" sym))
	(binding-where! binding where)
	(binding-val!   binding (or val #t)) ;; not #f
	(binding-used!  binding (or used (binding-used binding)))
	#t)

      (begin
	(and-let* ((        (env-toplevel env))
		   (binding (env-binding env sym))
		   (        (not (eq? dummy-usemod (binding-usemod binding)))))
	  ;; toplevel define of an imported binding
	  (lint-warn where "warning, define shadows import from ~a: ~a"
		     (module-name (usemod-module (binding-usemod binding)))
		     sym))
	(env-bind-one env where sym val used))))

;; return binding record for SYM, or #f
(define-public (env-binding env sym)
  (or (hashq-ref (env-binding-hash env) sym #f)
      (and (env-prev env)
	   (env-binding (env-prev env) sym))))

;; return just the value of SYM in ENV, or #f if unbound
(define (env-binding-val env sym)
  (cond ((env-binding env sym)
	 => binding-val)
	(else
	 #f)))
(define env-defined? env-binding-val)


;; return a new environment, chaining back to PREV
(define-public (env-new prev)
  (env-new-raw #f     ;; not toplevel
	       (make-hash-table 13)
	       '()
	       prev
	       '()))

;; add a reference to SYM within ENV
;; ARGLIST is the parameters, or #f for just a symbol ref
(define-public (env-reference-add env where sym arglist)
  (let ((bind (hashq-ref (env-binding-hash env) sym #f)))
    (if (and bind
	     (not arglist))
	;; plain symbol ref to local variable, just mark used
	(begin
	  (binding-used! bind #t)
	  (usemod-used! (binding-usemod bind) #t))
	;; otherwise note for later
	(env-reference-list! env
	  (cons (list where sym arglist
		      (->bool (memq sym lint-currently-defining)))
		(env-reference-list env))))))

;; call (THUNK), with SYM currently in the process of being defined (so that
;; references to it don't mark it as used)
(define (lint-defining sym thunk)
  (let ((old lint-currently-defining))
    (set! lint-currently-defining (cons sym lint-currently-defining))
    (thunk)
    (set! lint-currently-defining old)))
(define lint-currently-defining '())

;; append the reference-list from SUBENV onto ENV
(define (env-reference-list-append env subenv)
  (env-reference-list! env
		       (append! (env-reference-list subenv)
				(env-reference-list env))))

;; Mung the given MACRO to run in the environment of MODULE.  This is a
;; dirty hack, but without it ice-9 syncase will think the macro is unbound
;; in the environment and won't do any expansion at all.
;;
(define (module-sc-macro-mung module macro)
  (let ((trans       (macro-transformer macro))
	(module-env  (list (module-eval-closure module))))
    (procedure->memoizing-macro
     (lambda (expr env)
       (trans expr module-env)))))

;; return usemod record, or #f
(define (env-module-used? env module)
  (or (find (lambda (usemod)
	      (eq? module (usemod-module usemod)))
	    (env-usemod-list env))
      (and (env-prev env)
	   (env-module-used? (env-prev env) module))))

;; return a where record for VAR, using procedure source properties, or #f
;; this works in guile 1.6, but there's no properties in pre-1.8
(define (variable-where var)
  (and-let* ((       (variable-bound? var))
	     (proc   (variable-ref var))
	     (       (procedure? proc))
	     (source (procedure-source proc))) ;; #f for primitives
    ;; take source-properties from first list expression in source
    (any (lambda (elem)
	   (and (pair? elem)
		(let ((props (source-properties elem)))
		  (and (assq-ref props 'filename)
		       (where-new (assq-ref props 'filename)
				  (assq-ref props 'line)
				  (assq-ref props 'column))))))
	 ;; restrict to just first few exprs, one of them will be a list and
	 ;; if none has properties the the rest is unlikely to
	 (take-maybe source 5))))

;; return true if module NAME has been loaded
;; NAME is a list of symbols, like "(ice-9 q)"
(define (module-loaded? name)
  ;; per boot-9.scm
  (let ((full-name (append '(%app modules) name)))
    (nested-ref the-root-module full-name)))

(define-public (common-sc-macro)
  (if (module-loaded? '(ice-9 syncase))
      (module-symbol-binding (resolve-module '(ice-9 syncase)) 'define-syntax)
      ;; otherwise a dummy unique value
      (cons 1 2)))

;; add bindings from MODULE to ENV
(define*-public (env-bind-module env where top-module #:key (used #f))
  (define sc-macro (common-sc-macro))
  (cond ((env-module-used? env top-module)
	 => (lambda (prev-usemod)
	      (lint-error where "duplicate module import: ~a"
			  (module-name (usemod-module prev-usemod)))
	      (if (usemod-where prev-usemod)
		  (lint-error (usemod-where prev-usemod)
			      "  previous is here"))))
	(else
	 (let ((usemod (usemod-new top-module where used)))
	   (define (bind-module module)
	     (module-for-each
	      (lambda (sym var)
		(env-bind-one env (variable-where var) sym
			      (and-let* ((    (variable-bound? var))
					 (val (variable-ref var)))
				(if (eq? sc-macro val)
				    ;; macros setup to evaluate in given
				    ;; top-module, even those coming from
				    ;; `module-uses'
				    (module-sc-macro-mung top-module val)
				    val))
			      #t usemod))
	      module)
	     (for-each bind-module (module-uses module)))

	   (env-usemod-list! env (cons usemod (env-usemod-list env)))
	   (bind-module top-module)))))

;; return an environment which is all the symbols in MODULE
;; not sure where to get source position info
(define-public (env-from-module where module)
  (let ((env (env-new-raw #f  ;; not top-level
			  (make-hash-table (quotient (module-count module) 3))
			  '()
			  #f
			  '())))
    (env-bind-module env where module)
    env))

(define-public guile-module (resolve-module '(guile)))
(define-public guile-env (env-from-module #f guile-module))

;; lose elisp style `t'
(env-unbind guile-env 't)

;; need to traverse the module-uses to see all symbols
(define (r5rs-env)
  (define env (env-from-module #f (scheme-report-environment 5)))
  (for-each (lambda (m)
	      (env-bind-module env #f m #:used #t))
	    (module-uses (scheme-report-environment 5)))
  ;; macros don't show up, need to add them separately
  (env-bind-one env #f 'and        	 and)
  (env-bind-one env #f 'begin      	 begin)
  (env-bind-one env #f 'cond       	 cond)
  (env-bind-one env #f 'case       	 case)
  (env-bind-one env #f 'define     	 define)
  (env-bind-one env #f 'delay      	 delay)
  (env-bind-one env #f 'do         	 do)
  (env-bind-one env #f 'if         	 if)
  (env-bind-one env #f 'lambda     	 lambda)
  (env-bind-one env #f 'let        	 let)
  (env-bind-one env #f 'let*       	 let*)
  (env-bind-one env #f 'let-syntax 	 let-syntax)
  (env-bind-one env #f 'letrec-syntax letrec-syntax)
  (env-bind-one env #f 'letrec     	 letrec)
  (env-bind-one env #f 'or         	 or)
  (env-bind-one env #f 'quote      	 quote)
  (env-bind-one env #f 'quasiquote 	 quasiquote)
  (env-bind-one env #f 'set!       	 set!)

  ;; empty body not allowed in r5rs
  (let ()
    (define ([r5rs-begin] arg . rest)
      #f)
    (env-bind env #f 'begin [r5rs-begin])
    (lint-handler! [r5rs-begin] lint-begin))

  ;; generate just once
  (set! r5rs-env (lambda () env))
  env)


;;-----------------------------------------------------------------------------
;; gettext stuff

;; return true if PROC is `gettext' or bound to `gettext' in guile 1.8, or
;; is merely the symbol `gettext' or `_' in 1.6
(define (proc-gettext? proc env)
  (if (defined? 'gettext)
      (or (eq? gettext proc)
	  (and (symbol? proc)
	       (eq? gettext (env-binding-val env proc))))
      (memq proc '(gettext _))))

;; return true if PROC is `ngettext' or bound to `ngettext' in guile 1.8, or
;; is merely symbol `ngettext' in 1.6
(define (proc-ngettext? proc env)
  (if (defined? 'ngettext)
      (or (eq? ngettext proc)
	  (and (symbol? proc)
	       (eq? ngettext (env-binding-val env proc))))
      (eq? 'ngettext proc)))

;; Return list of literal strings to use as values of EXPR.
;;
;; EXPR can be a literal "foo", or (gettext "foo"), (ngettext "foo" "bar" n)
;; or (_ "foo").  Other forms result in an empty list.
;;
(define-public (string-gettext-list expr env)
  (filter string?
	  (cond ((and (list>=? expr 2)
		      (proc-gettext? (first expr) env))
		 (list (second expr)))
		((and (list>=? expr 3)
		      (proc-ngettext? (first expr) env))
		 (list (second expr) (third expr)))
		(else
		 (list expr)))))


;;-----------------------------------------------------------------------------
;; error messages

;; `lint-error' returns #f so it can be last in an `or' expression and have
;; that come out false
;; `lint-warn' is the same, but doesn't cause a non-zero final exit code

(define-public (lint-error where fmt . args)
  (set! result-all-good #f)
  (apply lint-warn where fmt args))
''(guile-lint format lint-error 1)

(define-public (lint-warn where fmt . args)
  ;; could use ~k here, but it doesn't show up well in backtraces
  (format #t "~a ~a\n"
	  (where-string where)
	  (apply format #f fmt args))
  #f)
''(guile-lint format lint-warn 1)

;; print an error with location information taken from PORT
(define (lint-error-port port fmt . args)
  (apply lint-error (where-port port) fmt args))
''(guile-lint format lint-error-port 1)

;; print LINT-MESSAGE, plus the message from an `scm-error' style exception
(define (lint-error-error where lint-message key subr message args rest)
  (apply lint-error where (string-append lint-message " " message) args))


;;-----------------------------------------------------------------------------
;; support

;; expressions (ie. lists) in ARGLIST replaced by dummy "x" symbols
(define (arglist-flatten arglist)
  (map (lambda (elem)
	 (if (pair? elem)
	     'x
	     elem))
       arglist))

;; call (PROC subenv) with a new sub-environment chained back to ENV, and
;; when PROC returns join any free variables from subenv onto ENV
;;
(define-public (lint-subenv env proc)
  (let ((subenv (env-new env)))
    (proc subenv)
    (lint-resolve-local subenv)
    (lint-show-unused-modules subenv)
    (lint-show-unused subenv)
    (env-reference-list-append env subenv)))

;; print warnings for all unused modules in ENV
(define (lint-show-unused-modules env)
  (if (not (null? (env-usemod-list env)))
      (let ((lst (remove usemod-used (env-usemod-list env))))
	(for-each (lambda (usemod)
		    (usemod-where! usemod (where-resolve
					   (usemod-where usemod))))
		  lst)
	(for-each (lambda (usemod)
		    (lint-warn (usemod-where usemod)
			       "warning, unused module: ~a"
			       (module-name (usemod-module usemod))))
		  (sort! lst
			 (lambda (x y)
			   (where<? (usemod-where x) (usemod-where y))))))))

;; print warnings for all unused local bindings in ENV
(define (lint-show-unused env)
  (for-each (lambda (elem) ;; list (sym where)
	      (lint-warn (second elem) "warning, unused binding: ~a"
			 (first elem)))
	    (sort! (hash-fold
		    (lambda (sym bind result)
		      (if (binding-used bind) ;; accumulate unuseds
			  result
			  (cons (list sym (where-resolve (binding-where bind)))
				result)))
		    '() (env-binding-hash env))
		   (lambda (x y)
		     (where<? (second x) (second y))))))

;; remove references in ENV which are covered by local bindings in ENV
;; the remaining references need to propagate out to parents of ENV
(define (lint-resolve-local env)
  (let ((hash (env-binding-hash env)))
    (env-reference-list! env
      (remove! (lambda (elem)
		 (receive-list (where sym arglist recursive)
		     elem
		   (cond ((hashq-ref hash sym)
			  => (lambda (bind)
			       ;; recursive use doesn't mark as used
			       (if (not recursive)
				   (begin
				     (binding-used! bind #t)
				     (usemod-used! (binding-usemod bind) #t)))
			       (if arglist
				   (lint-call-arglist
				    where
				    (binding-where bind) sym
				    (binding-val bind) arglist))
			       #t)) ;; remove
			 (else
			  (set-first! elem (where-resolve where))
			  #f))))    ;; keep
	       (env-reference-list env)))))

(define (lint-finalize-env env)
  (for-each (lambda (elem)
	      (set-first! elem (where-resolve (first elem))))
	    (env-reference-list env))
  (for-each (lambda (elem)
	      (receive-list (where sym arglist recursive)
		  elem
		(cond ((env-binding env sym)
		       => (lambda (bind)
			    (if arglist
				(lint-call-arglist
				 where
				 (binding-where bind) sym
				 (binding-val bind) arglist))))
		      (else
		       (lint-error where "unbound symbol: ~a" sym)))))
	    (sort! (env-reference-list env) env-reference<?)))

;; compare X and Y, being lists (WHERE SYM)
(define (env-reference<? x y)
  (if (not (where=? (first x) (first y)))
      (where<? (first x) (first y))
      (symbol<? (second x) (second y))))


;; call (PROC where val) for each element in LST, applied with ITER
(define-public (iterate-where iter where lst proc)
  (set! where (where-follow where lst))
  (iter (lambda (elem)
	  (proc (where-follow where elem) elem))
	lst))

;; call (PROC where sym val) for each binding in LST
;; extra binding values like in `do' are passed to PROC too
(define-public (for-each-bindings where lst proc)
  (set! where (where-follow where lst))
  (for-each (lambda (binding)
	      (apply proc (where-follow where binding) binding))
	    lst))

;; error if ENV is not the top-level
(define-public (lint-require-toplevel where now expr env)
  (if (or (not now)
	  (not (env-toplevel env)))
      (lint-error where "`~a' must be at top-level" (first expr))))

;; check FORMAL-LIST for duplicates
;; FORMAL-LIST can be a symbol or a list, the list can be symbols or (ice-9
;; optargs) elements
(define*-public (lint-formals-duplicates where name formal-list
					 #:optional (desc "formal"))
  (let ((seen '()))  ;; list of symbols
    (for-each-improper (lambda (formal)
			 (if (pair? formal)
			     (set! formal (first formal)))
			 (if (symbol? formal)
			     (if (memq formal seen)
				 (lint-error where "`~a' duplicate ~a: ~s"
					     name desc formal)
				 (set! seen (cons formal seen)))))
		       formal-list)))

;; check BINDING-LIST for duplicate symbol bindings
;; BINDING-LIST is a list of x or (x 1) or ((x y) (values 1 2))
;; extras like `do' step expressions are ignored
;; NAME is the originating syntax, eg. 'let
(define-public (lint-bindings-duplicates where name binding-list)
  (set! where (where-follow where binding-list))
  (let ((seen '()))  ;; alist (sym . where)
    (for-each
     (lambda (binding)
       (let ((where (where-follow where binding)))
	 (if (symbol? binding)
	     (set! binding (list binding)))
	 (set! where (where-follow where (first binding)))
	 (for-each-improper
	  (lambda (sym)
	    (cond ((assq-ref seen sym)
		   => (lambda (seen-where)
			(lint-error where "`~a' duplicate binding: ~s" name sym)
			;; don't bother showing prev location if it's the same
			;; (this can be the case in a let-values)
			(if (not (where=? (where-resolve where)
					  (where-resolve seen-where)))
			    (lint-error seen-where "  previous is here"))))
		  (else
		   (set! seen (alist-cons sym where seen)))))
	  (first binding))))
     binding-list)))

;; Return a value to use for EXPR in a binding.
;;
(define-public (lint-value where expr env)
  (set! where (where-follow where expr))
  (cond ((symbol? expr)
	 ;; symbols are looked up immediately, which is not quite right but
	 ;; good in practice
	 (env-binding-val env expr))

	((or (string? expr)
	     (number? expr)
	     (boolean? expr)
	     (vector? expr))
	 ;; self-evaluating literal
	 expr)

	((list>=? expr 1)
	 (let ((proc (first expr)))
	   (if (symbol? proc)
	       (set! proc (env-binding-val env proc)))

	   (cond

	    ((lint-return-property proc)
	     => (lambda (handler)
		  (handler expr)))

	    ;; begin, let, let*, letrec - last form in body
	    ((and (or (eq? proc begin)
		      (eq? proc let)
		      (eq? proc let*)
		      (eq? proc letrec))
		  (list>=? expr 3))
	     (lint-value where (last expr) env))

	    ;; delay - looked into immediately, which is not quite right,
	    ;; but may help sometimes
	    ((and (eq? proc delay)
		  (list>=? expr 2))
	     (lint-value where (last expr) env))

	    ;; lambda - fake procedure with given formals
	    ((eq? proc lambda)
	     (false-if-exception (lambda-with-args (second expr))))

	    ;; lambda* - fake procedure with given formals
	    ((eq? proc lambda*)
	     (false-if-exception (lambda*-with-args (second expr))))

	    ;; make-procedure-with-setter - faked up from getter and setter
	    ((and (eq? proc make-procedure-with-setter)
		  (list>=? expr 3))
	     (let ((get (lint-value where (second expr) env))
		   (set (lint-value where (third expr) env)))
	       (or (procedure? get)
		   (set! get noop))
	       (or (procedure? set)
		   (set! set noop))
	       (make-procedure-with-setter get set)))

	    ;; pk - see through debug printout
	    ((and (eq? proc pk)
		  (list>=? expr 2))
	     (lint-value where (last expr) env))

	    ;; procedure->macro, procedure->memoizing-macro - evaluated in
	    ;; the hope they can be run
	    ((or (eq? proc procedure->macro)
		 (eq? proc procedure->memoizing-macro))

	     (catch #t
	       (lambda ()
		 (eval expr guile-module))
	       (lambda args
		 (apply lint-error-error where "cannot use macro:" args)
		 #f)))

	    ;; quote - literal
	    ((and (eq? proc quote)
		  (list>=? expr 2))
	      (second expr))

	    ;; rec (srfi-31) - look through to expression value or fake
	    ;; procedure with given formals (the real macro expansions
	    ;; involving letrec are too hard to lint-value directly)
	    ((and srfi-31-available?
		  (eq? proc rec)
		  (list>=? expr 3))
	     (if (symbol? (second expr))
		 (lint-value where (third expr) env)
		 (false-if-exception
		  (lambda-with-args (cdr (second expr))))))

	    ((lint-types proc)
	     => (lambda (type-list)
		  (first type-list))))))))


;;-----------------------------------------------------------------------------
;; default args

(define-public lint-defaults-hash (make-hash-table 67))

(define-public (lint-defaults proc)
  (hashq-ref lint-defaults-hash proc))

(define-public (lint-defaults! proc . args)
  (if (hashq-ref lint-defaults-hash proc)
      (error "Already have lint-defaults for " obj))
  (hashq-set! lint-defaults-hash proc args))


;;-----------------------------------------------------------------------------
;; types

(define type-type (make-record-type "type" '(name test supertype-list
						  or-list)))
(define type-constructor     (record-constructor type-type))
(define type-name            (record-accessor type-type 'name))
(define type-test            (record-accessor type-type 'test))
(define type-test!           (record-modifier type-type 'test))
(define-public type-supertype-list  (record-accessor type-type 'supertype-list))
(define-public type-supertype-list! (record-modifier type-type 'supertype-list))
(define-public type-or-list         (record-accessor type-type 'or-list))
(define type-or-list!        (record-modifier type-type 'or-list))

(define-public type-any
  (type-constructor "any" (lambda (obj) #t) '() '()))
(type-supertype-list! type-any (list type-any))
(type-or-list! type-any (list type-any))

;; NAME is a string, the name to be printed for the type
;; #:test TEST is a predicate procedure for literals of the type
;;
(define*-public (make-type name #:key
			   (test      (lambda (obj) #f))
			   (supertype type-any))
  (let ((type (type-constructor name test '() '())))
    (type-or-list! type (list type))
    (type-supertype-list! type (cons type (type-supertype-list supertype)))
    type))

(define-public (type-or . type-list)
  (let* ((test-list (map type-test type-list))
	 (type      (type-constructor (string-join-str (map type-name
							    type-list) " or ")
				      (lambda (obj)
					(any (lambda (test)
					       (test obj))
					     test-list))
				      '() '())))
    (type-or-list! type (cons type (append-map type-or-list type-list)))
    (type-supertype-list! type
			  (cons type
				(delete-duplicates
				 (append-map type-supertype-list type-list)
				 eq?)))
    type))

(define (string-join-str lst str)
  (string-concatenate (append (append-map (lambda (elem)
					    (list elem str))
					  (drop-right lst 1))
			      (list (last lst)))))

;; return true if SUPER is a supertype of TYPE
(define (type-supertype? super type)
  (member super (type-supertype-list type)))

(define (type-compatible? want got)
  (or (any (lambda (type)
	     (type-supertype? type want))
	   (type-or-list got))
      (any (lambda (type)
	     (type-supertype? type got))
	   (type-or-list want))))

;; keyed by PROC, value list (ret arg1 ...)
(define types-hash (make-hash-table 999))

(define-public (lint-types-override! proc ret . arg-list)
  (if (eq? ret 'pure)
      (begin
	(set-object-property! proc 'pure #t)
	(set! ret (first arg-list))
	(set! arg-list (cdr arg-list))
	(if (eq? type-unspecified ret)
	    (error "lint-types!: oops, shouldn't return type-unspecified from pure func" proc))))
  (if (eq? ret 'noreturn)
      (begin
	(set-object-property! proc 'noreturn #t)
	(set! ret (first arg-list))
	(set! arg-list (cdr arg-list))
	(or (eq? type-unspecified ret)
	    (error "lint-types!: oops, noreturn func should return type-unspecified" proc))))
  (hashq-set! types-hash proc (cons ret arg-list)))

;; This is a macro checking the given proc is actually defined because
;; there's so many functions defined or not defined according to system
;; features and build options like --enable-networking.
(define-macro (lint-types! proc . rest)
  `(lint-types-proc! ',proc (false-if-exception ,proc) ,@rest))
(define-public (lint-types-proc! name proc ret . arg-list)
  ;; (if (not proc) (format #t "skip not defined: ~a\n" name))
  (if proc
      (begin
	(if (hashq-ref types-hash proc)
	    (error "lint-types!: already have types for" name proc))
	(apply lint-types-override! proc ret arg-list))))

;; return types list, or #f
(define (lint-types proc)
  (or (hashq-ref types-hash proc)
      (and (procedure-with-setter? proc)
	   (lint-types (procedure proc)))))

;; EXPR is a call like "(make-string 6 #\x)", check the arg types of the
;; actual parameters against what the called procedure expects.
;; No attention is paid to whether the call has the right number of actual
;; parameters.
;;
(define* (lint-check-types where expr env #:key type-list)
  (and-let* ((proc      (lint-value where (first expr) env))
	     (type-list (or type-list (lint-types proc))))
    (let ((type   type-any)
	  (argnum 0))
      (iterate-where
       for-each where (cdr expr)
       (lambda (where arg)
	 (set! argnum (1+ argnum))
	 ;; lock to last type
	 (if (not (null? type-list))
	     (set! type-list (cdr type-list)))
	 (if (not (null? type-list))
	     (set! type (first type-list)))

	 (if (not (eq? type type-unevaluated)) ;; ignore macro args
	     (cond ((and (list>=? arg 2)
			 (eq? quote (lint-value where (first arg) env)))
		    (or ((type-test type) (second arg))
			(lint-error where
				    "`~a' arg ~a invalid type, want ~a: ~s"
				    (procedure-name-maybe proc)
				    argnum
				    (type-name type) arg)))

		   ((and (not (pair? arg))
			 (not (symbol? arg)))
		    (or ((type-test type) arg)
			(lint-error where
				    "`~a' arg ~a invalid type, want ~a: ~s"
				    (procedure-name-maybe proc)
				    argnum
				    (type-name type) arg)))

		   ((list>=? arg 1)
		    (and-let* ((argproc     (lint-value where (first arg) env))
			       (argtype-list (lint-types argproc)))
		      (or (type-compatible? type (first argtype-list))
			  (lint-error where "`~a' arg ~a invalid type, got ~a, want ~a: ~s"
				      (procedure-name-maybe proc)
				      argnum
				      (type-name (first argtype-list))
				      (type-name type) arg)))))))))))


;;-----------------------------------------------------------------------------
;; type definitions

(define-macro (type-test-values . exprs)
  (set! exprs (map (lambda (expr)
		     `(false-if-exception ,expr))
		   exprs))
  `(let ((type-test-values-list (delq #f (list ,exprs))))
     (lambda (obj)
       (memv obj type-test-values-list))))

;; special marker for macro args
(define-public type-unevaluated
  (make-type "unevaluated"))

(define-public type-pair
  (make-type "pair"
	     #:test pair?))

(define-public type-list
  (make-type "list"
	     #:supertype type-pair
	     #:test list?))

(define-public type-list-of
  (pure-funcq (lambda (type)
		(make-type (format #f "list of ~as" (type-name type))
			   #:supertype type-list
			   #:test (lambda (obj)
				    (and (list? obj)
					 (every (type-test type) obj)))))))
(define-public type-boolean
  (make-type "boolean"
	     #:test boolean?))

(define-public type-f
  (make-type "#f"
	     #:supertype type-boolean
	     #:test not))

(define-public type-t
  (make-type "#t"
	     #:supertype type-boolean
	     #:test (lambda (obj) (eq? #t obj))))

(define-public type-symbol
  (make-type "symbol"
	     #:test symbol?))

(define-public type-symbol-or-t
  (type-or type-symbol type-t))

(define-public type-number
  (make-type "number" #:test number?))

(define-public (inexact-number? obj)
  (and (number? obj)
       (inexact? obj)))

(define-public type-inexact
  (make-type "inexact number"
	     #:supertype type-number
	     #:test inexact-number?))

(define-public type-real
  (make-type "real number"
	     #:supertype type-number
	     #:test real?))

;; in 1.6 `exact?' takes any object, in 1.8 only a number
(define-public (exact-number? obj)
  (and (number? obj)
       (exact? obj)))

(define-public type-exact
  (make-type "exact number"
	     #:supertype type-real
	     #:test exact-number?))

;; OBJ is any object, return true if it's an exact integer
(define-public (exact-integer? obj)
  (and (exact-number? obj)
       (integer? obj)))

(define-public type-exact-integer
  (make-type "exact integer"
	     #:supertype type-exact
	     #:test exact-integer?))

(define-public type-exact-integer-nonneg
  (make-type "exact integer non-negative"
	     #:supertype type-exact-integer
	     #:test (lambda (obj)
		      (and (exact-number? obj)
			   (integer? obj)
			   (not (negative? obj))))))

(define-public type-exact-integer-nonneg-or-f
  (type-or type-exact-integer-nonneg type-f))

(define-public type-object-property
  (make-type "object property"
	     #:test procedure-with-setter?))

(define (arbiter? obj)
  (is-a? obj <arbiter>))

(define-public type-arbiter
  (make-type "arbiter"
	     #:test arbiter?))

(define (async? obj)
  (is-a? obj <async>))

(define-public type-async
  (make-type "arbiter"
	     #:test async?))

(define-public type-array
  (make-type "array"
	     ;; in guile 1.6.7 array? is useless, it's true for everything
	     #:test array?))

(define-public type-character
  (make-type "character"
	     #:test char?))

(define (condition-variable? obj)
  (is-a? obj <condition-variable>))

(define-public type-condvar
  (make-type "condition variable"
	     #:test condition-variable?))

(define-public type-dirstream
  (make-type "dirstream"
	     #:test directory-stream?))

(define-public type-module
  (make-type "module"
	     #:test module?))

;; in the post 1.8 head eval and friends only take a module
(define-public type-environment
  (if (defined? 'environment?)
      (make-type "environment" #:test environment?)
      type-module))

;; LOCK_ constants only defined if the system has those constants
(define flock-operation-list
  (delq #f (list (false-if-exception LOCK_SH)
		 (false-if-exception LOCK_EX)
		 (false-if-exception LOCK_UN)
		 (false-if-exception (logior LOCK_SH LOCK_NB))
		 (false-if-exception (logior LOCK_EX LOCK_NB))
		 (false-if-exception (logior LOCK_UN LOCK_NB)))))

(define-public type-flock-operation
  (make-type "flock LOCK_ operation value"
	     #:supertype type-exact-integer ;; an "int" in C
	     #:test (lambda (obj)
		      (memv obj flock-operation-list))))

(define-public type-fluid
  (make-type "fluid"
	     #:test fluid?))

(define-public type-grent
  (make-type "grent object"
	     #:test vector?))

(define (guardian? obj)
  (is-a? obj <guardian>))

(define-public type-guardian
  (make-type "guardian"
	     #:test guardian?))

(define-public type-hashtable
  (make-type "hash table"
	     #:test (if (defined? 'hash-table?)
			hash-table?  ;; own smob in 1.8
			vector?)))   ;; just a vector in 1.6

(define-public type-hook
  (make-type "hook"
	     #:test hook?))

(define-public type-hostent
  (make-type "hostent object"
	     #:test vector?))

(define-public type-keyword
  (make-type "keyword"
	     #:test keyword?))

(define (mutex? obj)
  (is-a? obj <mutex>))

(define-public type-mutex
  (make-type "mutex object"
	     #:test mutex?))

(define-public type-netent
  (make-type "netent object"
	     #:test vector?))

(define-public (alist? obj)
  (and (list? obj)
       (every pair? obj)))

(define-public type-circular-list
  (make-type "circular list"
	     #:supertype type-pair
	     #:test circular-list?))

(define-public type-list-or-f
  (type-or type-list type-f))

;; a single array bound, being either a number like "4" or a list oflow high
;; range like "(3 5)"
;;
(define-public type-arraybound
  (type-or type-exact-integer type-list))
(type-test! type-arraybound
	    (lambda (obj)
	      (or (and (exact-number? obj)
		       (positive? obj))
		  (and (list? obj)
		       (= 2 (length obj))
		       (exact-number? (first obj))
		       (exact-number? (second obj))
		       (>= (second obj) (first obj))))))

(define-public type-port
  (make-type "port"
	     #:test port?))

(define-public type-input-port
  (make-type "input port"
	     #:test input-port?
	     #:supertype type-port))

(define-public type-output-port
  (make-type "output port"
	     #:test output-port?
	     #:supertype type-port))

(define-public (output-string-port? obj)
  (and (port? obj)
       (false-if-exception (get-output-string obj))))

(define-public type-output-string-port
  (make-type "output string port"
	     #:test output-string-port?
	     #:supertype type-output-port))

(define-public type-fd-or-port
  (type-or type-port type-exact-integer-nonneg))

(define-public type-primprop
  (make-type "primitive property"))

(define-public type-print-state
  (make-type "print state"))

(define-public type-procedure
  (make-type "procedure"
	     #:test procedure?))

(define-public type-procedure-with-setter
  (make-type "procedure with setter"
	     #:test procedure-with-setter?
	     #:supertype type-procedure))

(define-public type-procedure-or-f
  (type-or type-procedure type-f))

(define-public type-macro
  (make-type "macro"
	     #:test macro?))

(define-public type-thunk
  (make-type "thunk"
	     #:test thunk?
	     #:supertype type-procedure))

(define-public type-predicate
  (make-type "predicate"
	     #:test (lambda (obj)
		      (and-let* ((      (procedure? obj))
				 (arity (procedure-property obj 'arity)))
			(arity-match? arity 1)))))

(define-public type-equal-proc
  (make-type "equality procedure"
	     #:test (lambda (obj)
		      (and-let* ((      (procedure? obj))
				 (arity (procedure-property obj 'arity)))
			(arity-match? arity 2)))))

;; PRIO bindings only defined if the system has those constants, and for
;; example mingw doesn't
(define prio-which-list
  (delq #f (list (false-if-exception PRIO_PROCESS)
		 (false-if-exception PRIO_PGRP)
		 (false-if-exception PRIO_USER))))

(define-public type-prio-which
  (make-type "getpriority `which' value"
	     #:supertype type-exact-integer ;; an "int" in C
	     #:test (lambda (obj)
		      (memv obj prio-which-list))))

(define-public type-protoent
  (make-type "protoent object"
	     #:test vector?))

(define-public type-pwent
  (make-type "pwent object"
	     #:test vector?))

(define-public type-record
  (make-type "record type"
	     #:test record?))

(define-public type-rectype
  (make-type "record type"
	     #:test record-type?))

(define-public type-stat
  (make-type "stat object"
	     #:test vector?))

(define-public type-sockaddr
  (make-type "sockaddr object"
	     #:test vector?))

(define-public type-servent
  (make-type "servent object"
	     #:test vector?))

(define-public type-string
  (make-type "string"
	     #:test string?))

(define-public type-string-or-f
  (type-or type-string type-f))

(define-public char-set:open-flags-extras (char-set #\+ #\0 #\l #\b))

(define-public (open-file-mode-string? mode)
  (and (string? mode)
       (not (string-null? mode))                        ;; not empty
       (string-index "rwa" (string-ref mode 0))         ;; start rwa
       (string-every char-set:open-flags-extras mode 1) ;; rest some of "+0lb"
       (or (not (string-index mode #\0 1))              ;; not 0 and l together
	   (not (string-index mode #\l 1)))
       (or (open-file-mode-output? mode)                ;; no l on output
	   (not (string-index mode #\l 1)))))

(define-public type-open-file-mode-string
  (make-type "open mode string"
	     #:test open-file-mode-string?))

(define-public type-port-or-string
  (type-or type-string type-port))

(define-public type-promise
  (make-type "promise"
	     #:test promise?))

(define (random-state? obj)
  (is-a? obj <random-state>))

(define-public type-random-state
  (make-type "random state"
	     #:test random-state?))

(define-public type-vectag
  (make-type "uniform vector tag type"
	     #:supertype type-symbol
	     #:test (lambda (obj)
		      (memq obj '(s8 u8 s16 u16 s32 u32 s64 u64
				     f32 f64 c32 c64)))))

(define-public type-thread
  (make-type "thread"))

(define-public type-vector
  (make-type "vector"
	     #:test vector?))

(define-public type-list-or-vector
  (type-or type-list type-vector))

(define-public type-uname
  (make-type "uname object"
	     #:test vector?))

(define-public type-unspecified
  (make-type "unspecified"
	     #:test (let ((unspecified (if #f #f)))
		      (lambda (obj) (eq? unspecified obj)))))

(define-public type-values
  (make-type "multiple values"))

(define-public type-exit-status
  ;; actually it has a specific range, but start with this
  type-exact-integer-nonneg)

(define-public type-pid
  ;; actually limited to range of a pid_t, either 16 or 32 bits normally
  type-exact-integer-nonneg)

(define seek-whence-list (list SEEK_SET SEEK_CUR SEEK_END))

(define-public type-seek-whence
  (make-type "seek `whence' value"
	     #:supertype type-exact-integer-nonneg
	     #:test (lambda (obj)
		      (memv obj seek-whence-list))))

(define (continuation? obj)
  (is-a? obj <continuation>))

(define-public type-continuation
  (make-type "continuation"
	     #:test continuation?))

(define-public type-debug-object
  (make-type "debug object"
	     #:test debug-object?))

(define-public type-stack
  (make-type "stack"
	     #:test stack?))

(define-public type-frame
  (make-type "frame"
	     #:test frame?))

(define-public type-frame-or-f
  (type-or type-frame type-f))

;; all these are conditionalized in posix.c according to whether they're
;; defined by the system
(define locale-category-list
  (delq #f (list (false-if-exception LC_COLLATE)
		 (false-if-exception LC_CTYPE)
		 (false-if-exception LC_MONETARY)
		 (false-if-exception LC_NUMERIC)
		 (false-if-exception LC_TIME)
		 (false-if-exception LC_MESSAGES)
		 (false-if-exception LC_ALL)
		 (false-if-exception LC_PAPER)
		 (false-if-exception LC_NAME)
		 (false-if-exception LC_ADDRESS)
		 (false-if-exception LC_TELEPHONE)
		 (false-if-exception LC_MEASUREMENT)
		 (false-if-exception LC_IDENTIFICATION))))

(define-public type-locale-category
  (make-type "locale `LC' category"
	     #:supertype type-exact-integer ;; an "int" in C
	     #:test (lambda (obj)
		      (memv obj locale-category-list))))

(define-public type-tm
  (make-type "tm object"
	     #:test vector?))

(define-public type-tms
  (make-type "tms object"
	     #:test vector?))

;; return true if X is a bitwise subset of Y, ie. any 1 bit in X is also in Y
(define (logsubset? x y)
  (not (logtest x (lognot y))))

;; WNOHANG and WUNTRACED conditionalized in guile posix.c, and in particular
;; don't seem to exist on mingw
(define waitpid-flags-mask
  (logior (or (false-if-exception WNOHANG) 0)
	  (or (false-if-exception WNOHANG) 0)))

(define-public type-waitpid-options
  (make-type "waitpid `options' value"
	     #:supertype type-exact-integer-nonneg
	     #:test (lambda (obj)
		      (and (exact-integer? obj)
			   (logsubset? obj waitpid-flags-mask)))))

(define-public type-pair-of
  (pure-funcq (lambda* (type #:optional (type-cdr type))
		(make-type (if (eq? type type-cdr)
			       (format #f "pair of ~as" (type-name type))
			       (format #f "pair of ~a and ~a"
				       (type-name type) (type-name type-cdr)))
			   #:supertype type-pair
			   #:test (lambda (obj)
				    (and (pair? obj)
					 ((type-test type) (car obj))
					 ((type-test type-cdr) (cdr obj))))))))

(define type-pair-of-sec-usec (type-pair-of type-exact-integer-nonneg))

(define-public type-alist
  (make-type "association list"
	     #:supertype type-list
	     #:test alist?))

;; return true if every element of VEC satisfies PRED
(define (vector-every pred vec)
  (every pred (vector->list vec)))

(define-public type-vector-of
  (pure-funcq (lambda (type)
		(make-type (format #f "vector of ~as" (type-name type))
			   #:supertype type-vector
			   #:test (lambda (obj)
				    (and (vector? obj)
					 (vector-every (type-test type)
						       obj)))))))


;;-----------------------------------------------------------------------------
;; core types

;; srfi-4, srfi-13 and srfi-14 in the core in guile 1.8
(cond-expand (srfi-4 (use-modules (guile-lint srfi srfi-4)))
	     (else #f))
(cond-expand (srfi-13 (use-modules (guile-lint srfi srfi-13)))
	     (else #f))
(cond-expand (srfi-14 (use-modules (guile-lint srfi srfi-14)))
	     (else #f))

;; weak vectors in the core in guile 1.6
(if (defined? 'weak-vector)
    (use-modules (guile-lint ice-9 weak-vector)))

(lint-types! %library-dir       'pure type-string)
(lint-types! %package-data-dir  'pure type-string)
(lint-types! %site-dir          'pure type-string)
(lint-types! %search-load-path  'pure type-string type-string)

(lint-types! +             'pure type-number type-number)
(lint-types! -             'pure type-number type-number)
(lint-types! *             'pure type-number type-number)
(lint-types! 1+            'pure type-number type-number)
(lint-types! 1-            'pure type-number type-number)

(lint-types! abs           'pure type-number type-number)
(lint-types! $abs          'pure type-real type-real)
(lint-types! accept       (type-pair-of type-port type-sockaddr) type-port)
(lint-types! access?       type-boolean type-string type-exact-integer-nonneg)
(lint-types! acons         'pure type-alist type-any type-any type-alist)
(lint-types! add-hook!  type-unspecified type-hook type-procedure type-any)
(lint-types! all-threads (type-list-of type-thread)) ;; not in 1.6.7
(lint-types! alarm         type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! angle         'pure type-number type-number)

(lint-types! array?            'pure type-boolean type-any)
(lint-types! array->list       'pure type-list type-array)
(lint-types! array-contents    'pure type-array type-array)
(lint-types! array-copy!       type-array type-array)
(lint-types! array-copy-in-order!  type-array type-array)
(lint-types! array-dimensions  'pure type-list type-array)
(lint-types! array-equal?      'pure type-boolean type-array)
(lint-types! array-fill!       type-array type-any)
(lint-types! array-in-bounds?  'pure type-boolean
	     type-array type-exact-integer)
(lint-types! array-index-map!  type-unspecified type-array type-procedure)
(lint-types! array-rank        'pure type-exact-integer-nonneg type-array)
(lint-types! array-ref         'pure type-any type-array type-exact-integer)
(lint-types! array-set!        type-unspecified
	     type-array type-any type-exact-integer)
(lint-types! array-shape       'pure type-list type-array)
(lint-types! array-type    'pure type-vectag type-array) ;; new in 1.8

(lint-types! ash           'pure type-exact-integer
	     type-exact-integer type-exact-integer)

(lint-types! core:assoc    'pure type-pair type-any type-alist)
(lint-types! assoc-ref     'pure type-any type-alist type-any)
(lint-types! assoc-remove! type-alist type-alist type-any)
(lint-types! assoc-set!    type-alist type-alist type-any type-any)

(lint-types! async         'pure type-async type-thunk)
(lint-types! async-mark    type-unspecified type-async)
(lint-types! run-asyncs    type-unspecified (type-list-of type-async))

(lint-types! acos          'pure type-number type-number)
(lint-types! $acos         'pure type-real type-real)
(lint-types! acosh         'pure type-number type-number)
(lint-types! $acosh        'pure type-real type-real)
(lint-types! asin          'pure type-number type-number)
(lint-types! $asin         'pure type-real type-real)
(lint-types! asinh         'pure type-number type-number)
(lint-types! $asinh        'pure type-real type-real)
(lint-types! atan          'pure type-number type-number type-number)
(lint-arity! atan '(1 1 #f)) ;; only (1 0 #t) in boot-9
(lint-types! $atan         'pure type-real type-real)
(lint-types! $atan2        'pure type-real type-real type-real)
(lint-types! atanh         'pure type-number type-number)
(lint-types! $atanh        'pure type-real type-real)

(lint-types! basename      'pure type-string type-string)
(lint-types! bind  type-unspecified
	     type-port type-exact-integer-nonneg type-any)
(lint-types! bindtextdomain ;; new in 1.8
	     type-string type-string type-string)
(lint-types! bind-textdomain-codeset ;; new in guile 1.8
	     type-string type-string type-string)
(lint-types! bit-extract  'pure type-exact-integer
	     type-exact-integer
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! boolean?     'pure type-boolean type-any)
(lint-types! broadcast-condition-variable ;; not in 1.6.7
	     type-unspecified type-condvar)

(lint-types! call-with-new-thread ;; not in Guile 1.6 if --with-threads=no
	     type-thread type-thunk type-procedure)

(lint-types! car    'pure type-any type-pair)
(lint-types! cdr    'pure type-any type-pair)

(lint-types! caar   'pure type-any type-pair)
(lint-types! cdar   'pure type-any type-pair)
(lint-types! cadr   'pure type-any type-pair)
(lint-types! cddr   'pure type-any type-pair)

(lint-types! caaar  'pure type-any type-pair)
(lint-types! cdaar  'pure type-any type-pair)
(lint-types! cadar  'pure type-any type-pair)
(lint-types! cddar  'pure type-any type-pair)
(lint-types! caadr  'pure type-any type-pair)
(lint-types! cdadr  'pure type-any type-pair)
(lint-types! caddr  'pure type-any type-pair)
(lint-types! cdddr  'pure type-any type-pair)

(lint-types! caaaar 'pure type-any type-pair)
(lint-types! cdaaar 'pure type-any type-pair)
(lint-types! cadaar 'pure type-any type-pair)
(lint-types! cddaar 'pure type-any type-pair)
(lint-types! caadar 'pure type-any type-pair)
(lint-types! cdadar 'pure type-any type-pair)
(lint-types! caddar 'pure type-any type-pair)
(lint-types! cdddar 'pure type-any type-pair)
(lint-types! caaadr 'pure type-any type-pair)
(lint-types! cdaadr 'pure type-any type-pair)
(lint-types! cadadr 'pure type-any type-pair)
(lint-types! cddadr 'pure type-any type-pair)
(lint-types! caaddr 'pure type-any type-pair)
(lint-types! cdaddr 'pure type-any type-pair)
(lint-types! cadddr 'pure type-any type-pair)
(lint-types! cddddr 'pure type-any type-pair)

(lint-types! char?            'pure type-boolean type-any)
(lint-types! char=?           'pure type-boolean type-character type-character)
(lint-types! char<?           'pure type-boolean type-character type-character)
(lint-types! char<=?          'pure type-boolean type-character type-character)
(lint-types! char>?           'pure type-boolean type-character type-character)
(lint-types! char>=?          'pure type-boolean type-character type-character)
(lint-types! char-ci=?        'pure type-boolean type-character type-character)
(lint-types! char-ci<?        'pure type-boolean type-character type-character)
(lint-types! char-ci<=?       'pure type-boolean type-character type-character)
(lint-types! char-ci>?        'pure type-boolean type-character type-character)
(lint-types! char-ci>=?       'pure type-boolean type-character type-character)
(lint-types! char-alphabetic? 'pure type-boolean type-character)
(lint-types! char-numeric?    'pure type-boolean type-character)
(lint-types! char-whitespace? 'pure type-boolean type-character)
(lint-types! char-upper-case? 'pure type-boolean type-character)
(lint-types! char-lower-case? 'pure type-boolean type-character)
(lint-types! char-is-both?    'pure type-boolean type-character)
(lint-types! char->integer    'pure type-exact-integer-nonneg type-character)
(lint-types! char-downcase    'pure type-character type-character)
(lint-types! char-upcase      'pure type-character type-character)
(lint-types! char-ready?      'pure type-boolean type-input-port)

(lint-types! chdir               type-unspecified type-string)
(lint-types! chroot              type-unspecified type-string)
(lint-types! ceiling      	 'pure type-real type-real)
(lint-types! chmod  type-unspecified type-string type-exact-integer-nonneg)
(lint-types! chown  type-unspecified
	     (type-or type-port type-string type-exact-integer-nonneg)
	     type-exact-integer-nonneg)
(lint-types! close            	 type-unspecified type-fd-or-port)
(lint-types! close-fdes        	 type-unspecified type-exact-integer-nonneg)
(lint-types! close-port   	 type-unspecified type-port)
(lint-types! close-input-port    type-unspecified type-input-port)
(lint-types! close-output-port   type-unspecified type-output-port)
(lint-types! closedir            type-unspecified type-dirstream)
(lint-types! closure?            'pure type-boolean type-any)
(lint-types! complex?            'pure type-boolean type-any)
(lint-types! command-line        'pure (type-list-of type-string))
(lint-types! connect   type-unspecified
	     type-port type-exact-integer-nonneg type-any)
(lint-types! cos        	 'pure type-number type-number)
(lint-types! $cos                'pure type-real type-real)
(lint-types! cosh       	 'pure type-number type-number)
(lint-types! $cosh      'pure type-real type-real)
(lint-types! cons       	 'pure type-pair type-any)
(lint-types! cons*      	 'pure type-pair type-any)
(lint-types! cons-source       	 'pure type-pair type-any type-any type-any)
(lint-types! copy-file           type-unspecified type-string type-string)
(lint-types! copy-random-state   type-random-state type-random-state)
(lint-types! copy-tree           'pure type-any type-any)
(lint-types! crypt               'pure type-string type-string type-string)
(lint-types! ctermid             'pure type-string)
(lint-types! current-error-port  'pure type-output-port)
(lint-types! current-input-port  'pure type-input-port)
(lint-types! current-load-port   'pure type-input-port)
(lint-types! current-output-port 'pure type-output-port)
(lint-types! current-thread     'pure type-thread) ;; not in 1.6.7
(lint-types! current-time        'pure type-exact-integer-nonneg)
(lint-types! cuserid             'pure type-string)

(lint-types! core:delete   'pure type-list type-any type-list)
(lint-types! core:delete!  type-list type-any type-list)
(lint-types! delete1! type-list type-any type-list)
(lint-types! delete-file type-unspecified type-string)
(lint-types! denominator         'pure type-exact-integer type-real)
(lint-types! destroy-guardian!   type-unspecified type-guardian)
(lint-types! display  type-unspecified type-any type-output-port)
(lint-types! display-application  type-unspecified
	     type-frame type-port type-exact-integer-nonneg)
(lint-types! display-backtrace  type-unspecified  type-stack type-output-port
	     (type-or type-exact-integer-nonneg type-f)
	     (type-or type-exact-integer-nonneg type-f)
	     type-list)
;; guile 1.6 and 1.8 allow #f for the stack arg, and though that's not
;; actually documented it's probably intentional
(lint-types! display-error  type-unspecified
	     (type-or type-stack type-f) ;; STACK
	     type-output-port  		 ;; PORT
	     type-any          		 ;; SUBR, usually symbol
	     type-string       		 ;; MESSAGE
	     type-list         		 ;; ARGS
	     type-any)         		 ;; REST - unused
(lint-types! directory-stream?  'pure type-boolean type-any)
(lint-types! dirname            'pure type-string type-string)
(lint-types! drain-input        type-unspecified type-input-port)
(lint-types! dup->fdes          type-exact-integer-nonneg
	     type-fd-or-port type-exact-integer-nonneg)
(lint-types! dup2               type-unspecified
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! duplicate-port     type-port type-port type-open-file-mode-string)

(lint-types! effective-version  type-string)
(lint-types! eof-object?    	'pure type-boolean type-any)
(lint-types! endhostent     	type-unspecified)
(lint-types! endnetent      	type-unspecified)
(lint-types! endprotoent    	type-unspecified)
(lint-types! endservent     	type-unspecified)
(lint-types! environ    (type-list-of type-string) (type-list-of type-string))

(lint-types! even?          	'pure type-boolean type-real)

;; in 1.6 `exact?' took any object, in 1.8 only a number
(lint-types! exact?         'pure type-boolean type-number)

(lint-types! exact->inexact 'pure type-inexact type-number)
(lint-types! execl          type-unspecified type-string)
(lint-types! execlp         type-unspecified type-string)
(lint-types! execle         type-unspecified
	     type-string (type-list-of type-string) type-string)

;; scm_exit_status() treats (exit #f) as (exit 1), though that's not
;; documented in 1.6 or 1.8
(lint-types! exit           'noreturn type-unspecified
	     (type-or type-exact-integer-nonneg type-f))
;; any number of args are allowed, and tollerated by scm_exit_status(), but
;; only no args or an exit code are really meaningful
(lint-arity! exit '(0 1 #f))

(lint-types! exp            'pure type-number type-number)
(lint-types! $exp           'pure type-real type-real)
(lint-types! expt           'pure type-number type-number type-number)
(lint-types! $expt          'pure type-real type-real type-real)

(lint-types! fcntl type-any type-fd-or-port type-exact-integer-nonneg type-any)
(lint-types! fdes->ports    (type-list-of type-port) type-exact-integer-nonneg)
(lint-types! fdes->inport   type-input-port type-exact-integer-nonneg)
(lint-types! fdes->outport  type-input-port type-exact-integer-nonneg)
(lint-types! fdopen         type-port type-exact-integer-nonneg type-string)
(lint-types! feature?       'pure type-boolean type-symbol)
(lint-types! file-exists?   'pure type-boolean type-string)
(lint-types! file-port?     'pure type-boolean type-any)
(lint-types! fileno         'pure type-exact-integer-nonneg type-port)
(lint-types! flock type-unspecified type-fd-or-port type-flock-operation)
(lint-types! floor          'pure type-real type-real)
(lint-types! fluid?         'pure type-boolean type-any)
(lint-types! fluid-ref      'pure type-any type-fluid)
(lint-types! fluid-set!     type-any type-fluid type-any)
(lint-types! flush-all-ports type-unspecified)
(lint-types! force           type-any type-promise)
(lint-types! force-output    type-unspecified type-output-port)
(lint-types! frame?          'pure type-boolean type-any)
(lint-types! frame-arguments         type-list type-frame)
(lint-types! frame-evaluating-args?  'pure type-boolean type-frame)
(lint-types! frame-next        'pure type-frame-or-f type-frame)
(lint-types! frame-number      'pure type-exact-integer-nonneg type-frame)
(lint-types! frame-previous    'pure type-frame-or-f type-frame)
(lint-types! frame-procedure?  'pure type-boolean type-frame)
(lint-types! frame-procedure  'pure (type-or type-procedure type-f) type-frame)
(lint-types! frame-source     'pure type-any type-frame)
(lint-types! frame-overflow?  'pure type-boolean type-frame)
(lint-types! frame-real?      'pure type-boolean type-frame)
(lint-types! fsync            type-unspecified type-fd-or-port)
(lint-types! ftell            'pure type-exact-integer-nonneg type-fd-or-port)

(lint-types! gc             type-unspecified)
(lint-types! gc-stats       type-alist)
(lint-types! gcd            'pure type-exact-integer-nonneg type-exact-integer)
(lint-types! gensym             'pure type-symbol type-string)
(lint-types! get-internal-real-time  'pure type-exact-integer-nonneg)
(lint-types! get-internal-run-time   'pure type-exact-integer-nonneg)
(lint-types! get-output-string  'pure type-string type-output-string-port)
(lint-types! get-print-state    'pure type-print-state type-port)
(lint-types! getcwd             'pure type-string)
(lint-types! getenv             'pure type-string-or-f type-string)
(lint-types! getegid            'pure type-exact-integer-nonneg)
(lint-types! geteuid            'pure type-exact-integer-nonneg)

(lint-types! getgr              'pure type-grent
	     (type-or type-exact-integer-nonneg type-string))
(lint-types! getgrgid           'pure type-grent type-exact-integer-nonneg)
(lint-types! getgrnam           'pure type-grent type-string)
(lint-types! setgrent           type-unspecified)
(lint-types! getgrent           type-grent)
(lint-types! endgrent           type-unspecified)
(lint-types! group:name         type-string type-grent)
(lint-types! group:passwd       type-string type-grent)
(lint-types! group:gid          type-exact-integer-nonneg type-grent)
(lint-types! group:mem    (type-list-of type-exact-integer-nonneg) type-grent)

(lint-types! getgid       	'pure type-exact-integer-nonneg)
(lint-types! getgroups        'pure (type-vector-of type-exact-integer-nonneg))
(lint-types! gethost           	'pure type-hostent type-string)
(lint-types! gethostbyname     	'pure type-hostent type-string)
(lint-types! gethostbyaddr     	'pure type-hostent type-exact-integer-nonneg)
(lint-types! gethostent        	type-hostent)
(lint-types! gethostname        'pure type-string)
(lint-types! getlogin           'pure type-string-or-f)
(lint-types! getitimer         	'pure (type-list-of type-pair-of-sec-usec)
	     type-exact-integer-nonneg)
(lint-types! getnet            	'pure type-netent type-string)
(lint-types! getnetbyname      	'pure type-netent type-string)
(lint-types! getnetbyaddr      	'pure type-netent type-exact-integer-nonneg)
(lint-types! getnetent         	type-netent)
(lint-types! getpass       	type-string type-string)
(lint-types! getpeername       	'pure type-sockaddr type-port)
(lint-types! getpgrp       	'pure type-exact-integer-nonneg)
(lint-types! getpid       	'pure type-exact-integer-nonneg)
(lint-types! getppid      	'pure type-exact-integer-nonneg)
(lint-types! getpriority       	'pure type-exact-integer-nonneg
	     type-prio-which type-exact-integer)
(lint-types! getproto          	'pure type-protoent (type-or type-string
						    type-exact-integer-nonneg))
(lint-types! getprotobyname     'pure type-protoent type-string)
(lint-types! getprotobynumber   'pure type-protoent type-exact-integer-nonneg)
(lint-types! getprotoent        'pure type-protoent)
(lint-types! getpw              'pure type-pwent
	     (type-or type-exact-integer-nonneg type-string))
(lint-types! getpwuid           'pure type-pwent type-exact-integer-nonneg)
(lint-types! getpwnam           'pure type-pwent type-string)
(lint-types! setpw              'pure type-pwent type-any)
(lint-types! setpwent           type-unspecified)
(lint-types! getpwent           type-pwent)
(lint-types! endpwent           type-unspecified)
(lint-types! passwd:name        type-string type-pwent)
(lint-types! passwd:passwd      type-string type-pwent)
(lint-types! passwd:uid         type-exact-integer-nonneg type-pwent)
(lint-types! passwd:gid         type-exact-integer-nonneg type-pwent)
(lint-types! passwd:gecos       type-string type-pwent)
(lint-types! passwd:dir         type-string type-pwent)
(lint-types! passwd:shell       type-string type-pwent)

(lint-types! getserv            'pure type-servent
	     type-string type-exact-integer-nonneg)
(lint-types! getservbyname      'pure type-servent type-string type-string)
(lint-types! getservbyport      'pure type-servent
	     type-exact-integer-nonneg type-string)
(lint-types! getservent         type-servent)
(lint-types! getsockname        'pure type-sockaddr type-port)
(lint-types! getsockopt         'pure type-any
	     type-port type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! gettext 'pure type-string ;; new in guile 1.8
	     type-string type-string type-locale-category)
(lint-types! gettimeofday       'pure type-pair-of-sec-usec)
(lint-types! getuid             'pure type-exact-integer-nonneg)
(lint-types! gmtime             'pure type-tm type-exact-integer-nonneg)
(lint-types! guardian-destroyed? 'pure type-boolean type-guardian)
(lint-types! guardian-greedy?    'pure type-boolean type-guardian)

(lint-types! hash              'pure type-exact-integer-nonneg
	     type-any type-exact-integer-nonneg)
(lint-types! hash-clear!      type-unspecified type-hashtable) ;; new in 1.8
(lint-types! hash-create-handle!  'pure type-pair
	     type-hashtable type-any)
(lint-types! hash-get-handle   'pure type-pair
	     type-hashtable type-any)
(lint-types! hash-table?      'pure type-boolean type-any) ;; not in 1.6.7
(lint-types! hash-ref          'pure type-any type-hashtable type-any type-any)
(lint-types! hashq             'pure type-exact-integer-nonneg
	     type-any type-exact-integer-nonneg)
(lint-types! hashq-create-handle!  'pure type-pair
	     type-hashtable type-any)
(lint-types! hashq-get-handle  'pure type-pair
	     type-hashtable type-any)
(lint-types! hashq-ref         'pure type-any type-hashtable type-any type-any)
(lint-types! hashv             'pure type-exact-integer-nonneg
	     type-any type-exact-integer-nonneg)
(lint-types! hashv-create-handle!  'pure type-pair
	     type-hashtable type-any)
(lint-types! hashv-get-handle  'pure type-pair
	     type-hashtable type-any)
(lint-types! hashv-ref         'pure type-any type-hashtable type-any type-any)
(lint-types! hashx-get-handle  'pure type-pair
	     type-procedure type-procedure type-hashtable type-any)
(lint-types! hashx-ref         'pure type-any
	     type-procedure type-procedure type-hashtable type-any type-any)
(lint-types! hook?             'pure type-boolean type-any)
(lint-types! hook-empty?       'pure type-boolean type-hook)
(lint-types! hook->list        'pure (type-list-of type-procedure) type-hook)
(lint-types! hostent:name      'pure type-string type-hostent)
(lint-types! hostent:aliases   'pure (type-list-of type-string) type-hostent)
(lint-types! hostent:addrtype  'pure type-exact-integer-nonneg type-hostent)
(lint-types! hostent:length    'pure type-exact-integer-nonneg type-hostent)
(lint-types! hostent:addr-list 'pure type-list type-hostent)
(lint-types! htons             'pure type-exact-integer type-exact-integer)
(lint-types! htonl             'pure type-exact-integer type-exact-integer)

(lint-types! imag-part    'pure type-real type-number)
(lint-types! inet-aton    'pure type-exact-integer-nonneg type-string)
(lint-types! inet-ntoa    'pure type-string type-exact-integer-nonneg)
(lint-types! inet-lnaof   'pure type-exact-integer-nonneg
	     type-exact-integer-nonneg)
(lint-types! inet-netof   'pure type-exact-integer-nonneg
	     type-exact-integer-nonneg)
(lint-types! inet-makeaddr  'pure type-exact-integer-nonneg
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! inet-ntop  'pure type-string ;; only with IPv6
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! inet-pton  'pure type-exact-integer-nonneg ;; only with IPv6
	     type-exact-integer-nonneg type-string)
(lint-types! inexact?       'pure type-boolean
	     ;; in 1.6 `inexact?' takes any object, in 1.8 only a number
	     (catch 'wrong-type-arg
	       (lambda () (inexact? 'x) type-any)
	       (lambda args             type-number)))
(lint-types! inexact->exact 'pure type-exact type-number)

(lint-types! inf?           'pure type-boolean type-any) ;; new in guile 1.8
(lint-types! input-port?    'pure type-boolean type-any)
(lint-types! integer?       'pure type-boolean type-any)
(lint-types! integer->char  'pure type-character type-exact-integer-nonneg)
(lint-types! integer-expt   'pure type-number type-number
	     ;; no negative exponents in 1.6
	     (if (false-if-exception (integer-expt 4 -1))
		 type-exact-integer
		 type-exact-integer-nonneg))
(lint-types! integer-length 'pure type-exact-integer-nonneg type-exact-integer)
(lint-types! isatty?        'pure type-boolean type-port)

(lint-types! join-thread ;; not in Guile 1.6 if --with-threads=no
	     type-any type-thread)

(lint-types! keyword?        'pure type-boolean type-any)
(lint-types! keyword-dash-symbol 'pure type-symbol type-keyword)
(lint-types! keyword->symbol 'pure type-symbol type-keyword)
(lint-types! kill         type-unspecified
	     type-exact-integer-nonneg type-exact-integer-nonneg)

(lint-types! last-pair         'pure type-pair type-list)
(lint-types! last-stack-frame  'pure type-stack
	     (type-or type-continuation type-debug-object))
(lint-types! lazy-catch   type-any type-symbol-or-t type-thunk type-procedure)
(lint-types! lcm          'pure type-exact-integer type-exact-integer)
(lint-types! length       'pure type-exact-integer-nonneg type-list)
(lint-types! link         type-unspecified type-string type-string)
(lint-types! list         'pure type-list type-any)
(lint-types! list?        'pure type-boolean type-any)
(lint-types! list-copy    'pure type-list type-list)
(lint-types! list-cdr-ref 'pure type-list type-list type-exact-integer-nonneg)
(lint-types! list-cdr-set! type-unspecified
	     type-list type-exact-integer-nonneg type-any)
(lint-types! list-head    'pure type-list type-list type-exact-integer-nonneg)
(lint-types! list-ref     'pure type-any  type-list type-exact-integer-nonneg)
(lint-types! list-set! type-unspecified
	     type-list type-exact-integer-nonneg type-any)
(lint-types! list-tail    'pure type-list type-list type-exact-integer-nonneg)
(lint-types! list->array  'pure type-array
	     (type-or type-exact-integer-nonneg
		      (type-list-of type-exact-integer-nonneg))
	     type-list)
(lint-types! list->string 'pure type-string (type-list-of type-character))
(lint-types! list->typed-array ;; new in guile 1.8
	     'pure type-array type-vectag type-list)
(lint-types! list->vector 'pure type-vector type-list)
(lint-types! listen   type-unspecified ;; only when networking available
	     type-port type-exact-integer-nonneg)
(lint-types! eval  type-any
	     (type-or type-symbol type-list type-number type-string)
	     type-environment)
(lint-types! local-eval   type-any type-any type-environment)
(lint-types! localtime    'pure type-tm type-exact-integer-nonneg type-string)

(lint-types! lock-mutex ;; not in Guile 1.6 if --with-threads=no
	     type-unspecified type-mutex)

(lint-types! log           'pure type-number type-number)
(lint-types! $log          'pure type-real type-real)
(lint-types! log10         'pure type-number type-number)

(lint-types! logand        'pure type-exact-integer type-exact-integer)
(lint-types! logbit?       'pure type-boolean
	     type-exact-integer-nonneg type-exact-integer)
(lint-types! logcount      'pure type-exact-integer-nonneg type-exact-integer)
(lint-types! logior        'pure type-exact-integer type-exact-integer)
(lint-types! lognot        'pure type-exact-integer type-exact-integer)
(lint-types! logtest       'pure type-boolean type-exact-integer)
(lint-types! logxor        'pure type-exact-integer type-exact-integer)
(lint-types! lstat         'pure type-stat type-port-or-string)

(lint-types! macro?              'pure type-boolean)
(lint-types! macro-name          'pure type-string)
(lint-types! macro-transformer   'pure type-procedure type-macro)
(lint-types! macro-type          'pure (type-or type-symbol type-f))
(lint-types! major-version       'pure type-string)
(lint-types! make-arbiter        'pure type-arbiter type-any)
(lint-types! make-array          'pure type-array type-any type-arraybound)
(lint-types! make-condition-variable ;; not in Guile 1.6 if --with-threads=no
	     'pure type-condvar)

(lint-types! make-fluid       'pure type-fluid type-exact-integer-nonneg)
(lint-types! make-hash-table  'pure type-hashtable type-exact-integer-nonneg)
(lint-types! make-hook        'pure type-hook type-exact-integer-nonneg)
(lint-types! make-keyword-from-dash-symbol 'pure type-keyword type-symbol)

(lint-types! make-mutex  ;; not in Guile 1.6 if --with-threads=no
	     'pure type-mutex)

(lint-types! make-polar        'pure type-inexact type-number type-number)
(lint-types! make-soft-port    'pure type-port
	     (type-vector-of type-procedure) type-open-file-mode-string)
(lint-types! make-rectangular  'pure type-inexact type-number type-number)
(lint-types! make-recursive-mutex 'pure type-mutex) ;; not in 1.6.7
(lint-types! make-stack  'pure type-stack
	     (type-or type-continuation type-debug-object type-t)
	     (type-or type-t type-exact-integer-nonneg type-procedure))
(lint-types! make-string  'pure type-string
	     type-exact-integer-nonneg type-character)
(lint-types! make-symbol  'pure type-symbol type-string) ;; not in 1.6.7
(lint-types! make-typed-array 'pure type-array ;; new in guile 1.8
	     type-vectag type-any type-arraybound)
(lint-types! make-vector  'pure type-vector type-exact-integer-nonneg type-any)
(lint-types! %make-void-port  'pure type-port type-open-file-mode-string)
(lint-types! make-weak-key-hash-table    'pure type-hashtable
	     type-exact-integer-nonneg)
(lint-types! make-weak-value-hash-table  'pure type-hashtable
	     type-exact-integer-nonneg)
(lint-types! make-doubly-weak-hash-table 'pure type-hashtable
	     type-exact-integer-nonneg)

(lint-types! magnitude     'pure type-number type-number)
(lint-types! max           'pure type-real type-real)
(lint-arity! max '(1 0 #t)) ;; only (0 0 #t) from scm_tc7_asubr
(lint-types! core:member   'pure type-list type-any type-list)
(lint-types! memoized?     'pure type-boolean type-any)
(lint-types! micro-version 'pure type-string)
(lint-types! min           'pure type-real type-real)
(lint-arity! min '(1 0 #t)) ;; only (0 0 #t) from scm_tc7_asubr
(lint-types! minor-version 'pure type-string)
(lint-types! mkdir    type-unspecified type-string type-exact-integer-nonneg)
(lint-types! mknod    type-unspecified type-string type-symbol
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! mkstemp! type-port type-string)
(lint-types! mktime   'pure (type-pair-of type-exact-integer-nonneg type-tm)
	     type-tm type-string)
(lint-types! modulo   'pure type-exact-integer
	     type-exact-integer type-exact-integer)
(lint-types! modulo-expt 'pure type-exact-integer ;; new in 1.8
	     type-exact-integer type-exact-integer type-exact-integer)
(lint-types! move->fdes type-unspecified
	     type-input-port type-exact-integer-nonneg)

(lint-types! nan?            'pure type-boolean type-any) ;; new in 1.8
(lint-types! negative?       'pure type-boolean type-number)
(lint-types! netent:name     'pure type-string type-netent)
(lint-types! netent:aliases  'pure (type-list-of type-string) type-netent)
(lint-types! netent:addrtype 'pure type-exact-integer-nonneg type-netent)
(lint-types! netent:net      'pure type-exact-integer-nonneg type-netent)
(lint-types! newline         type-unspecified type-output-port)
(lint-types! ngettext  'pure type-string ;; not in 1.6.7
	     type-string type-string type-number
	     type-string type-locale-category)
(lint-types! nice           type-unspecified type-exact-integer)
(lint-types! noop           'pure type-any)
(lint-types! not            'pure type-boolean type-any)
(lint-types! ntohs          'pure type-exact-integer type-exact-integer)
(lint-types! ntohl          'pure type-exact-integer type-exact-integer)
(lint-types! number?        'pure type-boolean type-any)
(lint-types! number->string 'pure type-string
	     type-number type-exact-integer-nonneg)
(lint-types! null?          'pure type-boolean type-any)
(lint-types! numerator      'pure type-exact-integer type-real)

(lint-types! object-properties  'pure type-alist type-any)
(lint-types! object-property    'pure type-any type-any type-symbol)
(lint-types! odd?               'pure type-boolean type-number)
(lint-types! open               'pure type-port
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! open-fdes          'pure type-exact-integer-nonneg
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! open-input-file    'pure type-input-port type-string)
(lint-types! open-input-string  'pure type-input-port type-string)
(lint-types! open-output-file   'pure type-output-port type-string)
(lint-types! open-output-string 'pure type-output-string-port type-string)
(lint-types! opendir            'pure type-dirstream type-string)
(lint-types! output-port?       'pure type-boolean type-any)

(lint-types! pair?                 'pure type-boolean type-any)
(lint-types! parse-path         (type-list-of type-string)
	     type-string (type-list-of type-string))
(lint-types! pause                 type-unspecified)
(lint-types! peek-char             'pure type-character type-input-port)
(lint-types! pipe                  'pure (type-pair-of type-port))
(lint-types! pk                    type-any type-any)
(lint-types! port?                 'pure type-boolean type-any)
(lint-types! port-closed?          'pure type-boolean type-port)
(lint-types! port-column           'pure type-exact-integer-nonneg type-port)
(lint-types! port-filename  	   'pure type-string type-port)
(lint-types! port-line             'pure type-exact-integer-nonneg type-port)
(lint-types! port-mode      	   'pure type-open-file-mode-string type-port)
(lint-types! port-revealed         'pure type-exact-integer-nonneg type-port)
(lint-types! port-with-print-state 'pure type-port type-port type-print-state)
(lint-types! port->fdes            type-exact-integer-nonneg type-port)
(lint-types! positive?             'pure type-boolean type-number)
(lint-types! primitive-exit        'noreturn type-unspecified
	     type-exact-integer-nonneg)
(lint-types! primitive-_exit       'noreturn type-unspecified ;; new in 1.8.1
	     type-exact-integer-nonneg)
(lint-types! primitive-fork        type-exact-integer-nonneg)
(lint-types! primitive-load        type-any type-string)
(lint-types! primitive-load-path   type-any type-string)
(lint-types! primitive-move->fdes type-boolean
	     type-input-port type-exact-integer-nonneg)
(lint-types! program-arguments        'pure (type-list-of type-string))
(lint-types! provide                  type-unspecified type-symbol)
(lint-types! provided?                'pure type-boolean type-symbol)
(lint-types! primitive-property-del!  type-unspecified type-primprop type-any)
(lint-types! primitive-property-ref   'pure type-any type-primprop type-any)
(lint-types! primitive-property-set!  type-unspecified
	     type-primprop type-any type-any)
(lint-types! procedure?               'pure type-boolean type-any)
(lint-types! procedure-documentation  'pure type-string-or-f type-procedure)
(lint-types! procedure-name           'pure type-symbol type-procedure)
(lint-types! procedure-source         'pure type-any type-procedure)
(lint-types! procedure-properties     'pure type-alist type-procedure)
(lint-types! procedure-property      'pure type-any type-procedure type-symbol)
(lint-types! set-procedure-properties! type-unspecified
	     type-procedure type-list)
(lint-types! set-procedure-property! type-unspecified
	     type-procedure type-symbol type-any)
(lint-types! procedure-with-setter?   'pure type-boolean type-any)
(lint-types! promise?                 'pure type-boolean type-any)
(lint-types! protoent:name     'pure type-string type-protoent)
(lint-types! protoent:aliases  'pure (type-list-of type-string) type-protoent)
(lint-types! protoent:proto    'pure type-exact-integer-nonneg type-protoent)
(lint-types! putenv            type-unspecified type-string)

;; quit -- alias for `exit'
(lint-types! quotient   'pure type-exact type-exact type-exact)

(lint-types! raise           type-unspecified type-exact-integer-nonneg)
(lint-types! random          'pure type-real type-real type-random-state)
(lint-types! random:exp      'pure type-real type-random-state)
(lint-types! random:hollow-sphere! type-unspecified 
	     type-vector type-random-state)
(lint-types! random:normal   'pure type-real type-random-state)
(lint-types! random:normal-vector!  type-unspecified
	     type-vector type-random-state)
(lint-types! random:solid-sphere! type-unspecified 
	     type-vector type-random-state)
(lint-types! random:uniform  'pure type-real type-random-state)
(lint-types! seed->random-state 'pure type-random-state
	     (type-or type-number type-string))

(lint-types! rational?       'pure type-boolean type-any)
(lint-types! rationalize  ;; new in guile 1.8
	     'pure type-real type-real type-real)
(lint-types! read            type-any type-input-port)
(lint-types! read-char       type-character type-input-port)
(lint-types! readdir         type-string type-dirstream)
(lint-types! readlink        'pure type-string type-string)
(lint-types! real?           'pure type-boolean type-any)
(lint-types! real-part       'pure type-number type-number)
(lint-types! record?         'pure type-boolean type-any)
(lint-types! record-type-descriptor  'pure type-rectype type-record)
(lint-types! record-type-fields  'pure (type-list-of type-symbol) type-rectype)
(lint-types! record-type-name    'pure type-string type-rectype)
(lint-types! recv!           type-exact-integer-nonneg
	     type-port type-string type-exact-integer-nonneg)
(lint-types! recvfrom!       (type-pair-of type-exact-integer-nonneg
					   (type-or type-sockaddr type-f))
	     type-port type-string type-exact-integer-nonneg)
(lint-types! redirect-port   type-unspecified type-port type-port)
(lint-types! release-arbiter type-unspecified type-arbiter)
(lint-types! remainder       'pure type-exact-integer
	     type-exact-integer type-exact-integer)
(lint-types! remove-hook!    type-unspecified type-hook type-procedure)
(lint-types! rename-file     type-unspecified type-string type-string)
(lint-types! reset-hook!     type-unspecified type-hook)
(lint-types! restore-signals type-unspecified)
(lint-types! reverse         'pure type-list type-list)
(lint-types! reverse!        type-list type-list)
(lint-types! reverse-list->string  'pure type-string type-list)
(lint-types! rewinddir       type-unspecified type-dirstream)
(lint-types! rmdir           type-unspecified type-string)
(lint-types! round           'pure type-number type-number)
(lint-types! run-hook        type-unspecified type-hook type-any)

(lint-types! search-path       'pure type-string-or-f
	     (type-list-of type-string) type-string (type-list-of type-string))
(lint-types! seek              type-exact-integer-nonneg
	     type-fd-or-port type-exact-integer type-seek-whence)
(lint-types! select            (type-list-of type-list-or-vector)
	     type-list-or-vector type-list-or-vector type-list-or-vector
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! send              type-exact-integer-nonneg
	     type-port type-string type-exact-integer-nonneg)
(lint-types! sendto            type-exact-integer-nonneg
	     type-port type-string type-exact-integer-nonneg type-any)
(lint-types! servent:name      'pure type-string type-servent)
(lint-types! servent:aliases   'pure (type-list-of type-string) type-servent)
(lint-types! servent:port      'pure type-exact-integer-nonneg type-servent)
(lint-types! servent:proto     'pure type-exact-integer-nonneg type-servent)

(lint-types! set-car!          type-unspecified type-pair type-any)
(lint-types! set-cdr!          type-unspecified type-pair type-any)
(lint-types! set-current-input-port  type-unspecified type-input-port)
(lint-types! set-current-output-port type-unspecified type-output-port)
(lint-types! set-current-error-port  type-unspecified type-output-port)
(lint-types! set-object-properties!  type-unspecified type-any type-alist)
(lint-types! set-object-property!  type-unspecified
	     type-any type-symbol type-any)
(lint-types! set-port-column!    type-unspecified
	     type-port type-exact-integer-nonneg)
(lint-types! set-port-line!      type-unspecified
	     type-port type-exact-integer-nonneg)
(lint-types! set-port-revealed!  type-unspecified
	     type-port type-exact-integer-nonneg)
(lint-types! set-source-properties!  type-unspecified type-any type-alist)
(lint-types! set-source-property!    type-unspecified
	     type-any type-symbol type-any)

(lint-types! set-tm:sec     type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:min     type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:hour    type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:mday    type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:mon     type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:year    type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:wday    type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:yday    type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:isdst   type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:gmtoff  type-unspecified type-tm type-exact-integer)
(lint-types! set-tm:zone    type-unspecified type-tm type-string)

(lint-types! setenv            type-unspecified type-string type-string)
(lint-types! setegid           type-unspecified type-exact-integer-nonneg)
(lint-types! seteuid           type-unspecified type-exact-integer-nonneg)
(lint-types! setgid            type-unspecified type-exact-integer-nonneg)
(lint-types! setgroups  ;; new in guile 1.8
	     type-unspecified (type-vector-of type-exact-integer-nonneg))
(lint-types! sethost           type-unspecified type-any)
(lint-types! sethostname       type-unspecified type-string)
(lint-types! setitimer         (type-list-of type-pair-of-sec-usec)
	     type-exact-integer-nonneg)
(lint-types! setlocale         type-unspecified
	     type-locale-category type-string)
(lint-types! setnet            type-unspecified type-any)
(lint-types! setpriority       	'pure type-exact-integer-nonneg
	     type-prio-which type-exact-integer type-exact-integer)
(lint-types! setproto          type-unspecified type-any)
(lint-types! setpgid           type-unspecified type-exact-integer-nonneg)
(lint-types! setserv           type-unspecified type-any)
(lint-types! setsid            type-exact-integer-nonneg)
(lint-types! setsockopt        type-unspecified
	     type-port type-exact-integer-nonneg type-exact-integer-nonneg
	     type-any)
(lint-types! setter            'pure type-procedure type-procedure-with-setter)
(lint-types! setuid            type-unspecified type-exact-integer-nonneg)
(lint-types! setvbuf           type-unspecified
	     type-port type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! shared-array-increments  'pure type-exact-integer type-array)
(lint-types! shared-array-offset      'pure type-exact-integer type-array)
(lint-types! shared-array-root        'pure type-array type-array)
(lint-types! shutdown          type-unspecified
	     type-port type-exact-integer-nonneg)

(lint-types! signal-condition-variable ;; not in Guile 1.6 if --with-threads=no
	     type-unspecified type-condvar)
(lint-types! sigaction  type-procedure
	     type-exact-integer-nonneg type-procedure-or-f
	     type-exact-integer-nonneg type-thread)
(lint-types! sin        'pure type-number type-number)
(lint-types! $sin       'pure type-real type-real)
(lint-types! sinh       'pure type-number type-number)
(lint-types! $sinh      'pure type-real type-real)
(lint-types! sleep      type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! sockaddr:fam   'pure type-exact-integer-nonneg type-sockaddr)
(lint-types! sockaddr:path  'pure type-string type-sockaddr)
(lint-types! sockaddr:addr  'pure type-exact-integer-nonneg type-sockaddr)
(lint-types! sockaddr:port  'pure type-exact-integer-nonneg type-sockaddr)
(lint-types! sockaddr:flowinfo ;; new in guile 1.8
	     'pure type-exact-integer-nonneg type-sockaddr)
(lint-types! sockaddr:scopeid  ;; new in guile 1.8
	     'pure type-exact-integer-nonneg type-sockaddr)
(lint-types! socket  type-port
	     type-exact-integer-nonneg type-exact-integer-nonneg
	     type-exact-integer-nonneg)
(lint-types! socketpair  type-port
	     type-exact-integer-nonneg type-exact-integer-nonneg
	     type-exact-integer-nonneg)
(lint-types! source-properties  'pure type-alist type-any)
(lint-types! source-property    'pure type-any type-any type-symbol)
(lint-types! sqrt          'pure type-number type-number)
(lint-types! $sqrt         'pure type-real type-real)
(lint-types! stack?        'pure type-boolean type-any)
(lint-types! stack-id      'pure type-any
	     (type-or type-t type-debug-object type-continuation type-stack))
(lint-types! stack-length  'pure type-exact-integer-nonneg type-stack)
(lint-types! stack-ref   'pure type-frame type-stack type-exact-integer-nonneg)
(lint-types! stat          'pure type-stat type-port-or-string)
(lint-types! stat:dev      'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:ino      'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:mode     'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:nlink    'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:uid      'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:gid      'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:rdev     'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:size     'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:atime    'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:mtime    'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:ctime    'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:blksize  'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:blocks   'pure type-exact-integer-nonneg type-stat)
(lint-types! stat:type     'pure type-symbol type-stat)
(lint-types! stat:perms    'pure type-exact-integer-nonneg 'pure type-stat)

(lint-types! status:exit-val 'pure (type-or type-exact-integer-nonneg type-f)
	     type-exact-integer-nonneg)
(lint-types! status:term-sig 'pure (type-or type-exact-integer-nonneg type-f)
	     type-exact-integer-nonneg)
(lint-types! status:stop-sig 'pure (type-or type-exact-integer-nonneg type-f)
	     type-exact-integer-nonneg)

(lint-types! strerror      'pure type-string  type-exact-integer-nonneg)
(lint-types! strftime      'pure type-string type-string type-tm)

(lint-types! string?       'pure type-boolean type-any)
(lint-types! string=?      'pure type-boolean type-string type-string)
(lint-types! string<?      'pure type-boolean type-string type-string)
(lint-types! string<=?     'pure type-boolean type-string type-string)
(lint-types! string>?      'pure type-boolean type-string type-string)
(lint-types! string>=?     'pure type-boolean type-string type-string)
(lint-types! string-ci=?   'pure type-boolean type-string type-string)
(lint-types! string-ci<?   'pure type-boolean type-string type-string)
(lint-types! string-ci<=?  'pure type-boolean type-string type-string)
(lint-types! string-ci>?   'pure type-boolean type-string type-string)
(lint-types! string-ci>=?  'pure type-boolean type-string type-string)
(lint-types! string-ci->symbol 'pure type-symbol type-string)
(lint-types! string->number 'pure type-number type-string type-exact-integer-nonneg)
(lint-types! string->list   'pure (type-list-of type-character) type-string)
(lint-types! string->symbol 'pure type-symbol type-string)
(lint-types! string-capitalize  'pure type-string type-string)
(lint-types! string-capitalize! type-unspecified type-string)
(lint-types! string-copy    'pure type-string type-string)
(lint-types! string-copy!   type-unspecified
	     type-string type-exact-integer-nonneg
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! core:string-downcase 'pure type-string
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! core:string-downcase! type-unspecified
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! string-fill! type-unspecified
	     type-string type-character
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(or (false-if-exception (core:string-index "0" char-numeric?))
    ;; guile 1.6 core version only takes a char as predicate
    (lint-types! core:string-index  'pure type-exact-integer-nonneg-or-f
		 type-string type-character
		 type-exact-integer-nonneg type-exact-integer-nonneg))
(lint-types! string-length   'pure type-exact-integer-nonneg type-string)
(lint-types! string-null?    'pure type-boolean type-string)
(lint-types! string-ref      'pure type-character
	     type-string type-exact-integer-nonneg)
(or (false-if-exception (string-rindex "0" char-numeric?))
    ;; guile 1.6 core version only takes a char as predicate
    (lint-types! string-rindex   'pure type-exact-integer-nonneg-or-f
		 type-string type-character
		 type-exact-integer-nonneg type-exact-integer-nonneg))
(lint-types! string-set! type-unspecified
	     type-string type-exact-integer-nonneg type-character)
(lint-types! string-split  'pure (type-list-of type-string)
	     type-string type-character)
(lint-types! core:string-upcase 'pure type-string
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! core:string-upcase! type-unspecified
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)

(lint-types! strptime 'pure (type-pair-of type-tm type-exact-integer-nonneg)
	     type-string type-string)
(lint-types! symlink       type-unspecified type-string type-string)
(lint-types! system        type-exit-status type-string)
(lint-types! system*  ;; new in guile 1.8
	     'pure type-exit-status type-string)
(lint-types! system-error-errno  'pure
	     (type-or type-exact-integer-nonneg type-f) type-list)
(lint-types! symbol?           'pure type-boolean type-any)
(lint-types! symbol-hash       'pure type-exact-integer-nonneg type-symbol)
(lint-types! symbol-interned?  'pure type-boolean type-symbol)
(lint-types! symbol->keyword   'pure type-keyword type-symbol)
(lint-types! symbol->string    'pure type-string type-symbol)

(lint-types! substring  'pure type-string
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! substring/copy  'pure type-string ;; new in guile 1.8
	     type-string
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! substring/read-only  'pure type-string ;; new in guile 1.8
	     type-string
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! substring/shared  'pure type-string
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! substring-fill! type-unspecified
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg
	     type-character)
(lint-types! substring-move! type-unspecified
	     type-string type-exact-integer-nonneg type-exact-integer-nonneg
	     type-string type-exact-integer-nonneg)
(lint-types! sync type-unspecified)

(lint-types! tan        'pure type-number type-number)
(lint-types! $tan       'pure type-real type-real)
(lint-types! tanh       'pure type-number type-number)
(lint-types! $tanh      'pure type-real type-real)
(lint-types! tcgetpgrp  'pure type-exact-integer-nonneg type-port)
(lint-types! tcsetpgrp  type-unspecified type-port type-exact-integer-nonneg)
(lint-types! textdomain     type-string type-string)        ;; new in 1.8
(lint-types! thread-exited? 'pure type-boolean type-thread) ;; new in 1.8
(lint-types! throw      type-unspecified type-symbol type-any)
(lint-types! thunk?     'pure type-boolean type-any)
(lint-types! times      'pure type-tms)
(lint-types! tm:sec     'pure type-exact-integer type-tm)
(lint-types! tm:min     'pure type-exact-integer type-tm)
(lint-types! tm:hour    'pure type-exact-integer type-tm)
(lint-types! tm:mday    'pure type-exact-integer type-tm)
(lint-types! tm:mon     'pure type-exact-integer type-tm)
(lint-types! tm:year    'pure type-exact-integer type-tm)
(lint-types! tm:wday    'pure type-exact-integer type-tm)
(lint-types! tm:yday    'pure type-exact-integer type-tm)
(lint-types! tm:isdst   'pure type-exact-integer type-tm)
(lint-types! tm:gmtoff  'pure type-exact-integer type-tm)
(lint-types! tm:zone    'pure type-string type-tm)
(lint-types! tmpnam     type-string)
(lint-types! tms:clock  'pure type-exact-integer-nonneg type-tms)
(lint-types! tms:utime  'pure type-exact-integer-nonneg type-tms)
(lint-types! tms:stime  'pure type-exact-integer-nonneg type-tms)
(lint-types! tms:cutime 'pure type-exact-integer-nonneg type-tms)
(lint-types! transpose-array  'pure type-array type-array type-exact-integer-nonneg)
(lint-types! truncate   'pure type-number type-number)
(lint-types! truncate-file  type-unspecified
	     type-port-or-string type-exact-integer-nonneg)
(lint-types! try-arbiter type-unspecified type-arbiter)
(lint-types! try-mutex  type-boolean type-mutex) ;; not in 1.6.7
(lint-types! ttyname    'pure type-string type-port)
(lint-types! typed-array? ;; new in guile 1.8
	     'pure type-boolean type-any type-vectag)
(lint-types! tzset      type-unspecified)

(lint-types! umask  ;; ENHANCE-ME: pure if no arg
	     type-exact-integer-nonneg type-exact-integer-nonneg)

;; uniform-vector-read! and uniform-vector-write are aliases
(lint-types! uniform-array-read!  type-unspecified
	     type-array type-fd-or-port
	     type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! uniform-array-write  type-unspecified
	     type-array type-fd-or-port
	     type-exact-integer-nonneg type-exact-integer-nonneg)

(if (not (eq? uniform-vector? array?)) ;; not distinguished in 1.6 
    (lint-types! uniform-vector?       type-boolean type-any))
(lint-types! uniform-vector-length type-exact-integer-nonneg type-vector)
(lint-types! uniform-vector-ref    'pure type-real
	     type-vector type-exact-integer-nonneg)
(lint-types! uniform-vector-set!   type-unspecified
	     type-vector type-exact-integer-nonneg type-real)
(lint-types! uniform-vector->list  ;; new in guile 1.8
	     'pure (type-list-of type-real) type-vector)

(lint-types! unlock-mutex  ;; not in Guile 1.6 if --with-threads=no
	     type-unspecified type-mutex)
(lint-types! unmemoize     'pure type-any type-any) ;; gone in guile 1.8

(lint-types! unread-char       type-unspecified type-character type-input-port)
(lint-types! unread-string     type-unspecified type-string type-input-port)
(lint-types! unsetenv          type-unspecified type-string) ;; new in 1.8
(lint-types! uname             'pure type-uname)
(lint-types! usleep        type-exact-integer-nonneg type-exact-integer-nonneg)
(lint-types! utime             type-unspecified type-string type-exact-integer)
(lint-types! utsname:sysname   'pure type-string type-uname)
(lint-types! utsname:nodename  'pure type-string type-uname)
(lint-types! utsname:release   'pure type-string type-uname)
(lint-types! utsname:version   'pure type-string type-uname)
(lint-types! utsname:machine   'pure type-string type-uname)

(lint-types! values            'pure type-values type-any)
(lint-types! vector?           'pure type-boolean type-any)
(lint-types! vector-copy   'pure type-vector type-vector) ;; not in 1.6.7
(lint-types! vector-move-left! type-unspecified
	     type-vector type-exact-integer-nonneg type-exact-integer-nonneg
	     type-vector type-exact-integer-nonneg)
(lint-types! vector-move-right! type-unspecified
	     type-vector type-exact-integer-nonneg type-exact-integer-nonneg
	     type-vector type-exact-integer-nonneg)
(lint-types! vector-length 'pure type-exact-integer-nonneg type-vector)
(lint-types! vector-ref    'pure type-any type-vector type-exact-integer-nonneg)
(lint-types! vector-set!   type-unspecified
	     type-vector type-exact-integer-nonneg type-any)
(lint-types! vector-fill!  type-unspecified type-vector type-any)
(lint-types! vector->list  'pure type-list type-vector)
(lint-types! version       'pure type-string)

(lint-types! waitpid (type-pair-of type-pid type-exit-status)
	     type-exact-integer ;; actually positive or negative PID
	     type-waitpid-options)
(lint-types! wait-condition-variable ;; not in Guile 1.6 --with-threads=no
	     type-unspecified
	     type-condvar type-mutex
	     (type-or type-exact-integer type-pair))
(lint-types! weak-key-hash-table?    'pure type-boolean type-any)
(lint-types! weak-value-hash-table?  'pure type-boolean type-any)
(lint-types! doubly-weak-hash-table? 'pure type-boolean type-any)
(lint-types! with-output-to-string  type-string type-thunk)
(lint-types! with-input-from-string type-any type-string type-thunk)
(lint-types! write            type-unspecified type-any type-output-port)
(lint-types! write-char       type-unspecified type-character type-output-port)

(lint-types! yield  ;; not in guile 1.6 if --with-threads=no
	     type-unspecified)

(lint-types! zero?            'pure type-boolean type-number)


;;-----------------------------------------------------------------------------
;; @

(if (defined? '@) ;; new in guile 1.8
    (begin

      (define (lint-@ where now expr env)
	(let ((spec (second expr))
	      (sym  (third expr)))
	  (if (not (module-exists? spec))
	      (lint-error where "`@' no such module: ~s" spec)

	      (let ((mod (module-public-interface (resolve-module spec))))
		(if (not (module-variable mod sym))
		    (lint-error where "`@' no such public variable in ~s: ~s"
				spec sym)

		    ;; ENHANCE-ME: could look for the requested variable
		    ;; bound under a different name by a renamer
		    (and-let* ((bind   (env-binding env sym))
			       (usemod (binding-usemod bind))
			       (       (eq? mod (usemod-module usemod))))
		      (lint-warn where "hint, `@' variable already in use-modules: ~s" sym)))))))

      (lint-arity! @ '(2 0 #f)) ;; macro
      (lint-handler! @ lint-@)))


;;-----------------------------------------------------------------------------
;; @@

(if (defined? '@@) ;; new in guile 1.8
    (begin

      (define (lint-@@ where now expr env)
	(let ((spec (second expr))
	      (sym  (third expr)))
	  (cond ((not (module-exists? spec))
		 (lint-error where "`@@' no such module: ~s" spec))
		((not (module-variable (resolve-module spec) sym))
		 (lint-error where "`@@' no such variable in ~s: ~s" spec sym))
		((module-variable (module-public-interface
				   (resolve-module spec)) sym)
		 (lint-warn where "hint, `@@' variable is public in ~s, can use plain `@': ~s" spec sym)))))

      (lint-arity! @@ '(2 0 #f))
      (lint-handler! @@ lint-@@)))


;;-----------------------------------------------------------------------------
;; <= < = > >=

(lint-types! <= 'pure type-boolean type-number)
(lint-types! <  'pure type-boolean type-number)
(lint-types! =  'pure type-boolean type-number)
(lint-types! >  'pure type-boolean type-number)
(lint-types! >= 'pure type-boolean type-number)

(define (lint-< where now expr env)
  (or (length>=? expr 3)
      (lint-warn where "warning, `~a' always true on less than two arguments: ~s" (first expr) expr))
  (lint-list  where now (cdr expr) env))

(lint-handler! <= lint-<)
(lint-handler! <  lint-<)
(lint-handler! =  lint-<)
(lint-handler! >  lint-<)
(lint-handler! >= lint-<)


;;-----------------------------------------------------------------------------
;; /

(lint-types! / 'pure type-number type-number)

;; zero args allowed by arity (dispatched through generics ...)
(lint-arity! / '(1 0 #t))

(define (lint-/ where now expr env)

  ;; note same is not true of (/ 1.0 X), since that forces to an inexact
  (and (= 3 (length expr))
       (eqv? 1 (second expr))
       (lint-warn where "hint, `(/ 1 X)' can be just `(/ X)'"))

  (lint-list where now expr env))

(lint-handler! / lint-/)


;;-----------------------------------------------------------------------------
;; and

(lint-arity! and '(0 0 #t))

;; stop looking after literal #f, to ignore tests deliberately dummied out
(define (lint-and where now expr env)
  (case (length expr)
    ((1) (lint-warn where "warning, `and' always true on no arguments"))
    ((2) (lint-warn where "warning, `and' unnecessary on one argument")))
  (let more ((lst (cdr expr)))
    (or (null? lst)
	(case (lint-if-condition where now (first lst) env)
	  ((#f)
	   ;; compile-time false, stop testing
	   #f)
	  (else
	   (lint-object where now (first lst) env)
	   (more (cdr lst)))))))

(lint-handler! and lint-and)


;;-----------------------------------------------------------------------------
;; append

(set-object-property! append 'pure #t)

(define-public (lint-append where now expr env)
  ;; lint-types system can't easily express different type-any for last arg
  (let ((numargs (1- (length expr))))
    (lint-check-types where expr env
		      ;; one `type-list' for the return, then numargs-1 more
		      ;; for those args, and a final `type-any'
		      #:type-list (append! (make-list numargs type-list)
					   (list type-any))))
  (lint-list where now (cdr expr) env))

(lint-handler! append  lint-append)
(lint-handler! append! lint-append)

(define (lint-append-apply where now expr env)
  (if (and (= 3 (length expr))
	   ;; srfi-1 is in scope
	   (eq? concatenate (lint-value where 'concatenate env)))
      (lint-warn where "hint, `(apply append LST)' can be `(concatenate LST)'")))
(lint-apply-handler! append lint-append-apply)


;;-----------------------------------------------------------------------------
;; apply

;; only (1 0 #t) in its zany eval.c definition
(lint-arity! apply '(2 0 #t))

(define-public (lint-apply where now expr env)

  ;; lint-types system can't easily express different type-list for last arg
  (let ((numargs (1- (length expr))))
    (lint-check-types where expr env
		      ;; one `type-any' for the return, then numargs-1 more
		      ;; for those args, followed by final `type-list'
		      #:type-list (append! (make-list numargs type-any)
					   (list type-list))))

  (and-let* ((proc (lint-value where (second expr) env))
	     (handler (object-property proc 'lint-apply-handler)))
    (handler where now expr env))

  ;; arity match of call against fixed + variable args
  (let* ((proc       (second expr))
	 (fixed-args (drop-right (cddr expr) 1))
	 (arglist    (append (arglist-flatten fixed-args) 'apply-list)))
    (lint-procexpr-arglist where now proc env arglist))

  ;; type check of fixed args in the call
  (lint-check-types where (drop-right (cdr expr) 1) env)

  (lint-list where now (cdr expr) env))

(lint-handler! apply lint-apply)


;;-----------------------------------------------------------------------------
;; array-for-each

(lint-types! array-for-each  type-unspecified type-procedure type-array)

;; (for-each PROC lst...), calling (PROC elem...)
(define-public (lint-for-each where now expr env)
  (lint-procexpr-arglist where now (second expr) env
			 (make-list (length (cddr expr)) 'elem))
  (lint-list where now expr env))

(lint-handler! array-for-each lint-for-each)


;;-----------------------------------------------------------------------------
;; array-map!

;; 1.6.7 and a few other versions had demanded at least one source arg, but
;; that was a bug, the arity (2 0 #t) which is in the code is right
;;
;; (lint-arity! array-map! '(2 0 #t))

(lint-types! array-map!  type-unspecified type-array type-procedure type-array)

(define (lint-array-map! where now expr env)
  (lint-procexpr-arglist where now (third expr) env
			 (make-list (length (cdddr expr)) 'elem))
  (lint-list where now expr env))

(lint-handler! array-map! lint-array-map!)


;;-----------------------------------------------------------------------------
;; array-map-in-order!

;; in 1.6.x (procedure-property array-map-in-order! 'arity) is only (2 0 #t),
;; but at least one source array is required (and that's checked as the
;; first thing in the function)
(lint-arity! array-map-in-order! '(3 0 #t))

(lint-types! array-map-in-order!  type-unspecified
	     type-array type-procedure type-array)

(lint-handler! array-map-in-order! lint-array-map!)


;;-----------------------------------------------------------------------------
;; assq

;; used for delq, memq too
(define (lint-assq where now expr env)
  (let ((arg (second expr)))
    (and-let* ((pair (find (lambda (pair)
			     ((car pair) arg))
			   (if option-r5rs
			       lint-eq?-r5rs-alist
			       lint-eq?-guile-alist))))
      (lint-error where "`~a' test `eq?' is ~a: ~s"
		  (first expr) (cdr pair) arg)))
  (lint-list where now (cdr expr) env))

(lint-types! assq 'pure type-pair type-any type-alist)
(lint-handler! assq lint-assq)


;;-----------------------------------------------------------------------------
;; assq-ref

;; also used by delq, memq
(define (lint-assq-ref where now expr env)
  (lint-assq where now (cons* (first expr)
			      (third expr)
			      (second expr)
			      (cdddr expr)) env))

(lint-types! assq-ref  'pure type-any type-alist type-any)
(lint-handler! assq-ref lint-assq-ref)


;;-----------------------------------------------------------------------------
;; assq-remove!

(lint-types! assq-remove!  type-alist type-alist type-any)
(lint-handler! assq-remove! lint-assq-ref)

;;-----------------------------------------------------------------------------
;; assq-set!

(lint-types! assq-set!     type-alist type-alist type-any type-any)
(lint-handler! assq-set! lint-assq-ref)


;;-----------------------------------------------------------------------------
;; assv

;; also used for delv, memv
(define (lint-assv where now expr env)
  (let ((arg (second expr)))
    (and-let* ((pair (find (lambda (pair)
			     ((car pair) arg))
			   lint-eqv?-alist)))
      (lint-error where "`~a' test `eqv?' is ~a: ~s"
		  (first expr) (cdr pair) arg)))
  (lint-list where now (cdr expr) env))

(lint-types! assv 'pure type-pair type-any type-alist)
(lint-handler! assv lint-assv)


;;-----------------------------------------------------------------------------
;; assv-ref

(define (lint-assv-ref where now expr env)
  (lint-assv where now (cons* (first expr)
			      (third expr)
			      (second expr)
			      (cdddr expr)) env))

(lint-types! assv-ref  'pure type-any type-alist type-any)
(lint-handler! assv-ref lint-assv-ref)


;;-----------------------------------------------------------------------------
;; assv-remove!

(lint-types! assv-remove!  type-alist type-alist type-any)
(lint-handler! assv-remove! lint-assv-ref)


;;-----------------------------------------------------------------------------
;; assv-set!

(lint-types! assv-set!     type-alist type-alist type-any type-any)
(lint-handler! assv-set! lint-assv-ref)


;;-----------------------------------------------------------------------------
;; begin

;; empty (begin) allowed in guile 1.8, but not in 1.6
(define begin-accepts-empty-body?
  (catch #t
    (lambda ()
      (begin)
      #t)
    (lambda args
      #f)))

(lint-arity! begin (if begin-accepts-empty-body? '(0 0 #t) '(1 0 #t)))

(define (lint-begin where now expr env)
  (if (= (length expr) 2)
      (lint-warn where "hint, `begin' unnecessary around single expression:"))      
  (lint-body-list where now (cdr expr) env))

(lint-handler! begin lint-begin)


;;-----------------------------------------------------------------------------
;; call-with-blocked-asyncs

(if (defined? 'call-with-blocked-asyncs) ;; new in guile 1.8
    (begin

      (define (lint-call-with-blocked-asyncs where now expr env)
	(lint-procexpr-arglist where now (second expr) env '())
	(lint-list where now expr env))

      (lint-types! call-with-blocked-asyncs   type-any type-thunk)

      (lint-handler! call-with-blocked-asyncs lint-call-with-blocked-asyncs)))


;;-----------------------------------------------------------------------------
;; call-with-current-continuation

(lint-types! call-with-current-continuation  type-any type-predicate)

(define (lint-call-with-current-continuation where now expr env)
  (lint-procexpr-arglist where now (second expr) env '(cont))
  (lint-list where now expr env))

(lint-handler! call-with-current-continuation lint-call-with-current-continuation)


;;-----------------------------------------------------------------------------
;; call-with-dynamic-root

(lint-types! call-with-dynamic-root  type-any type-thunk type-procedure)

(define (lint-call-with-dynamic-root where now expr env)
  (lint-procexpr-arglist where now (second expr) env '())
  (lint-procexpr-arglist where now (third expr) env '(key . args))
  (lint-list where now expr env))

(lint-handler! call-with-dynamic-root lint-call-with-dynamic-root)


;;-----------------------------------------------------------------------------
;; call-with-input-file
;; call-with-input-string
;; call-with-output-file

(lint-types! call-with-input-file    type-any type-string type-predicate)
(lint-types! call-with-input-string  type-any type-string type-predicate)
(lint-types! call-with-output-file   type-any type-string type-predicate)

(define (lint-call-with-input-file where now expr env)
  (lint-procexpr-arglist where now (third expr) env '(port))
  (lint-list where now expr env))

(lint-handler! call-with-input-file   lint-call-with-input-file)
(lint-handler! call-with-input-string lint-call-with-input-file)
(lint-handler! call-with-output-file  lint-call-with-input-file)


;;-----------------------------------------------------------------------------
;; call-with-output-string

(lint-types! call-with-output-string    type-string type-predicate)

(define (lint-call-with-output-string where now expr env)
  (lint-procexpr-arglist where now (second expr) env '(port))
  (lint-list where now expr env))

(lint-handler! call-with-output-string lint-call-with-output-string)


;;-----------------------------------------------------------------------------
;; call-with-unblocked-asyncs

(if (defined? 'call-with-unblocked-asyncs) ;; new in guile 1.8
    (begin
      (lint-types! call-with-unblocked-asyncs  type-any type-thunk)
      (lint-handler! call-with-unblocked-asyncs
		     lint-call-with-blocked-asyncs)))


;;-----------------------------------------------------------------------------
;; call-with-values

(lint-types! call-with-values           type-any type-thunk type-thunk)

(define (lint-call-with-values where now expr env)
  (lint-procexpr-arglist where now (second expr) env '())
  (lint-procexpr-arglist where now (third expr) env 'args)
  (lint-list where now expr env))

(lint-handler! call-with-values lint-call-with-values)


;;-----------------------------------------------------------------------------
;; case

(lint-arity! case '(2 0 #t))

(define (form-case-datum where datum last?)
  (set! where (where-follow where datum))
  (if (and (eq? 'else datum)
	   (not last?))
      (lint-error where "`case' `else' must be last"))
  (or (eq? 'else datum)
      (list? datum)
      (lint-error where "`case' datum must be a (proper) list, or `else': ~s" datum)))

(define (form-case-clause where clause last?)
  (set! where (where-follow where clause))
  (if (list>=? clause 1)
      (begin
	(if (not (list>=? clause 2))
	    (lint-error where "`case' empty expression in clause: ~s" clause))
	(form-case-datum where (first clause) last?))
      (lint-error where "`case' clause must be a (proper) list of at least one element: ~s" clause)))

(define (form-case where expr)
  (let ((last-clause (last expr)))
    (apply and?
	   (map (lambda (clause)
		  (form-case-clause where clause (eq? clause last-clause)))
		(cddr expr)))))

;; return true if OBJ is a type that oculd match in a `case' datum
(define (case-datum-could-match? obj)
  (or (symbol?  obj)
      (keyword? obj)
      (number?  obj)
      (boolean? obj)
      (char?    obj)
      (null?    obj)))

(define (lint-case-data where expr)
  (let ((seen '()))
    (iterate-where
     for-each where (cddr expr)
     (lambda (where clause)
       (if (pair? (first clause))
	   (iterate-where
	    for-each where (first clause)
	    (lambda (where datum)
	      (cond ((not (case-datum-could-match? datum))
		     (lint-error where "`case' datum type cannot match: ~s" datum))
		    ((assv-ref seen datum)
		     => (lambda (orig-where)
			  (lint-error where "`case' datum duplicate: ~s" datum)
			  (lint-error orig-where "  previous use is here")))
		    (else
		     (set! seen (alist-cons datum where
					    seen)))))))))))

(define (lint-case where now expr env)
  (if (form-case where expr)
      (begin
	(lint-case-data where expr)
	;; key
	(lint-object where now (second expr) env)
	;; consequents
	(for-each (lambda (elem)
		    (lint-body-list (where-follow where elem)
				    now (cdr elem) env))
		  (cddr expr)))))

(lint-handler! case lint-case)


;;-----------------------------------------------------------------------------
;; catch

(lint-types! catch  type-any type-symbol-or-t type-thunk type-procedure)

(define (lint-catch where now expr env)
  (lint-procexpr-arglist where now (second expr) env '())
  (lint-procexpr-arglist where now (third expr) env '())
  (lint-list where now expr env))

(lint-handler! catch lint-catch)


;;-----------------------------------------------------------------------------
;; cond

(lint-arity! cond '(1 0 #t))

(define (form-cond where expr)
  (apply and?
	 (iterate-where map-in-order where (cdr expr)
			(lambda (where clause)
			  (or (list>=? clause 1)
			      (lint-error where "`cond' clause must be a (proper) list containing at least a test: ~s" clause))))))

;; FIXME: for `=>', propagate "now" into the check of the lambda
;;
(define (lint-cond-clause where now clause env last?)
  (let ((test (first clause))
	(body (cdr clause)))
    
    (cond ((eq? 'else test)
	   (begin
	     (or last?
		 (lint-error where "`cond' `else' clause must be last"))
	     (if (null? body)
		 (lint-error where "`cond' `else' clause must have non-empty body: ~s" clause))
	     (if (and (pair? body)
		      (eq? '=> (first body)))
		 (begin
		   (set! body (cdr body))
		   (lint-error where "`cond' `else' clause cannot have `=>'")))
	     (lint-body-list where now body env)))
	  
	  ((and (not (null? body))
		(eq? '=> (first body)))
	   
	   (if (not (= 2 (length body)))
	       (lint-error where "`cond' `=>' must have single recipient")
	       (lint-expr where now (list (second body) test) env)))
	  
	  (else
	   (lint-body-list where now clause env #:initial-used 1)))))

(define (lint-cond where now expr env)
  (if (form-cond where expr)
      (let ((last-clause (last expr)))
      (iterate-where for-each where (cdr expr)
		     (lambda (where clause)
		       (lint-cond-clause where now clause env
					 (eq? clause last-clause)))))))

(lint-handler! cond lint-cond)


;;-----------------------------------------------------------------------------
;; cond-expand

;; test cond-expand feature SYM within MODULE
(define (module-cond-expand-symbol? module sym)
  (memq sym (hashq-ref %cond-expand-table module '())))

;; test cond-expand feature SYM in ENV
(define (env-cond-expand-symbol? env sym)
  (or (find (lambda (module)
	      (module-cond-expand-symbol? module sym))
	    (map usemod-module (env-usemod-list env)))
      (and-let* ((prev (env-prev env)))
	(env-cond-expand-symbol? prev sym))))

;; test cond-expand FEATURE expression, in ENV
;; FEATURE must have already been validated
(define (env-cond-expand-feature? env feature)
  (define func-alist `((and . ,every)
		       (or  . ,any)
		       (not . ,not-any)))
  (or (eq? 'else feature)
      (let more ((feature feature))
	(cond ((symbol? feature)
	       (or (memq feature %cond-expand-features) ;; globals
		   (env-cond-expand-symbol? env feature)))
	      (else
	       ;; recurse on and/or/not
	       ((assq-ref func-alist (first feature)) more (cdr feature)))))))


(lint-arity! cond-expand '(0 0 #t))

(define (form-cond-expand-feature where feature)
  (set! where (where-follow where feature))
  (cond ((symbol? feature)
	 #t)
	((and (list>=? feature 1)
	      (eq? 'not (first feature)))
	 (if (= 2 (length feature))
	     (form-cond-expand-feature where (second feature))
	     (lint-error where "`cond-expand' feature `not' takes one argument")))
	((and (list>=? feature 1)
	      (memq (first feature) '(and or)))
	 (if (null? (cdr feature))
	     (begin
	       ;; warn about an empty (and) or (or), but continue
	       (lint-error where "`cond-expand' empty `~a' feature form" (first feature))
	       #t) ;;
	     (apply and? (map (lambda (elem)
				(form-cond-expand-feature where elem))
			      (cdr feature)))))
	(else
	 (lint-error where "`cond-expand' invalid feature form: ~s" feature))))

(define (form-cond-expand-clause where clause last?)
  (set! where (where-follow where clause))
  (if (and (list>=? clause 1)
	   (eq? 'else (first clause))
	   (not last?))
      (lint-error where "`cond-expand' `else' must be last"))
  (and (or (list>=? clause 1)
	   (lint-error where "`cond-expand' invalid clause"))
       (form-cond-expand-feature (where-follow where clause)
				 (first clause))))

(define (form-cond-expand where expr)
  (if (null? (cdr expr))
      (lint-error where "`cond-expand' without clauses fails")
      (let ((last-clause (last expr)))
	(apply and?
	       (map (lambda (clause)
		      (form-cond-expand-clause (where-follow where clause)
					       clause (eq? clause last-clause)))
		    (cdr expr))))))

(define (lint-cond-expand where now expr env)
  (if (form-cond-expand where expr)
      (let ((match (find (lambda (clause)
			   (env-cond-expand-feature? env (first clause)))
			 (cdr expr))))
	(if match
	    (lint-body-list where now (cdr match) env)
	    (lint-error where "`cond-expand' no clause succeeds")))))
	    
(lint-handler! cond-expand lint-cond-expand)


;;-----------------------------------------------------------------------------
;; define

(lint-arity! define '(2 0 #t))

(define (form-define-arglist where expr)
  (and (or (not (null? expr))
	   (lint-error where "`define' name-spec cannot be empty: ~s" expr))
       (if option-r5rs
	   (or (symbol? (first expr))
	       (lint-error where "`define' name-spec must start with symbol: ~s" expr))
	   (or (symbol? (first expr))
	       (pair? (first expr))
	       (lint-error where "`define' name-spec must start with symbol or currying list: ~s" expr)))
       (or (symbol? (first expr))
	   (form-define-arglist where (first expr)))
       (or (symbol? (cdr expr))
	   (list-of-symbols-improper? (cdr expr))
	   (lint-error where "`define' formal parameters must be symbols: ~s" (cdr expr)))))

(define (form-define where expr)
  (and (or (symbol? (second expr))
	   (form-define-arglist where (second expr)))
       (or (not (symbol? (second expr)))
	   (= 3 (length expr))
	   (lint-error where "`define' of symbol needs exactly two arguments"))))

;; call (PROC where lst) for each currying level in LST
(define (currying-where where lst proc)
  (set! where (where-follow where lst))
  (proc where lst)
  (if (pair? (first lst))
      (currying-where where (first lst) proc)))

(define-public (currying-formals-flatten formals)
  (let more ((formals formals)
	     (lst     '()))
    (if (symbol? (car formals))
	(append lst (cdr formals))
	(more (car formals)
	      (append lst (improper->proper (cdr formals)))))))

;; return BODY with possible initial docstring removed
;; don't remove if the sole content is a string, that's the return value
(define-public (lint-strip-docstring body)
  (if (and (list>=? body 2)
	   (string? (first body)))
      (cdr body)
      body))

(define*-public (lint-define where now expr env #:optional (used #f))
  (or (env-toplevel env)
      (lint-error where "`define' bad placement, can only be at start of body"))
  (if option-toplevel-used
      (set! used #t))
  (lint-define-allowed where now expr env used))

(define*-public (lint-define-allowed where now expr env #:optional (used #f))
  (if (form-define where expr)
      (cond
       ((symbol? (second expr))
	;; (define SYM VAL)
	(let ((sym (second expr))
	      (val (third expr)))
	  (lint-defining sym
	    (lambda ()
	      (lint-object where now val env)))
	  (env-bind-define env where sym (lint-value where val env) used)))

       (else
	;; (define (SYM ...) ...)
	(let ((formals (second expr)))
	  ;; duplicates between currying levels don't make sense,
	  ;; though they're not an error
	  (lint-formals-duplicates (where-follow where formals) 'define
				   (currying-formals-flatten formals))

	  ;; new definition
	  (let ((name (car-to-object formals))
		(val  (lint-value where
				  (list lambda (cdr (car-to-last-list
						     formals)) #f) env)))
	    (set-procedure-property! val 'name name)
	    (env-bind-define env where name val used)

	    (lint-defining name
	      (lambda ()
		(lint-subenv env
		  (lambda (subenv)
		    (currying-where where formals
				    (lambda (where defs)
				      (env-bind subenv where
						(cdr defs) #t
						#t))) ;; used
		    (lint-body where #f
			       (lint-strip-docstring (cddr expr))
			       subenv)))))))))))

(lint-handler! define lint-define)


;;-----------------------------------------------------------------------------
;; define-macro

(lint-arity! define-macro '(2 0 #t))

(define (form-define-macro where expr)
  (or (list-of-symbols-improper? (second expr))
      (lint-error where "`define-macro' arglist must be a list of symbols: ~s" (second expr))))

(define (lint-define-macro where now expr env)
  (if (form-define-macro where expr)
      (let* ((params (second expr))
	     (body   (cddr expr))
	     (name   (car params))
	     (args   (cdr params)))
	(set! body (lint-strip-docstring body))
	(lint-subenv env
		     (lambda (subenv)
		       (env-bind subenv where args #t #t)
		       (lint-body where #f body subenv)))
	(env-bind env where name
		  (eval-defmacro where name args body)))))

(lint-handler! define-macro lint-define-macro)


;;-----------------------------------------------------------------------------
;; define-module

(lint-arity! define-module '(1 0 #t))

(define (form-define-module-export where arg)
  (or (and (list? arg)
	   (every symbol? arg))
      (lint-error where "`define-module' #:export arguments must be symbols: ~s" arg)))

;; NAME is a symbol like `last', return true if it's a valid duplicates
;; handler strategy for "define-module ... #:duplicates (NAME)"
;; `lookup-duplicates-handlers' used here is undocumented ...
(define (lint-duplicates-handler-name? name)
  (and (symbol? name)
       (false-if-exception (lookup-duplicates-handlers (list name)))))

(define (form-define-module-duplicates where handler-names)
  (for-each (lambda (name)
	      (or (lint-duplicates-handler-name? name)
		  (lint-error where "`define-module' #:duplicates invalid handler: ~s" name)))
	    handler-names))

(define (string->keyword str)
  (symbol->keyword (string->symbol str)))

;; if SYM is a symbol :foo then return a keyword #:foo, otherwise return SYM
;; unchanged
;; this oddity is like guile 1.8 boot-9.scm `keyword-like-symbol->keyword',
;; but taking a stricter approach, in demanding a ":", not just stripping
;; the first char from any symbol
;;
(define (colon-symbol->keyword sym)
  (or (and-let* ((    (symbol? sym))
		 (str (symbol->string sym))
		 (    (string-prefix? ":" str)))
	(string->keyword (string-drop str 1)))
      sym))

;; return true if OBJ is either a keyword, or a symbol that looks like a
;; keyword like `:foo'
(define (keyword-or-colon-symbol? obj)
  (or (keyword? obj)
      (keyword? (colon-symbol->keyword obj))))

(define (lint-define-module where now expr env)
  (define (need n lst)
    (or (length>=? lst (1+ n))
	(lint-error where "`define-module' ~a needs ~a argument~:p"
		    (first lst) n)))
  (lint-require-toplevel where now expr env)

  (let ((modname (second expr)))
    ;; look for existing module under checking
    (set! env (find (lambda (env)
		      (equal? modname (env-toplevel env)))
		    lint-modules-env-list))
    ;; otherwise create a new one
    (if (not env)
	(begin
	  (set! env (lint-make-user-env))
	  (env-toplevel! env modname)
	  (set! lint-modules-env-list (cons env lint-modules-env-list))))
    (set! lint-current-module-env env))

  (let loop ((lst (cddr expr)))
    (if (not (null? lst))
	;; could put symbols `:use-module' etc in the `case' instead of
	;; doing symbol to keyword conversion for the test, but that would
	;; depend on the keywords reader option when loading this code, and
	;; would prefer to safe against either setting of that
	(case (colon-symbol->keyword (first lst))
	  ((#:use-module #:use-syntax)
	   (if (need 1 lst)
	       (begin
		 (env-bind-module env
				  (where-follow where (second lst))
				  (apply-resolve-interface (second lst)))
		 (loop (cddr lst)))))
	  ((#:autoload)
	   (if (need 2 lst)
	       (begin
		 (env-bind-module env
				  (where-follow where (second lst))
				  (apply-resolve-interface (second lst)))
		 (loop (cdddr lst)))))
	  ((#:export #:export-syntax #:re-export #:re-export-syntax #:replace)
	   ;; ENHANCE-ME: make #:export and #:replace references demand
	   ;; local bindings, make #:re-export demand imported bindings
	   (if (need 1 lst)
	       (let* ((syms  (second lst))
		      (where (where-follow where syms)))
		 (if (form-define-module-export where (cons 'export syms))
		     (lint-list where #f syms env))
		 (loop (cddr lst)))))
	  ((#:pure)
	   ;; FIXME: restrict env to standard syntax forms only
	   (loop (cdr lst)))
	  ((#:duplicates)
	   ;; FIXME: this is syntax checked, but has no influence on the
	   ;; bindings imported by #:use-module here or `use-modules' later
	   (if (need 1 lst)
	       (let* ((handler-names (second lst))
		      (where         (where-follow where handler-names)))
		 (form-define-module-duplicates where handler-names)
		 (loop (cddr lst)))))
	  ((#:no-backtrace)
	   (loop (cdr lst)))
	  (else
	   (lint-error where "`define-module' invalid option: ~s"
		       (first lst))
	   ;; attempt to keep going on next keyword
	   (loop (find-tail keyword-or-colon-symbol? (cdr lst))))))))

(lint-handler! define-module lint-define-module)


;;-----------------------------------------------------------------------------
;; define-public

(lint-arity! define-public '(2 0 #t))

(define-public (lint-define-public where now expr env)
  (lint-require-toplevel where now expr env)
  (lint-define where now expr env #t))

(lint-handler! define-public lint-define-public)


;;-----------------------------------------------------------------------------
;; defined?

(lint-types! defined? 'pure type-boolean type-symbol)


;;-----------------------------------------------------------------------------
;; defmacro

(define (eval-defmacro where name args body)
  (and-let* ((proc  (catch #t
		      (lambda ()
			(eval (cons* 'lambda args body) guile-module))
		      (lambda (key . args)
			(lint-value where (cons* lambda args body) #f))))
	     (trans (lambda (expr env)
		      (apply proc (cdr expr))))
	     (macro (procedure->memoizing-macro trans)))
    (set-procedure-property! trans 'name name)
    (lint-arity! macro (formals->arity args))
    macro))

(lint-arity! defmacro '(3 0 #t))

(define (form-defmacro where expr)
  (and? (or (symbol? (second expr))
	    (lint-error where "`defmacro' name must be a symbol: ~s" (second expr)))
	(or (symbol? (third expr))
	    (list-of-symbols-improper? (third expr))
	    (lint-error where "`defmacro' arglist must be a symbol or list of symbols: ~s" (third expr)))))

(define (lint-defmacro where now expr env)
  (if (form-defmacro where expr)
      (begin
	(lint-require-toplevel where now expr env)
	(let ((name (second expr))
	      (args (third expr))
	      (body (cdddr expr)))
	  (lint-subenv env
		       (lambda (subenv)
			 (env-bind subenv where args #t)
			 (lint-body where #f body subenv)))
	  (env-bind-one env where name
			(eval-defmacro where name args body)
			(and option-toplevel-used
			     (env-toplevel env)))))))

(lint-handler! defmacro lint-defmacro)


;;-----------------------------------------------------------------------------
;; delay

(lint-arity! delay '(1 0 #t))

(lint-types! delay               type-promise type-any)

(define (lint-delay where now expr env)
  ;; now becomes false
  (lint-list where #f (cdr expr) env))

(lint-handler! delay lint-delay)


;;-----------------------------------------------------------------------------
;; delq
;; delq!
;; delq1!

(lint-types! delq     'pure type-list type-any type-list)
(lint-types! delq!    type-list type-any type-list)
(lint-types! delq1!   type-list type-any type-list)

(lint-handler! delq   lint-assq)
(lint-handler! delq!  lint-assq)
(lint-handler! delq1! lint-assq)


;;-----------------------------------------------------------------------------
;; delv
;; delv!
;; delv1!

(lint-types! delv     'pure type-list type-any type-list)
(lint-types! delv!    type-list type-any type-list)
(lint-types! delv1!   type-list type-any type-list)

(lint-handler! delv   lint-assv)
(lint-handler! delv!  lint-assv)
(lint-handler! delv1! lint-assv)


;;-----------------------------------------------------------------------------
;; do

(lint-arity! do '(2 0 #t))

(define (form-do where expr)
  (and? (and (or (list? (second expr))
		 (lint-error (where-follow where (second expr))
			     "`do' bindings must be a (proper) list: ~s" (second expr)))
	     (apply and?
		    (iterate-where
		     map-in-order where (second expr)
		     (lambda (where binding)
		       (or (and (or (= 2 (length binding))
				    (= 3 (length binding)))
				(symbol? (first binding)))
			   (lint-error where "`do' binding must be symbol, initializer and optional step: ~s" binding))))))

	(or (list>=? (third expr) 1)
	    (lint-error (where-follow where (third expr))
			"`do' test+result must be a non-empty (proper) list: ~s" (third expr)))))

(define (lint-do where now expr env)
  (if (form-do where expr)
      (begin
	(lint-bindings-duplicates where 'do (second expr))
	;; initializers
	(for-each-bindings where (second expr)
			   (lambda (where sym val . step)
			     (lint-object where now val env)))
	(lint-subenv env
		     (lambda (subenv)
		       ;; add bindings
		       (for-each-bindings where (second expr)
					  (lambda (where sym val . step)
					    (env-bind-one subenv where sym
							  (lint-value where
								      val
								      env))))
		       ;; steps, if present
		       (for-each-bindings where (second expr)
					  (lambda (where sym val . step)
					    (if (not (null? step))
						(lint-object where now
							     (car step)
							     subenv))))
		       ;; test and final calculations
		       (lint-body-list where now (third expr) subenv
				       #:initial-used 1)
		       ;; body
		       (lint-body-list where now (cdddr expr) subenv))))))
(lint-handler! do lint-do)


;;-----------------------------------------------------------------------------
;; dup

;; only (1 0 #t) in boot-9
(lint-arity! dup '(1 1 #f))

(lint-types! dup type-fd-or-port
	     type-fd-or-port type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; dup->inport

;; only (1 0 #t) in boot-9
(lint-arity! dup->inport '(1 1 #f))

(lint-types! dup->inport type-input-port
	     type-fd-or-port type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; dup->outport

;; only (1 0 #t) in boot-9
(lint-arity! dup->outport '(1 1 #f))
(lint-types! dup->outport type-output-port
	     type-fd-or-port type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; dup->port

;; only (2 0 #t) in boot-9
(lint-arity! dup->port '(2 1 #f))

(lint-types! dup->port type-port
	     type-fd-or-port type-open-file-mode-string
	     type-exact-integer-nonneg)


;;-----------------------------------------------------------------------------
;; dynamic-wind

(lint-types! dynamic-wind type-any type-thunk type-thunk type-thunk)

(define (lint-dynamic-wind where now expr env)
  (lint-procexpr-arglist where now (second expr) env '())
  (lint-procexpr-arglist where now (third expr) env '())
  (lint-procexpr-arglist where now (fourth expr) env '())
  (lint-list where now expr env))

(lint-handler! dynamic-wind lint-dynamic-wind)


;;-----------------------------------------------------------------------------
;; environments

;; Despite an environment being a fundamental scheme concept, it's no longer
;; a first class object in guile post 1.8, which is frankly unbelievable,
;; and for that matter hurts something like the linting here which could
;; have done some nice introspection/replication of environments to get
;; proper macro evaluation etc.  The commented out bits here were started
;; with a view to using some of those features -- best omitted if they're
;; going to be shifting sands.

(lint-types! environment?       'pure type-boolean type-any)
(lint-types! environment-bound? 'pure type-boolean
	     type-environment type-symbol)
(lint-types! procedure-environment 'pure type-environment type-procedure)

;; (if (defined? 'environment?)
;;     (begin
;;       (define (observer? obj)
;; 	(is-a? obj <observer>))

;;       (define-public type-observer
;; 	(make-type "observer token"
;; 		   #:test observer?))

;;       (define type-environment-cell
;; 	(type-or type-f        ;; no cell: SCM_ENVIRONMENT_LOCATION_NO_CELL
;; 		 type-pair))   ;; leaf-environment (symbol . value)

;;       (define-public type-leaf-environment
;; 	(make-type "leaf environment"
;; 		   #:supertype type-environment
;; 		   #:test leaf-environment?))

;;       (define-public type-eval-environment
;; 	(make-type "eval environment"
;; 		   #:supertype type-environment
;; 		   #:test eval-environment?))

;;       (define-public type-import-environment
;; 	(make-type "import environment"
;; 		   #:supertype type-environment
;; 		   #:test import-environment?))

;;       (define-public type-export-environment
;; 	(make-type "export environment"
;; 		   #:supertype type-environment
;; 		   #:test export-environment?))

;;       ;; list of ((SYMBOL [mutable-location|immutable-locatation]...)
;;       ;;          (SYMBOL ...))
;;       (define type-export-environment-signature
;; 	(type-list-of (type-list-of type-symbol))))

;;     (lint-types! environment-cell   'pure type-environment-cell
;; 		 type-environment type-symbol type-any)
;;     (lint-types! environment-define type-unspecified
;; 		 type-environment type-symbol type-any)
;;     (lint-types! environment-ref    'pure type-any
;; 		 type-environment type-symbol)
;;     (lint-types! environment-set!   type-unspecified
;; 		 type-environment type-symbol type-any)
;;     (lint-types! environment-undefine type-unspecified
;; 		 type-environment type-symbol)
;;     (lint-types! environment-unobserve type-unspecified
;; 		 type-environment type-observer)

;;     (lint-types! make-leaf-environment  'pure type-leaf-environment)
;;     (lint-types! leaf-environment?      'pure type-boolean type-any)

;;     (lint-types! make-eval-environment          'pure type-eval-environment
;; 		 type-environment type-environment)
;;     (lint-types! eval-environment?              'pure type-boolean type-any)
;;     (lint-types! eval-environment-local         'pure type-environment
;; 		 type-eval-environment)
;;     (lint-types! eval-environment-set-local!     type-unspecified
;; 		 type-eval-environment type-environment)
;;     (lint-types! eval-environment-imported      'pure type-environment
;; 		 type-eval-environment)
;;     (lint-types! eval-environment-set-imported!  type-unspecified
;; 		 type-eval-environment type-environment)

;;     (lint-types! import-environment?         'pure type-boolean type-any)
;;     (lint-types! import-environment-imports  'pure (type-list-of type-environment)
;; 		 type-import-environment)
;;     (lint-types! import-environment-set-imports!  type-unspecified
;; 		 type-import-environment (type-list-of type-environment))

;;     (lint-types! make-export-environment     'pure type-export-environment
;; 		 type-environment type-export-environment-signature)
;;     (lint-types! export-environment?         'pure type-boolean type-any)
;;     (lint-types! export-environment-private  'pure type-environment
;; 		 type-export-environment)
;;     (lint-types! export-environment-set-private!  type-unspecified
;; 		 type-export-environment type-environment)
;;     (lint-types! export-environment-signature
;; 		 'pure type-export-environment-signature
;; 		 type-export-environment)
;;     (lint-types! export-environment-set-signature!  type-unspecified
;; 		 type-export-environment type-export-environment-signature)

;;     (lint-types! memoized-environment 'pure type-environment type-any)


;;     ;;-----------------------------------------------------------------------
;;     ;; make-import-environment

;;     (lint-types! make-import-environment  'pure type-import-environment
;; 		 (type-list-of type-environment) type-procedure)

;;     (define (lint-make-import-environment where now expr env)
;;       (lint-procexpr-arglist where now (third expr) env '(env symbol imports-list))
;;       (lint-list where now expr env))

;;     (lint-handler! make-import-environment lint-make-import-environment)


;;     ;;-----------------------------------------------------------------------
;;     ;; environment-fold

;;     (lint-types! environment-fold   type-any
;; 		 type-environment type-procedure type-any)

;;     (define (lint-environment-fold where now expr env)
;;       (lint-procexpr-arglist where now (third expr) env '(symbol value prev))
;;       (lint-list where now expr env))

;;     (lint-handler! environment-fold lint-environment-fold)


;;     ;;-----------------------------------------------------------------------
;;     ;; environment-observe
;;     ;; environment-observe-weak

;;     (lint-types! environment-observe  type-observer
;; 		 type-environment type-procedure)
;;     (lint-types! environment-observe-weak  type-observer
;; 		 type-environment type-procedure)

;;     (define (lint-environment-observe where now expr env)
;;       (lint-procexpr-arglist where now (third expr) env '(env))
;;       (lint-list where now expr env))

;;     (lint-handler! environment-observe      lint-environment-observe)
;;     (lint-handler! environment-observe-weak lint-environment-observe)
;;     )


;;-----------------------------------------------------------------------------
;; eq?

;; primitive is (1 0 #t), but one arg is always true and doesn't make sense
(lint-arity! eq? '(2 0 #t))

(lint-types! eq?  'pure type-boolean type-any)

(define (bignum? obj)
  (and (integer? obj)
       (or (> obj most-positive-fixnum)
	   (< obj most-negative-fixnum)
	   ;; hard code limits from 32-bit systems, so we detect bad cases
	   ;; for those even when on a 64-bit system
	   (> obj 536870911)
	   (< obj -536870912))))

(define (fracnum? obj)
  (and (rational? obj)
       (not (integer? obj))))

(define (flonum? obj)
  (and (number? obj)
       (inexact? obj)))

;; return true if OBJ is a form (quote X) of the given TYPE?
;; FIXME: Check the binding of "quote" in the current environment?
(define (quoted-type? type? obj)
  (and (pair? obj)
       (eq? 'quote (car obj))
       (pair? (cdr obj))
       (type? (second obj))))

;; return true if OBJ is a TYPE? or a form (quote X) of that TYPE?
(define (type-or-quoted-type? type? obj)
  (or (type? obj)
      (quoted-type? type? obj)))

;; return true if OBJ is a vector, or a (quote #(...)) quoted vector
(define (vector-or-quoted-vector? obj)
  (type-or-quoted-type? vector? obj))

;; return true if OBJ is a quoted pair
(define (quoted-pair? obj)
  (quoted-type? pair? obj))

(define lint-eq?-r5rs-alist
  `((,number? . "implementation dependent on numbers")
    (,char?   . "implementation dependent on characters")
    (,vector-or-quoted-vector? . "not for literal vectors")
    (,quoted-pair?             . "not for literal lists")
    (,string?                  . "not for literal strings")))

;; in guile flonums, bignums and fractions are non-immediate hence not for eq?
;; in guile 1.6 fractions are merely flonums, so inexact? is ahead of fracnum?
(define lint-eq?-guile-alist
  `((,bignum?  . "not for big numbers")
    (,flonum?  . "not for float numbers")
    (,fracnum? . "not for fractions")
    (,vector-or-quoted-vector? . "not for literal vectors")
    (,quoted-pair?             . "not for literal lists")
    (,string?                  . "not for literal strings")))

;; in guile 1.6.7 array? is useless, it's true for everything; we can't
;; identify literal arrays if it doesn't work
(if (not (array? car))
    (begin
      (set! lint-eq?-r5rs-alist  (cons `(,array? . "not for literal arrays")
				       lint-eq?-r5rs-alist))
      (set! lint-eq?-guile-alist (cons `(,array? . "not for literal arrays")
				       lint-eq?-guile-alist))))

(define (lint-eq? where now expr env)
  (for-each (lambda (arg)
	      (and-let* ((pair (find (lambda (pair)
				       ((car pair) arg))
				     (if option-r5rs
					 lint-eq?-r5rs-alist
					 lint-eq?-guile-alist))))
		(lint-error where "`eq?' is ~a: ~s" (cdr pair) arg)))
	    (cdr expr))
  (lint-list where now (cdr expr) env))

(lint-handler! eq? lint-eq?)


;;-----------------------------------------------------------------------------
;; equal?

;; primitive is (1 0 #t), but one arg is always true and doesn't make sense
(lint-arity! equal? '(2 0 #t))

(lint-types! equal?  'pure type-boolean type-any)


;;-----------------------------------------------------------------------------
;; eqv?

;; primitive is (1 0 #t), but one arg is always true and doesn't make sense
(lint-arity! eqv? '(2 0 #t))

(lint-types! eqv?  'pure type-boolean type-any)

(define lint-eqv?-alist
  `((,vector-or-quoted-vector? . "not for literal vectors")
    (,quoted-pair?             . "not for literal lists")
    (,string?                  . "not for literal strings")))

;; in guile 1.6.7 array? is useless, it's true for everything; we can't
;; identify literal arrays if it doesn't work
(if (not (array? car))
    (set! lint-eqv?-alist (cons `(,array? . "not for literal arrays")
				lint-eqv?-alist)))

(define (lint-eqv? where now expr env)
  (for-each (lambda (arg)
	      (and-let* ((pair (find (lambda (pair)
				       ((car pair) arg))
				     lint-eqv?-alist)))
		(lint-error where "`eqv?' is ~a: ~s" (cdr pair) arg)))
	    (cdr expr))
  (lint-list where now (cdr expr) env))

(lint-handler! eqv? lint-eqv?)


;;-----------------------------------------------------------------------------
;; error

;; usually a string for the message, but anything (eg. a symbol) is allowed
(lint-types! error  type-unspecified type-any type-any)

(define (lint-error-proc where now expr env)

  ;; it's possible an error message really has a literal ~a or ~s, but much
  ;; more likely it's a mistaken expectation that it works like format
  (if (list>=? expr 2)
      (for-each (lambda (msg)
		  (if (or (string-contains-ci msg "~a")
			  (string-contains-ci msg "~s"))
		      (lint-error where "`error' message string doesn't take `~~a' or `~~s' formatting: ~s" msg)))
		(string-gettext-list (second expr) env)))

  (lint-list where now (cdr expr) env))

(lint-handler! error lint-error-proc)


;;-----------------------------------------------------------------------------
;; export

;; empty export list is allowed, not sure if that's sensible but no need to
;; enforce more than guile demands
(lint-arity! export '(0 0 #t))

(define (form-export where expr)
  (or (and (list? expr)
	   (every symbol? expr))
      (lint-error where "`export' arguments must be symbols: ~s" expr)))

(define (lint-export where now expr env)
  (if (form-export where expr)
      (begin
	(lint-require-toplevel where now expr env)
	;; reference symbols, not `now' since they can be defined later
	(lint-list where #f (cdr expr) env))))

(lint-handler! export lint-export)


;;-----------------------------------------------------------------------------
;; false-if-exception

;; better error message for macro
(lint-arity! false-if-exception '(1 0 #f))


;;-----------------------------------------------------------------------------
;; filter
;; filter!

(lint-types! filter         type-list type-procedure type-list)
(lint-types! filter!        type-list type-procedure type-list)

(lint-handler! filter lint-for-each)
(lint-handler! filter! lint-for-each)


;;-----------------------------------------------------------------------------
;; for-each

(lint-handler! core:for-each lint-for-each)
(lint-types! core:for-each  type-unspecified type-procedure type-list)


;;-----------------------------------------------------------------------------
;; future

(if (defined? 'future)  ;; guile 1.8, but withdrawn
    (begin

      ;; execution begins immediately, so keep NOW
      (define (lint-future where now expr env)
	(lint-list where now (cdr expr) env))

      ;; macro
      (lint-arity! future '(1 0 #f))

      (lint-types! future  type-any type-any)
      (lint-handler! future lint-future)))


;;-----------------------------------------------------------------------------
;; hash-fold

(lint-types! hash-fold  type-any type-procedure type-any type-hashtable)

(define (lint-hash-fold where now expr env)
  (lint-procexpr-arglist where now (second expr) env '(key value prev))
  (lint-list where now expr env))

(lint-handler! hash-fold lint-hash-fold)


;;-----------------------------------------------------------------------------
;; hash-for-each
;; hash-for-each-handle

(if (defined? 'hash-for-each) ;; not in guile 1.6.7
    (begin

      (define (lint-hash-for-each where now expr env)
	(lint-procexpr-arglist where now (second expr) env '(key value))
	(lint-list where now expr env))

      (lint-types! hash-for-each         type-unspecified
		   type-procedure type-hashtable)
      (lint-types! hash-for-each-handle  type-unspecified
		   type-procedure type-hashtable)

      (lint-handler! hash-for-each lint-hash-for-each)
      (lint-handler! hash-for-each-handle lint-hash-for-each)))


;;-----------------------------------------------------------------------------
;; hash-map->list

(if (defined? 'hash-map->list) ;; guile 1.8
    (begin
      (lint-types! hash-map->list  type-list type-procedure type-hashtable)
      (lint-handler! hash-map->list lint-hash-for-each)))


;;-----------------------------------------------------------------------------
;; if

;; primitive macro, no arity
(lint-arity! if '(2 1 #f))

;; #t, #f or 'unknown
;; test (defined? 'foo) tests for foo
;;
(define (lint-if-condition where now expr env)
  (set! where (where-follow where expr))
  (cond
   ;; form "(and x y z ...)"
   ((and (list? expr)
	 (length>=? expr 1)
	 (eq? and (lint-value where (first expr) env)))
    ;; all #t gives #t, any non-#t (ie. #f or 'unknown) stops at that point
    (let more ((args (cdr expr)))
      (or (null? args)
	  (let ((this (lint-if-condition where now (first args) env)))
	    (if (eq? #t this)
		(more (cdr args))
		this)))))

   ;; form "(defined? 'foo)", being evaluated "now", probably as a
   ;; top-level conditional
   ((and now
	 (list? expr)
	 (= 2 (length expr))
	 (eq? defined? (lint-value where (first expr) env))
	 (quoted-type? symbol? (second expr)))
    (->bool (env-defined? env (second (second expr)))))

   ;; form "(not something)"
   ((and (list? expr)
	 (= 2 (length expr))
	 (eq? not (lint-value where (first expr) env)))
    (let ((ret (lint-if-condition where now (second expr) env)))
      (if (boolean? ret)
	  (not ret)
	  ret)))

   ;; form "(or x y z ...)"
   ((and (list? expr)
	 (length>=? expr 1)
	 (eq? or (lint-value where (first expr) env)))
    ;; looking for any non-#f, ie. either #t or 'unknown
    (any (lambda (arg)
	   (lint-if-condition where now arg env))
	 (cdr expr)))

   ;; form "(provided? 'something)" or "(feature? 'something)"
   ((and (list? expr)
	 (= 2 (length expr))
	 (memq (lint-value where (first expr) env)
	       (list feature? core:provided?))
	 (quoted-type? symbol? (second expr)))
    (core:provided? (second (second expr))))

   ;; literal #t or #f
   ((boolean? expr)
    expr)

   (else
    'unknown)))

(define (lint-if where now expr env)
  (case (lint-if-condition where now (second expr) env)
    ((#t)
     (lint-object where now (second expr) env)       ;; condition
     (lint-object where now (third expr) env))       ;; true leg
    ((#f)
     (lint-object where now (second expr) env)       ;; condition
     (if (length>=? expr 4)
	 (lint-object where now (fourth expr) env))) ;; false leg
    (else
     (lint-list where now (cdr expr) env))))         ;; all parts

(lint-handler! if lint-if)


;;-----------------------------------------------------------------------------
;; lambda

(lint-arity! lambda '(2 0 #t))
(lint-types! lambda 'pure type-any)

;; used also by form-case-lambda
(define-public (form-lambda where name expr)
  (and (or (length>=? expr 3)
	   (lint-error where "`~a' must have bindings and at least one expression: ~s" name expr))
       (or (symbol? (second expr))
	   (list-of-symbols-improper? (second expr))
	   (lint-error (where-follow where (second expr))
		       "`~a' formal parameters must be a symbol or a list of symbols: ~s" name (second expr)))))

(define (lint-lambda where now expr env)
  (if (form-lambda where 'lambda expr)
      (let ((formals (second expr)))
	(lint-formals-duplicates where 'lambda formals)
	(lint-subenv env
		     (lambda (subenv)
		       (env-bind subenv where formals #t #t) ;; args used
		       (lint-body where #f (cddr expr) subenv))))))
(lint-handler! lambda lint-lambda)


;;-----------------------------------------------------------------------------
;; let

(lint-arity! let '(2 0 #t))

;; check bindings form for a let, let*, etc
;; BINDING-LIST doesn't have to be a list, that will be checked
(define-public (form-bindings where name binding-list)
  (set! where (where-follow where binding-list))
  (and (or (list? binding-list)
	   (lint-error where "`~a' bindings must be a (proper) list: ~s"
		       name binding-list))
       (apply and?
	      (iterate-where
	       map-in-order where binding-list
	       (lambda (where binding)
		 (or (and (list? binding)
			  (= 2 (length binding))
			  (symbol? (first binding)))
		     (lint-error where "`~a' binding must be symbol and value: ~s" name binding)))))))

(define (form-let where expr)
  (define (pseudo-named-let-func name bindings form . rest)
    #f)
  (and (or (not (symbol? (second expr)))   ;; extra arity if named let
	   (form-arity where #f pseudo-named-let-func expr))
       (form-bindings where (first expr)
		      (if (symbol? (second expr))
			  (third expr)
			  (second expr)))))

;; also used for `letpar'
(define-public (lint-let where now expr env)
  (if (form-let where expr)
      (let ((name         (first expr))
	    (loop         #f)
	    (binding-list (second expr)))
	;; named let
	(if (symbol? binding-list)
	    (begin
	      (set! loop binding-list)
	      (set! expr (cdr expr))
	      (set! binding-list (second expr))))
	(lint-bindings-duplicates where name binding-list)
	;; values
	(for-each-bindings where binding-list
			   (lambda (where sym val)
			     (lint-object where now val env)))
	(lint-subenv env
		     (lambda (subenv)
		       ;; bindings
		       (if loop
			   (env-bind-one subenv where loop
					 (lint-value where
						     (list lambda 
							   (map car
								binding-list)
							   #f)
						     env)))
		       (for-each-bindings where (second expr)
					  (lambda (where sym val)
					    (env-bind-one subenv where sym
							  (lint-value where
								      val
								      env))))
		       ;; body
		       (lint-body where now (cddr expr) subenv))))))
(lint-handler! let lint-let)


;;-----------------------------------------------------------------------------
;; let*

(lint-arity! let* '(2 0 #t))

(define (form-let* where expr)
  (form-bindings where (first expr) (second expr)))

;; must create a sub-environment even when bindings is an empty list, so as
;; to localize defines
(define (lint-let* where now expr env)
  (if (form-let* where expr)
      (let loop ((bindings    (second expr))
		 (env         env)
		 (subenv-list '()))
	(lint-subenv env
		     (lambda (subenv)
		       (if (null? bindings)
			   (lint-body where now (cddr expr) subenv)

			   (let* ((binding (first bindings))
				  (sym     (first binding))
				  (val     (second binding))
				  (where   (where-follow where binding)))
			     (lint-object where now val env)
			     (env-bind-one subenv where sym
					   (lint-value where val env))
			     (loop (cdr bindings) subenv (cons subenv subenv-list)))))))))

(lint-handler! let* lint-let*)


;;-----------------------------------------------------------------------------
;; letrec

(lint-arity! letrec '(2 0 #t))

(define (form-letrec where expr)
  (form-bindings where (first expr) (second expr)))

(define (lint-letrec where now expr env)
  (if (form-letrec where expr)
      (begin
	(lint-bindings-duplicates where 'letrec (second expr))
	(lint-subenv env
	  (lambda (subenv)
	    ;; variables bound before values tested, so arity of
	    ;; calls to them can be checked
	    (for-each-bindings where (second expr)
			       (lambda (where sym val)
				 (env-bind-one
				  subenv where sym
				  (lint-value where val subenv))))
	    ;; values
	    (for-each-bindings where (second expr)
			       (lambda (where sym val)
				 (lint-defining sym
				   (lambda ()
				     (lint-object where now val subenv)))))
	    ;; body
	    (lint-body where now (cddr expr) subenv))))))
(lint-handler! letrec lint-letrec)


;;-----------------------------------------------------------------------------
;; load

(lint-types! load  type-unspecified type-string type-procedure)

(define (load-relative-filename from filename)
  (if (and (not (string-prefix? "/" filename))
	   (not (string=? (dirname from) ".")))
      (string-append (dirname from) "/" filename)
      filename))

(define (lint-load where now expr env)
  (if (string? (second expr))
      (let ((port (open-input-file
		   (load-relative-filename lint-current-filename
					   (second expr)))))
	(do ((obj (read-object port) (read-object port)))
	    ((eof-object? obj))
	  (lint-object (where-port-object port obj)
		       #t obj env))))
  (if (list>=? expr 3)
      (lint-procexpr-arglist where now (third expr) env '(port)))
  (lint-list where now expr env))

(lint-handler! load lint-load)


;;-----------------------------------------------------------------------------
;; load-extension

(lint-types! load-extension  type-unspecified type-string type-string)

(define-public (lint-load-extension where now expr env)
  (lint-list where now expr env)
  (if (not (and (string? (second expr))
		(string? (third expr))))
      (lint-warn where "warning, cannot lint `load-extension' on non-literal arguments: ~s" expr)

      (let* ((filename  (second expr))
	     (init-func (third expr))
	     (m         (make-module))
	     (m-public  (make-module))
	     (usemod    (usemod-new m where #f)))
	;; must have a public interface for SCM_DEFINE_PUBLIC to work
	(set-module-public-interface! m m-public)
	(set-module-name! m        (format #f "load-extension ~s" filename))
	(set-module-name! m-public (format #f "load-extension (public) ~s"
					   filename))
	(save-module-excursion
	 (lambda ()
	   (set-current-module m)
	   (catch 'misc-error
	     (lambda ()
	       (load-extension filename init-func))
	     (lambda (key func msg args rest)
	       (lint-error where "cannot load extension")
	       (lint-error where "ERROR: ~s ~s: ~a"
			   key func (apply simple-format #f msg args))))))

	(module-for-each
	 (lambda (sym var)
	   (if (not (eq? '%module-public-interface sym))
	       (let ((public? (module-bound? m-public sym)))
		 (if public?  ;; publics exported mean the module is used
		     (usemod-used! usemod #t))
		 (env-bind-one env
			       (where-follow where (variable-where var))
			       sym
			       (and (variable-bound? var)
				    (variable-ref var))
			       public?  ;; an exported public binding is used
			       usemod))))
	 m)
	(env-usemod-list! env (cons usemod (env-usemod-list env))))))

(lint-handler! load-extension lint-load-extension)


;;-----------------------------------------------------------------------------
;; make-guardian

(lint-types! make-guardian  'pure type-guardian type-any)

;; the created [guardian] is pure if called with no args, and not if not, if
;; there was a way to express that ...
(let ()
  (define* ([guardian] #:optional obj) #f)
  (lint-types! [guardian] type-any type-any)
  (lint-return! make-guardian (lambda (expr)
				[guardian])))


;;-----------------------------------------------------------------------------
;; make-list

(lint-types! make-list    'pure type-list type-exact-integer-nonneg type-any)

;; only (1 0 #t) in guile 1.6.7 boot-9.scm (ok in 1.8 though)
(lint-arity! make-list '(1 1 #f))
(lint-defaults! make-list #f '())


;;-----------------------------------------------------------------------------
;; make-object-property

(lint-types! make-object-property  'pure type-object-property)

(let ()
  (define ([object-property] obj) #f)
  (lint-types! [object-property]  'pure type-any type-any)
  (lint-return! make-object-property (lambda (expr)
				       [object-property])))


;;-----------------------------------------------------------------------------
;; make-procedure-with-setter

(lint-types! make-procedure-with-setter  'pure type-procedure
	     type-procedure type-procedure)

(define (lint-make-procedure-with-setter where now expr env)
  ;; GETTER can be called with no args
  (lint-procexpr-arglist where now (second expr) env 'args)
  ;; SETTER gets at least one arg (the value set)
  (lint-procexpr-arglist where now (third expr) env '(arg . args))
  (lint-list where now expr env))

(lint-handler! make-procedure-with-setter lint-make-procedure-with-setter)


;;-----------------------------------------------------------------------------
;; make-record-type

(lint-arity! make-record-type '(2 1 #f))

;; string name is documented, but the code takes a symbol too and that's
;; used by the srfi-9 implementation
(lint-types! make-record-type
	     'pure type-rectype  (type-or type-string type-symbol) type-list)


;;-----------------------------------------------------------------------------
;; make-shared-array

;; only (2 0 #t) in unif.c, but zero dims doesn't really make sense
(lint-arity! make-shared-array '(3 0 #t))

(lint-types! make-shared-array  'pure type-array
	     type-array type-procedure type-arraybound)

(define (lint-make-shared-array where now expr env)
  (lint-procexpr-arglist where now (third expr) env
			 (make-list (length (cdddr expr)) 'idx))
  (lint-list where now expr env))

(lint-handler! make-shared-array lint-make-shared-array)


;;-----------------------------------------------------------------------------
;; map
;; map-in-order

(lint-types! core:map          type-list type-procedure type-list)
(lint-types! core:map-in-order type-list type-procedure type-list)

(lint-handler! core:map          lint-for-each)
(lint-handler! core:map-in-order lint-for-each)


;;-----------------------------------------------------------------------------
;; memq

(lint-types! memq 'pure type-list type-any type-list)
(lint-handler! memq lint-assq)


;;-----------------------------------------------------------------------------
;; memv

(lint-types! memv 'pure type-list type-any type-list)
(lint-handler! memv lint-assv)


;;-----------------------------------------------------------------------------
;; merge
;; merge!

(lint-types! merge   'pure type-list type-list type-list type-procedure)
(lint-types! merge!  type-list type-list type-list type-procedure)

(define (lint-merge where now expr env)
  (lint-procexpr-arglist where now (fourth expr) env '(a b))
  (lint-list where now expr env))

(lint-handler! merge lint-merge)
(lint-handler! merge! lint-merge)


;;-----------------------------------------------------------------------------
;; open-file

(lint-types! open-file  'pure type-port type-string type-open-file-mode-string)

;; return true if the MODE string implies writing (ie. w, a or +)
(define (open-file-mode-output? mode)
  (or (char=? #\w (string-ref mode 0))
      (char=? #\a (string-ref mode 0))
      (string-index mode #\+ 1)))

(define (lint-open-file-mode where mode)
  (if (zero? (string-length mode))
      (lint-error where "`open-file' mode string must not be empty")
      (begin
	(or (string-index "rwa" (string-ref mode 0))
	    (lint-error where "`open-file' mode string must start r, w or a: ~s" mode))
	(string-for-each (lambda (c)
			   (or (string-index "+0lb" c)
			       (lint-error where "`open-file' invalid flag `~a': ~s" c mode)))
			 mode 1)
	(or (not (string-index mode #\0 1))
	    (not (string-index mode #\l 1))
	    (lint-error where "`open-file' flags `0' and `l' are mutually exclusive: ~s" mode))
	(or (open-file-mode-output? mode)
	    (not (string-index mode #\l 1))
	    (lint-error where "`open-file' flag `l' doesn't apply to input: ~s" mode)))))

(define (lint-open-file where now expr env)
  ;; check mode argument if it's a literal string, which is quite likely
  (let ((mode (third expr)))
    (if (string? mode)
	(lint-open-file-mode where mode)))
  (lint-list where now (cdr expr) env))

(lint-handler! open-file lint-open-file)


;;-----------------------------------------------------------------------------
;; or

(lint-arity! or '(0 0 #t))

;; stop looking after literal #t, to ignore tests deliberately dummied out
(define (lint-or where now expr env)
  (case (length expr)
    ((1) (lint-warn where "warning, `or' always true on no arguments"))
    ((2) (lint-warn where "warning, `or' unnecessary on one argument")))
  (let more ((lst (cdr expr)))
    (or (null? lst)
	(case (lint-if-condition where now (first lst) env)
	  ((#t)
	   ;; compile-time true, stop testing
	   #f)
	  (else
	   (lint-object where now (first lst) env)
	   (more (cdr lst)))))))

(lint-handler! or lint-or)


;;-----------------------------------------------------------------------------
;; port-for-each

(lint-types! port-for-each  type-unspecified type-procedure)

(define (lint-port-for-each where now expr env)
  (lint-procexpr-arglist where now (second expr) env '(port))
  (lint-list where now expr env))

(lint-handler! port-for-each lint-port-for-each)


;;-----------------------------------------------------------------------------
;; primitive-make-property

(lint-types! primitive-make-property  'pure type-primprop type-procedure-or-f)

(define (lint-primitive-make-property where now expr env)
  (lint-procexpr-arglist where now (second expr) env '(prop obj))
  (lint-list where now expr env))

(lint-handler! primitive-make-property lint-primitive-make-property)


;;-----------------------------------------------------------------------------
;; procedure->macro
;; procedure->memoizing-macro
;; procedure->syntax

(define (lint-procedure->macro where now expr env)
  (lint-procexpr-arglist where now (second expr) env '(expr env))
  (lint-list where now expr env))

(lint-types! procedure->macro            'pure type-macro type-procedure)
(lint-types! procedure->memoizing-macro  'pure type-macro type-procedure)
(lint-types! procedure->syntax           'pure type-macro type-procedure)

(lint-handler! procedure->macro           lint-procedure->macro)
(lint-handler! procedure->memoizing-macro lint-procedure->macro)
(lint-handler! procedure->syntax          lint-procedure->macro)


;;-----------------------------------------------------------------------------
;; quasiquote

(lint-arity! quasiquote '(1 0 #t))

(define (form-unquote? where expr)
  (or (and (list? expr)          ;; proper list
	   (= 2 (length expr)))  ;; length 2
      (lint-error where "`unquote' takes just one argument")))

(define (lint-quasiquote where now expr env)
  ;; traverse the tree in EXPR looking for unquotes to lint
  (let traverse ((where where)
		 (expr  (cdr expr))
		 (depth 0))
    (set! where (where-follow where expr))
    (if (pair? expr)
	(cond ((memq (car expr) '(unquote unquote-splicing))
	       (if (zero? depth)
		   (if (form-unquote? where expr)
		       (lint-object where now (second expr) env))
		   (traverse where (cdr expr) (1- depth))))
	      ((eq? (car expr) 'quasiquote)
	       (traverse where (cdr expr) (1+ depth)))
	      (else
	       (traverse where (car expr) depth)
	       (traverse where (cdr expr) depth))))))
	
(lint-handler! quasiquote lint-quasiquote)


;;-----------------------------------------------------------------------------
;; quote

(lint-arity! quote '(1 0 #t))

(define (lint-quote where now expr env)
  (let ((obj (second expr)))
    (cond ((char? obj)
	   (lint-error where "no need to quote characters: ~s" obj))
	  ((number? obj)
	   (lint-error where "no need to quote numbers: ~s" obj))
	  ((memq obj '(#t #f))
	   (lint-error where "no need to quote booleans: ~s" obj))
	  ((string? obj)
	   (lint-error where "no need to quote strings: ~s" obj))))

  (and-let* ((  (pair? (cdr expr)))
	     (x (cadr expr))
	     (  (pair? x))
	     (  (eq? 'quote (car x)))
	     (  (pair? (cdr x)))
	     (x (cadr x))
	     (  (list>=? x 2))
	     (  (eq? 'guile-lint (first x))))
    (lint-hint where x env)))

(lint-handler! quote lint-quote)


;;-----------------------------------------------------------------------------
;; record-accessor

(let ()
  (define ([record-accessor] record) #f)
  (lint-types! [record-accessor]  type-any type-record)

  (lint-types! record-accessor 'pure type-procedure type-rectype type-symbol)
  (lint-return! record-accessor (lambda (expr)
				  [record-accessor])))


;;-----------------------------------------------------------------------------
;; record-constructor

;; only (1 0 #t) in boot-9
(lint-arity! record-constructor '(1 1 #f))

(lint-types! record-constructor 'pure type-procedure
	     type-rectype (type-list-of type-symbol))

(let ()
  ;; of unknown fields
  (define ([record-constructor] . fields) #f)

  ;; A quoted list of field args gets a procedure taking those args.  Use
  ;; just the length of the list, in case it's not symbols or there's
  ;; duplicates.  A quoted list is typical usage, and arises from srfi-9.
  ;;
  (define (lint-record-constructor-return expr)
    (if (and (list>=? expr 3)
	     (quoted-type? list? (third expr)))
	(set-procedure-name!
	 '[record-constructor]
	 (lambda-with-args (map gensym
				(make-list (length (second (third expr)))
					   "field"))))
	[record-constructor]))

  (lint-types! [record-constructor]  'pure type-record type-any)
  (lint-return! record-constructor lint-record-constructor-return))


;;-----------------------------------------------------------------------------
;; record-modifier

(let ()
  (define ([record-modifier] record value) #f)
  (lint-types! [record-modifier]  type-unspecified type-record type-any)

  (lint-types! record-modifier 'pure type-procedure type-rectype type-symbol)
  (lint-return! record-modifier (lambda (expr)
				  [record-modifier])))


;;-----------------------------------------------------------------------------
;; record-predicate

(lint-types! record-predicate 'pure type-predicate type-rectype)

(let ()
  (define ([record-predicate] record) #f)
  (lint-types! [record-predicate]  'pure type-boolean type-any)
  (lint-return! record-predicate (lambda (expr)
				   [record-predicate])))


;;-----------------------------------------------------------------------------
;; restricted-vector-sort!

(lint-types! restricted-vector-sort!  type-unspecified
	     type-vector type-procedure
	     type-exact-integer-nonneg type-exact-integer-nonneg)

(define (lint-sort where now expr env)
  (lint-procexpr-arglist where now (third expr) env '(a b))
  (lint-list where now expr env))

(lint-handler! restricted-vector-sort! lint-sort)


;;-----------------------------------------------------------------------------
;; scm-error

(lint-types! scm-error  type-unspecified
	     type-symbol
	     (type-or type-string type-symbol type-f)
	     type-string type-list type-list-or-f)

(define (lint-scm-error-string where msg)
  (let more ((upto 0))
    (let ((pos (string-index msg #\~ upto)))
      (cond ((not pos)
	     ;; end of string, done
	     )

	    ((= pos (1- (string-length msg)))
	     (lint-error where "`scm-error' message format string ends in `~~': ~s" msg))
	    
	    (else
	     (or (memv (string-ref msg (1+ pos)) '(#\~ #\a #\A #\s #\S))
		 (lint-error where "`scm-error' only takes `~~a' and `~~s' formatting: ~s" msg))
	     (more (+ 2 pos)))))))

(define (lint-scm-error where now expr env)
  (for-each (lambda (msg)
	      (lint-scm-error-string (where-follow where (fourth expr)) msg))
	    (string-gettext-list (fourth expr) env))
  (lint-list where now (cdr expr) env))

(lint-handler! scm-error lint-scm-error)


;;-----------------------------------------------------------------------------
;; set!

(lint-arity! set! '(2 0 #f))

;; VAL is checked before and after new value stored.  If it's an expression
;; then it wants any DST procedure binding in effect before.  If it's a
;; lambda then its contents want the DST procedure binding after.
;;
(define (lint-set! where now expr env)
  (let ((dst (second expr))
	(val (third expr)))
    (or (symbol? dst)
	(pair?   dst)
	(lint-error where "`set!' must have symbol (or setter) and value: ~s" expr))
    (lint-list where now (cdr expr) env)
    (if (symbol? dst)
	(and-let* ((bind  (env-binding env dst)))
	  (let ((value (or (lint-value where val env)
			   #t)))  ;; not #f in bindings
	    ;; nasty hack to prevent `noop' overwriting existing value, for
	    ;; the benefit of (set! init noop) for once-only initializations
	    (if (not (and (not now)
			  (eq? value noop)))
		(begin
		  (binding-where! bind (where-follow where val))
		  (binding-val! bind value))))))
    (lint-object where now val env)))

(lint-handler! set! lint-set!)


;;-----------------------------------------------------------------------------
;; sethostent

(if (defined? 'sethostent) ;; only with networking, not in mingw
    (begin
      ;; only (0 0 #t) in networking.scm
      (lint-arity!    sethostent '(0 1 #f))
      (lint-types!    sethostent type-unspecified type-any)
      (lint-defaults! sethostent #f)))


;;-----------------------------------------------------------------------------
;; setnetent

(if (defined? 'setnetent) ;; only with networking, not in mingw
    (begin
      ;; only (0 0 #t) in networking.scm
      (lint-arity! setnetent '(0 1 #f))
      (lint-types! setnetent type-unspecified type-any)))


;;-----------------------------------------------------------------------------
;; setprotoent

(if (defined? 'setprotoent) ;; only with networking, not in mingw
    (begin
      ;; only (0 0 #t) in networking.scm
      (lint-arity!    setprotoent '(0 1 #f))
      (lint-defaults! setprotoent #f)
      (lint-types!    setprotoent type-unspecified type-any)))


;;-----------------------------------------------------------------------------
;; setservent

(if (defined? 'setservent) ;; only with networking, not in mingw
    (begin
      ;; only (0 0 #t) in networking.scm
      (lint-arity! setservent '(0 1 #f))
      (lint-types! setservent type-unspecified type-any)
      (lint-defaults! setservent #f)))


;;-----------------------------------------------------------------------------
;; simple-format

(lint-types! simple-format  (type-or type-string type-unspecified)
	     (type-or type-port type-boolean) type-string type-any)


;; STR is a string like "~a things" to be validated
;; ARG-COUNT is how many arguments supplied to it
;; OR-MORE is true if there's an unknown 0 or more extra on top of ARG-COUNT
;;
(define* (lint-simple-format-string where str arg-count env #:key or-more)
  (if (negative? arg-count)
      (error "Oops, negative arg-count to lint-simple-format-string"))
  (let ((count   0)
	(str-len (string-length str)))
    (let more ((pos 0))
      (set! pos (string-index str #\~ pos))
      (if pos
	  (begin
	    (set! pos (1+ pos))
	    (if (>= pos str-len)
		(lint-error where "`simple-format' string ends with `~~': ~s" str)
		(begin
		  (case (string-ref str pos)
		    ((#\~ #\%)
		     #f)
		    ((#\a #\A #\s #\S)
		     (set! count (1+ count)))
		    (else
		     (lint-error where "`simple-format' string unrecognised conversion: ~~~a" (string-ref str pos))))
		  (more (1+ pos)))))))

    (or (if or-more
	    (>= count arg-count)
	    (= count arg-count))
	(lint-error where "`simple-format' wrong argument count, format is ~a, got ~a~a"
		    count arg-count (if or-more " or more" "")))))

(define (lint-simple-format where now expr env)
  (if (not (null? (string-gettext-list (second expr) env)))
      (lint-error where "`simple-format' probably missing port argument"))
  (for-each (lambda (str)
	      (if (not (string-any #\~ str))
		  (lint-warn where "hint, format string without `~~' can use `display': ~s" str))
	      (lint-simple-format-string where str (length (cdddr expr)) env))
	    (string-gettext-list (third expr) env))
  (lint-list where now (cdr expr) env))

(define (lint-simple-format-apply where now expr env)
  ;; `(apply simple-format port fmt X...X lst)' has fixed args of len-5
  (if (and (length>=? expr 5)
	   (string? (fourth expr)))
      (lint-simple-format-string where (fourth expr) (- (length expr) 5) env
				 #:or-more #t)))

(lint-handler! simple-format lint-simple-format)
(lint-apply-handler! simple-format lint-simple-format-apply)


;;-----------------------------------------------------------------------------
;; sort
;; sort!

(lint-types! sort   'pure type-list-or-vector type-list-or-vector type-procedure)
(lint-types! sort!  type-list-or-vector type-list-or-vector type-procedure)

(define (lint-sort where now expr env)
  (lint-procexpr-arglist where now (third expr) env '(a b))
  (lint-list where now expr env))

(lint-handler! sort lint-sort)
(lint-handler! sort! lint-sort)


;;-----------------------------------------------------------------------------
;; sort-list
;; sort-list!

(lint-types! sort-list   'pure type-list type-list type-procedure)
(lint-types! sort-list!  type-list type-list type-procedure)

(lint-handler! sort-list  lint-sort)
(lint-handler! sort-list! lint-sort)


;;-----------------------------------------------------------------------------
;; sorted?

(lint-types! sorted?  'pure type-boolean type-list-or-vector type-procedure)

(define (lint-sorted? where now expr env)
  (lint-procexpr-arglist where now (third expr) env '(a b))
  (lint-list where now expr env))

(lint-handler! sorted? lint-sorted?)


;;-----------------------------------------------------------------------------
;; stable-sort
;; stable-sort!

(lint-types! stable-sort   'pure type-list-or-vector
	     type-list-or-vector type-procedure)
(lint-types! stable-sort!  type-list-or-vector
	     type-list-or-vector type-procedure)

(lint-handler! stable-sort lint-sort)
(lint-handler! stable-sort! lint-sort)


;;-----------------------------------------------------------------------------
;; start-stack

(lint-arity! start-stack '(2 0 #f))
(lint-types! start-stack  type-any type-symbol type-any)

(define (lint-start-stack where now expr env)
  (lint-object where now (third expr) env))
(lint-handler! start-stack lint-start-stack)


;;-----------------------------------------------------------------------------
;; string

(lint-types! string 'pure type-string type-character)

(define (lint-string-apply where now expr env)
  (if (= 3 (length expr))
      (lint-warn where "hint, `(apply string LST)' can be `(list->string LST)'")))
(lint-apply-handler! string lint-string-apply)


;;-----------------------------------------------------------------------------
;; string-append

(lint-types! string-append  'pure type-string type-string)

(define (lint-string-append where now expr env)
  (let ((numargs (1- (length expr))))
    (or (>= numargs 2)
	(lint-warn where "warning, `string-append' with ~a arg~p is unusual"
		   numargs numargs)))
  (lint-list where now expr env))
(lint-handler! string-append lint-string-append)

(define (lint-string-append-apply where now expr env)
  (if (and (= 3 (length expr))
	   ;; if srfi-13 in scope in guile 1.6, or always in guile 1.8
	   (eq? string-concatenate (lint-value where 'string-concatenate env)))
      (lint-warn where "hint, `(apply string-append LST)' can be `(string-concatenate LST)'")))
(lint-apply-handler! string-append lint-string-append-apply)


;;-----------------------------------------------------------------------------
;; undefine

(lint-arity! undefine '(1 0 #f))

(lint-types! undefine  type-unspecified type-symbol)

(define (form-undefine where expr)
  (or (and (= 2 (length expr))
	   (symbol? (second expr)))
      (lint-error where "`undefine' must have one symbol: ~s" expr)))

(define (lint-undefine where now expr env)
  (and (form-undefine where expr)
       (or (env-toplevel env)
	   (lint-error where "`undefine' only at top level: ~s" expr))
       (or (env-defined? env (second expr))
	   (lint-error where "`undefine' symbol already unbound: ~s" (second expr)))
       (env-unbind env (second expr))))
(lint-handler! undefine lint-undefine)


;;-----------------------------------------------------------------------------
;; use-modules, use-syntax

;; empty use-modules is not really sensible, but it's not an error
(lint-arity! use-modules '(0 0 #t))

(define (lint-use-modules where now expr env)
  (lint-require-toplevel where now expr env)
  (for-each (lambda (spec)
	      (if (module-exists? spec)
		  (env-bind-module env
				   (where-follow where spec)
				   (apply-resolve-interface spec))
		  (lint-error where "`use-modules' no such module: ~s" spec)))
	    (cdr expr)))

(lint-handler! use-modules lint-use-modules)

;; FIXME: what do we need to do different for use-syntax?
(lint-handler! use-syntax lint-use-modules)


;;-----------------------------------------------------------------------------
;; vector

(lint-types! vector 'pure type-vector type-any)

(define (lint-vector-apply where now expr env)
  (if (= 3 (length expr))
      (lint-warn where "hint, `(apply vector LST)' can be `(list->vector LST)'")))
(lint-apply-handler! vector lint-vector-apply)


;;-----------------------------------------------------------------------------
;; with-fluid*
;; with-fluids*

(define (lint-with-fluid* where now expr env)
  (lint-procexpr-arglist where now (fourth expr) env '())
  (lint-list where now expr env))

(if (defined? 'with-fluid*) ;; new in guile 1.8
    (begin
      (lint-types! with-fluid*  type-any  type-fluid type-any type-thunk)
      (lint-handler! with-fluid*  lint-with-fluid*)))

(lint-types! with-fluids*  type-any
	     (type-list-of type-fluid) type-list type-thunk)
(lint-handler! with-fluids* lint-with-fluid*)


;;-----------------------------------------------------------------------------
;; with-fluids

;; friendlier error message than otherwise get from macro expansion
(lint-arity! with-fluids '(2 0 #t))


;;-----------------------------------------------------------------------------
;; with-input-from-file
;; with-output-to-file
;; with-error-to-file

(lint-types! with-input-from-file   type-any type-string type-thunk)
(lint-types! with-output-to-file    type-any type-string type-thunk)
(lint-types! with-error-to-file     type-any type-string type-thunk)

(define (lint-with-input-from-file where now expr env)
  (lint-procexpr-arglist where now (third expr) env '())
  (lint-body-list where now expr env
		  #:initial-used 2)) ;; macro symbol and filename

(lint-handler! with-input-from-file  lint-with-input-from-file)
(lint-handler! with-output-to-file lint-with-input-from-file)
(lint-handler! with-error-to-file  lint-with-input-from-file)


;;-----------------------------------------------------------------------------
;; while

(lint-arity! while '(1 0 #t))
(lint-types! while type-unspecified type-any)

(define (lint-while where now expr env)
  ;; insert dummy references to break and continue, so they don't look unused
  ;; and also insert into body so it doesn't look like an empty `begin'
  (lint-macro where now while (cons* (first expr)       ;; `while'
				     `(begin
					(break)
					(continue)
					,(second expr)) ;; condition
				     '(break)
				     '(continue)
				     (cddr expr))       ;; body
	      env))

(lint-handler! while lint-while)


;;-----------------------------------------------------------------------------
;; hints

(define (lint-format-string where str arg-count env)
  (let ((f (env-binding-val env 'format)))
    (cond ((eq? f simple-format)
	   (lint-simple-format-string where str arg-count env))
	  ((eq? f format)
	   (lint-ice-9-format-string where str arg-count env))
	  (else
	   (lint-error where "`format' binding unrecognised, cannot use `format' hint")))))

(define lint-hint-format-alist
  `((format        . ,lint-format-string)
    (ice-9-format  . ,(lambda (where str arg-count env) ;; hide from autoloader
			(lint-ice-9-format-string where str arg-count env)))
    (simple-format . ,lint-simple-format-string)))

;; TYPE is 'format, 'simple-format or 'ice-9-format
;; return a lint-handler procedure for checking with COUNT preceding args
(define-public (make-lint-format type count)
  (let ((proc (assq-ref lint-hint-format-alist type)))
    (lambda (where now expr env)
      (if (length>=? expr (+ 2 count))
	  (let ((fargs (list-tail expr (1+ count))))
	    (for-each (lambda (fmt)
			(proc where fmt (1- (length fargs)) env))
		      (string-gettext-list (first fargs) env))))
      (lint-list where now (cdr expr) env))))

;; TYPE is 'format 'simple-format or 'ice-9-format
(define (lint-hint-format where sym type count env)
  (let ((proc (env-binding-val env sym)))
    (cond ((not (integer-non-negative? count))
	   (lint-error where "guile-lint ~a hint: bad count: ~a" type count))
	  ((not (procedure? proc))
	   (lint-error where "guile-lint ~a hint: not a procedure: ~a" type sym))
	  ((lint-handler-property proc)
	   (lint-error where "guile-lint ~a hint: already have a hint: ~a" type sym))
	  (else
	   (lint-handler! proc (make-lint-format type count))))))

(define (lint-hint where lst env)
  (case (second lst)
    ((format ice-9-format simple-format)
     (if (and (or (= 3 (length lst))
		  (= 4 (length lst)))
	      (symbol? (third lst))
	      (or (= 3 (length lst))
		  (number? (fourth lst))))
	 (lint-hint-format where (third lst) (second lst)
			   (if (= 3 (length lst)) 0 (fourth lst))
			   env)
	 (lint-error where "guile-lint ~a hint: should be ''(guile-lint ~a symbol [number])" (second lst) (second lst))))))


;;-----------------------------------------------------------------------------
;; (srfi srfi-31) rec
;;
;; this just loads the module, for lint-value to recognise

(define srfi-31-available?  ;; not in guile 1.6.4
  (catch #t
    (lambda () (module-exists? '(srfi srfi-31)))
    (lambda args #f)))


;;-----------------------------------------------------------------------------
;; srfi-55 require-extension
;;
;; This is in the core when it exists (guile 1.8 and up).

(define (form-require-extension-clause where clause)
  (or (and (list? clause)
	   (not (null? clause))
	   (symbol? (first clause)))
      (lint-error where "`require-extension' clause must be a proper list starting with a symbol: ~s" clause)))

(define (lint-require-extension-srfi where n env)
  (if (and (number? n)
	   (exact? n)
	   (integer? n))
      (let ((spec (list 'srfi (string->symbol
			       (format #f "srfi-~a" n)))))
	(if (module-exists? spec)
	    (env-bind-module env where
			     (apply-resolve-interface spec))
	    (lint-error where "`require-extension' no such srfi: ~s" n)))
      (lint-error where "`require-extension' srfi argument must be a number: ~s" n)))

(define (lint-require-extension-clause where clause env)
  (if (form-require-extension-clause where clause)
      (case (first clause)
	((srfi)
	 (for-each (lambda (n)
		     (lint-require-extension-srfi where n env))
		   (cdr clause)))
	(else
	 (lint-error where "`require-extension' unknown feature identifier: ~s" (first clause))))))

(define (lint-require-extension where now expr env)
  (lint-require-toplevel where now expr env)
  (iterate-where for-each where (cdr expr)
		 (lambda (where clause)
		   (lint-require-extension-clause where clause env))))

(if (defined? 'require-extension) ;; guile 1.8 and up
    (lint-handler! require-extension lint-require-extension))


;;-----------------------------------------------------------------------------
;; special handlers

;; value indicating failed expansion
(define macro-no-expansion (cons 1 2))

(define (apply-macro-transformer where trans expr)
  (let ((new-expr
	 (catch #t
	   (lambda () (trans expr (procedure-environment trans)))
	   (lambda args
	     (apply lint-error-error where "error in macro expansion:" args)
	     macro-no-expansion))))
    (if (equal? expr new-expr)
	(begin
	  (lint-error where "oops, macro expansion didn't change expression: ~a" expr)
	  macro-no-expansion)
	new-expr)))

(define (lint-macro where now macro expr env)
  (let ((trans (macro-transformer macro)))
    (if trans
	(let ((expansion (apply-macro-transformer where trans expr)))
	  (if (not (eq? expansion macro-no-expansion))
	      (lint-object where now expansion env)))

	;; primitive macros are supposed to have handlers, warn when one
	;; without is encountered (but only warn once)
	(if (not (memq macro lint-macro-warned-list))
	    (begin
	      (set! lint-macro-warned-list (cons macro lint-macro-warned-list))
	      (lint-warn where "warning, unknown primitive macro not checked: ~s" (macro-name macro)))))))

(define lint-macro-warned-list '())


(define-public (form-arity where obj-where obj expr)
  (form-arity-n where obj-where (first expr) obj (1- (length expr))))

;; consider a call to OBJ with N arguments
;; N can be a pair (n . 'more) for n or more from an `apply'
(define (form-arity-n where obj-where sym obj n)
  (let ((arity (lint-arity obj)))
    (or (not arity)
	(arity-match? arity n)
	(let ((obj-name (or sym (procedure-name-maybe obj))))
	  (lint-error where "`~a' called with ~a arg~p, expect ~a"
		      obj-name
		      (if (pair? n) (format #f "~a or more" (car n)) n)
		      (if (pair? n) (car n) n)
		      (arity-desc arity))
	  (if obj-where
	      (lint-error obj-where "  `~a' definition is here" obj-name))
	  #f))))

(define (lint-call-arity where proc-where proc-name arity arglist)
  (or (arity-match-arglist? arity arglist)
      (let* ((or-more? (dotted-list? arglist))
	     (n        (- (length (improper->proper arglist))
			  (if or-more? 1 0))))
	(lint-error where "`~a' called with ~a~:[~; or more~] arg~:[s~;~], expect ~a"
		    proc-name n or-more? (and (= 1 n) (not or-more?))
		    (arity-desc arity))
	(if proc-where
	    (lint-error proc-where "  `~a' definition is here" proc-name))
	#f)))



(define (lint-call-arglist where obj-where obj-name obj-value arglist)
  (or (and-let* ((         (procedure? obj-value))
		 (optarity (procedure-property obj-value 'arglist)))
	(lint-call-optarity where obj-where obj-name optarity arglist))

      (and-let* ((arity (lint-arity obj-value)))
	(lint-call-arity where obj-where obj-name arity arglist))))

;; PROC-EXPR is a procedure expression (a symbol, lambda form, actual
;; procedure, etc) which will be called with ARGLIST, at time NOW
;; ARGLIST is a sample list per `arity-match-arglist?'
;;
(define-public (lint-procexpr-arglist where now proc-expr env arglist)
  (set! where (where-follow where proc-expr))
  (if (and (symbol? proc-expr)
	   (not now))
      (env-reference-add env where proc-expr arglist)
      (let ((proc (lint-value where proc-expr env)))
	(if (procedure? proc)
	    (lint-call-arglist where where
			       (procedure-name-maybe proc)
			       proc arglist)))))

(define-public (lint-call where now expr env)
  (let ((obj        (first expr))
	(name       '[unknown-procedure])
	(proc       #f)
	(proc-where #f))

    (cond ((or (macro? obj)
	       (procedure? obj))
	   (set! proc obj)
	   (set! name (procedure-name-maybe proc)))
	  ((symbol? obj)
	   (set! name obj)
	   (and-let* ((bind (env-binding env obj)))
	     (set! proc (binding-val bind))
	     (set! proc-where (binding-where bind)))))

    (and-let* ((defaults  (lint-defaults proc))
	       (arity     (lint-arity proc))
	       (mandatory (first arity))
	       (optional  (second arity))
	       (actuals   (cdr expr))
	       (          (<= (length actuals) (+ mandatory optional))))
      (set! defaults (take defaults (length actuals)))
      (let more ()
	(and (> (length actuals) mandatory)
	     (equal? (last defaults)
		     (lint-value where (last actuals) env))
	     (begin
	       (lint-warn where "hint, `~a' no need for arg ~a `~s', that's the default" name (length actuals) (last defaults))
	       (set! defaults (drop-right defaults 1))
	       (set! actuals  (drop-right actuals 1))
	       (more)))))

    (cond ((lint-handler-property proc)
	   => (lambda (handler)
		(if (symbol? obj) ;; binding used
		    (env-reference-add env where obj #f))
		(if (form-arity-n where proc-where name proc (1- (length expr)))
		    (handler where now expr env))))
	  ((macro? proc)
	   (if (symbol? obj) ;; binding used
	       (env-reference-add env where obj #f))
	   (if (form-arity-n where proc-where name proc (1- (length expr)))
	       (lint-macro where now proc expr env)))
	  (now
	   (if proc
	       (lint-call-arglist where proc-where name proc (cdr expr)))
	   (lint-list where now expr env))
	  ((procedure? obj)
	   (lint-call-arglist where proc-where name proc (cdr expr))
	   (lint-list where now expr env))
	  ((symbol? obj)
	   ;; defer check to end of env, to use local defines
	   (env-reference-add env where name (arglist-flatten (cdr expr)))
	   (lint-list where now (cdr expr) env)))))

;; check each element of the list LST
(define-public (lint-list where now lst env)
  (set! where (where-follow where lst))
  (for-each (lambda (elem)
	      (lint-object (where-follow where elem) now elem env))
	    lst))

;; like `lint-list' but `define' and `define*'s are allowed at the start of
;; the list
;;
(define-public (lint-body where now lst env)
  (set! where (where-follow where lst))
  (lint-subenv
   env
   (lambda (env)
     (let more ((lst lst))
       (if (null? lst)
	   (lint-error where "body cannot be empty")
	   (let ((form (and (list>=? (first lst) 1)
			    (if (symbol? (caar lst))
				(env-binding-val env (caar lst))
				(caar lst)))))
	     (cond ((assq-ref lint-body-define-alist form)
		    => (lambda (handler)
			 ;; reference define* so ice-9 optargs marked as used
			 (if (symbol? (caar lst))
			     (env-reference-add env #f (caar lst) #f))
			 (handler (where-follow where (first lst)) now
				  (first lst) env)
			 (more (cdr lst))))
		   ((eq? form begin)
		    ;; interpolate `begin' form, for internal defines there
		    (let ((new-lst (append (cdar lst) (cdr lst))))
		      (set! where (where-follow where new-lst))
		      (more new-lst)))
		   (else
		    (lint-body-list where now lst env)))))))))

;; check each element of LST, treating them as the forms in a body, which
;; means only the return value from the last is used
;;
(define*-public (lint-body-list where now lst env #:key (initial-used 0))
  (set! where (where-follow where lst))
  (let ((len (length lst))
	(pos 1))
    (for-each (lambda (elem)
		(let ((where (where-follow where elem)))
		  (lint-object where now elem env)
		  (if (and (> pos initial-used)
			   (< pos len))
		      (lint-called-for-effect where now elem env)))
		(set! pos (1+ pos)))
	      lst)))

;; check if EXPR is valid to be called only for its side effects
(define-public (lint-called-for-effect where now expr env)
  (cond ((symbol? expr)
	 (lint-error where "variable evaluated but result unused: ~s" expr))
	((list>=? expr 1)
	 (let ((value (lint-value where (first expr) env)))
	   (if (object-property value 'pure)
	       (lint-error where "pure function called for effect: ~s" expr))))
	(else
	 (lint-error where "literal evaluated but result unused: ~s" expr))))

;; the define forms allowed at the start of a body, and their handler
(define-public lint-body-define-alist
  (list (cons define lint-define-allowed)))

(define (form-expr where expr)
  (or (list? expr)
      (lint-error where "expression must be a proper list: ~s" expr)))

;; an expression, EXPR must be a pair (doesn't need to be a proper list)
(define-public (lint-expr where now expr env)
  (if (form-expr where expr)
      (let ((name (first expr)))
	(cond ((or (symbol? name)
		   (procedure? name)
		   (macro? name))
	       (lint-check-types where expr env)
	       (lint-call where now expr env))
	      ((not (pair? name))
	       (lint-error where "unlikely object to call: ~s" name)
	       (lint-list where now expr env))
	      (else
	       ;; try to eval the called expression, eg. a lambda form
	       (let ((val (lint-value where name env)))
		 (cond ((procedure? val)
			(lint-object where now name env)
			(lint-call where now (cons val (cdr expr)) env))
		       (else
			(lint-list where now expr env)))))))))

(define-public (lint-symbol where now sym env)
  (if (and now
	   (not (env-defined? env sym)))
      (lint-error where "unbound symbol: ~s" sym)
      (env-reference-add env where sym #f)))

(define-public (lint-object where now obj env)
  (set! where (where-follow where obj))
  (cond ((pair? obj)
	 (lint-expr where now obj env))
	((symbol? obj)
	 (lint-symbol where now obj env))
	((and option-r5rs
	      (vector? obj))
	 (lint-error where "literal vectors must be quoted: ~s" obj))))

;; the current top-level environment, either the initial `(guile-user)' or a
;; module created by define-module
(define-public lint-current-module-env #f)

;; list of all top-level environments created by define-module and the
;; initial guile-user
(define lint-modules-env-list '())


;;-----------------------------------------------------------------------------
;; files

(define (read-object port)
  (catch error-reader-key
    (lambda ()
      (read port))
    (lambda (key . args)
      (cond ((error-reader-close-paren? key args)
	     ;; stray ")" is printed, but continue
	     (lint-error-port port "unexpected \")\"")
	     (throw 'lint-continue #t))

	    ((error-reader? key args)
	     ;; other reader errors (eg. bad #) stop processing,
	     ;; don't show final unbounds etc
	     (apply lint-error-port port (second args) (third args))
	     (newline)
	     (throw 'lint-continue #f))

	    (else
	     ;; Something bad in a loaded module, or in guile-lint either
	     ;; way we probably can't sensibly continue.
	     ;;
	     ;; Do our own display-error here, since a re-throw will get
	     ;; printed as if it's guile-lint which is error source.  The
	     ;; diagnostics here are still not great though.
	     ;;
	     (save-stack)
	     (backtrace)
	     (apply display-error the-last-stack (current-error-port)
		    args)
	     (exit 1))))))

(define (lint-make-user-env)
  (let ((env (env-new (if option-r5rs (r5rs-env) guile-env))))

    ;; in 1.6 "(use-modules (ice-9 format))" in does something evil to the
    ;; "(guile)" module, making 'format bound to ice-9 format instead of
    ;; simple-format, so undo that
    (and-let* ((bind (env-binding env 'simple-format)))
      (env-bind-one env (binding-where bind) 'format simple-format #t))

    (env-toplevel! env '(guile-user))
    env))

(define-public lint-finalize-env-hook (make-hook 1))

(define (lint-port port)
  (read-enable 'positions)

  (set! lint-current-module-env (lint-make-user-env))
  (set! lint-modules-env-list (list lint-current-module-env))

  (for-each (lambda (n)
	      (env-bind-module lint-current-module-env #f
			       (apply-resolve-interface
				(list 'srfi (string->symbol
					     (format #f "srfi-~a" n))))
			       #:used #t))
	    option-use-srfi-list)

  ;; The outer 'lint-continue catch returns a thunk to be executed, it's
  ;; either noop when finished or a call to `more' to keep looping.
  ;;
  ;; A catch error handler is not in a tail-call position, so we can't
  ;; just make the call there, but must return a thunk and call that.
  ;;
  ;; The inner reader catch is so we can continue after certain reader
  ;; errors.  But that catch is established only over the read call, it's
  ;; not over the lint-object processing, since we don't want it when
  ;; reading code for use-modules etc.
  ;;
  (let more ()
    (if (catch 'lint-continue
	  (lambda ()
	    (do ((obj (read-object port) (read-object port)))
		((eof-object? obj))
	      (let ((where (where-port-object port obj)))
		(lint-object where #t obj
			     lint-current-module-env)
		(lint-called-for-effect where #t obj
					lint-current-module-env)))
	    (for-each (lambda (env)
			(lint-resolve-local env)
			(lint-finalize-env env)
			(lint-show-unused-modules env)
			(lint-show-unused env)
			(run-hook lint-finalize-env-hook env))
		      lint-modules-env-list)
	    #f)
	  (lambda (key proc)
	    proc))
	(more))))

(define lint-current-filename #f)

(define-public (lint-file filename)
  (set! lint-current-filename filename)
  (if option-verbose
      (format #t "~a:\n" filename))
  (let ((port (false-if-exception (open-input-file filename))))
    (if port
	(lint-port port)
	(format (current-error-port)
		"guile-lint: cannot read file ~a\n" filename))))
